/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef ORC_CONVERT_COLUMN_READER_HH
#define ORC_CONVERT_COLUMN_READER_HH

#include "ColumnReader.hh"
#include "SchemaEvolution.hh"

namespace orc {

  class ConvertColumnReader : public ColumnReader {
   public:
    ConvertColumnReader(const Type& readType, const Type& fileType, StripeStreams& stripe,
                        bool throwOnOverflow);

    // override next() to implement convert logic
    void next(ColumnVectorBatch& rowBatch, uint64_t numValues, char* notNull) override;

    uint64_t skip(uint64_t numValues) override;

    void seekToRowGroup(std::unordered_map<uint64_t, PositionProvider>& positions) override;

   protected:
    bool useTightNumericVector;
    const Type& readType;
    std::unique_ptr<ColumnReader> reader;
    std::unique_ptr<ColumnVectorBatch> data;
    const bool throwOnOverflow;
  };

  std::unique_ptr<ColumnReader> buildConvertReader(const Type& fileType, StripeStreams& stripe,
                                                   bool useTightNumericVector,
                                                   bool throwOnOverflow);

}  // namespace orc

#endif  // ORC_CONVERT_COLUMN_READER_HH
