// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: orc_proto.proto
// Protobuf C++ Version: 7.35.1

#include "orc_proto.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::orc::proto::IntegerStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::DoubleStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StringStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::BucketStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::DecimalStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::DateStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::TimestampStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::BinaryStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::CollectionStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::BoundingBox
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::GeospatialStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::ColumnStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::RowIndexEntry
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::RowIndex
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::BloomFilter
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::BloomFilterIndex
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::Stream
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::ColumnEncoding
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StripeEncryptionVariant
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StripeFooter
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StringPair
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::Type
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StripeInformation
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::UserMetadataItem
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::StripeStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::Metadata
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::ColumnarStripeStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::FileStatistics
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::DataMask
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::EncryptionKey
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::EncryptionVariant
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::Encryption
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::Footer
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::PostScript
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
        // ::orc::proto::FileTail
        {&::_pbi::kDescriptorMethods, &::descriptor_table_orc_5fproto_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace orc {
namespace proto {
class UserMetadataItem::_Internal {
 public:
  using HasBits = decltype(::std::declval<UserMetadataItem>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_._has_bits_);
};

constexpr UserMetadataItem::ParseTableT_ UserMetadataItem::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::UserMetadataItem>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional bytes value = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_.value_)}},
      // optional string name = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_.name_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string name = 1;
      {PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_.name_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional bytes value = 2;
      {PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr UserMetadataItem::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        value_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr UserMetadataItem::UserMetadataItem(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL UserMetadataItem::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) UserMetadataItem(arena);
}
constexpr auto UserMetadataItem::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(UserMetadataItem), alignof(UserMetadataItem));
}
constexpr auto UserMetadataItem::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &UserMetadataItem::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<UserMetadataItem>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &UserMetadataItem::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<UserMetadataItem>(), &UserMetadataItem::ByteSizeLong,
              &UserMetadataItem::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(UserMetadataItem, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[23],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct UserMetadataItemGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr UserMetadataItemGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 UserMetadataItem_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(UserMetadataItem::InternalGenerateClassData_(
            _default, &UserMetadataItem_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<UserMetadataItem>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~UserMetadataItemGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) UserMetadataItem _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<UserMetadataItem>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(UserMetadataItemGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST UserMetadataItemGlobalsTypeInternal UserMetadataItem_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* UserMetadataItem_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return UserMetadataItem_globals_.GetClassData();
#else
  return UserMetadataItem_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class TimestampStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<TimestampStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_._has_bits_);
};

constexpr TimestampStatistics::ParseTableT_ TimestampStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::TimestampStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional sint64 minimum = 1;
      {::_pbi::TcParser::FastZ64S1,
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_)}},
      // optional sint64 maximum = 2;
      {::_pbi::TcParser::FastZ64S1,
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_)}},
      // optional sint64 minimum_utc = 3;
      {::_pbi::TcParser::FastZ64S1,
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_utc_)}},
      // optional sint64 maximum_utc = 4;
      {::_pbi::TcParser::FastZ64S1,
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_utc_)}},
      // optional int32 minimum_nanos = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(TimestampStatistics, _impl_.minimum_nanos_), 4>(),
       {40, 4, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_nanos_)}},
      // optional int32 maximum_nanos = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(TimestampStatistics, _impl_.maximum_nanos_), 5>(),
       {48, 5, 0,
        PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_nanos_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional sint64 minimum = 1;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional sint64 maximum = 2;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional sint64 minimum_utc = 3;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_utc_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional sint64 maximum_utc = 4;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_utc_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional int32 minimum_nanos = 5;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_nanos_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // optional int32 maximum_nanos = 6;
      {PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_nanos_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr TimestampStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_{::int64_t{0}},
        maximum_{::int64_t{0}},
        minimum_utc_{::int64_t{0}},
        maximum_utc_{::int64_t{0}},
        minimum_nanos_{0},
        maximum_nanos_{0} {}

template <typename>
constexpr TimestampStatistics::TimestampStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL TimestampStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) TimestampStatistics(arena);
}
constexpr auto TimestampStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(TimestampStatistics), alignof(TimestampStatistics));
}
constexpr auto TimestampStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &TimestampStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<TimestampStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &TimestampStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<TimestampStatistics>(), &TimestampStatistics::ByteSizeLong,
              &TimestampStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[6],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct TimestampStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr TimestampStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 TimestampStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(TimestampStatistics::InternalGenerateClassData_(
            _default, &TimestampStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<TimestampStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~TimestampStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) TimestampStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<TimestampStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(TimestampStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST TimestampStatisticsGlobalsTypeInternal TimestampStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* TimestampStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return TimestampStatistics_globals_.GetClassData();
#else
  return TimestampStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StripeInformation::_Internal {
 public:
  using HasBits = decltype(::std::declval<StripeInformation>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_._has_bits_);
};

constexpr StripeInformation::ParseTableT_ StripeInformation::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_._has_bits_),
      0, // no _extensions_
      7, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967168,  // skipmap
      offsetof(ParseTableT_, field_entries),
      7,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StripeInformation>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint64 offset = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.offset_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.offset_)}},
      // optional uint64 index_length = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.index_length_), 2>(),
       {16, 2, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.index_length_)}},
      // optional uint64 data_length = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.data_length_), 3>(),
       {24, 3, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.data_length_)}},
      // optional uint64 footer_length = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.footer_length_), 4>(),
       {32, 4, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.footer_length_)}},
      // optional uint64 number_of_rows = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.number_of_rows_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.number_of_rows_)}},
      // optional uint64 encrypt_stripe_id = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(StripeInformation, _impl_.encrypt_stripe_id_), 6>(),
       {48, 6, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.encrypt_stripe_id_)}},
      // repeated bytes encrypted_local_keys = 7;
      {::_pbi::TcParser::FastBR1,
       {58, 0, 0,
        PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.encrypted_local_keys_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint64 offset = 1;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.offset_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 index_length = 2;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.index_length_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 data_length = 3;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.data_length_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 footer_length = 4;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.footer_length_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 number_of_rows = 5;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.number_of_rows_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 encrypt_stripe_id = 6;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.encrypt_stripe_id_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // repeated bytes encrypted_local_keys = 7;
      {PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.encrypted_local_keys_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kBytes | ::_fl::kRepSString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr StripeInformation::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        encrypted_local_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeInformation,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.encrypted_local_keys_)>()
         }
        ,
        offset_{::uint64_t{0u}},
        index_length_{::uint64_t{0u}},
        data_length_{::uint64_t{0u}},
        footer_length_{::uint64_t{0u}},
        number_of_rows_{::uint64_t{0u}},
        encrypt_stripe_id_{::uint64_t{0u}} {}

template <typename>
constexpr StripeInformation::StripeInformation(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StripeInformation::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StripeInformation(arena);
}
constexpr auto StripeInformation::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StripeInformation), alignof(StripeInformation));
}
constexpr auto StripeInformation::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StripeInformation::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StripeInformation>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StripeInformation::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StripeInformation>(), &StripeInformation::ByteSizeLong,
              &StripeInformation::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[22],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StripeInformationGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StripeInformationGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StripeInformation_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StripeInformation::InternalGenerateClassData_(
            _default, &StripeInformation_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StripeInformation>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StripeInformationGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StripeInformation _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StripeInformation>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StripeInformationGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StripeInformationGlobalsTypeInternal StripeInformation_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StripeInformation_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StripeInformation_globals_.GetClassData();
#else
  return StripeInformation_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StringStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<StringStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_._has_bits_);
};

constexpr StringStatistics::ParseTableT_ StringStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StringStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string minimum = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.minimum_)}},
      // optional string maximum = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.maximum_)}},
      // optional sint64 sum = 3;
      {::_pbi::TcParser::FastZ64S1,
       {24, 4, 0,
        PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.sum_)}},
      // optional string lower_bound = 4;
      {::_pbi::TcParser::FastBS1,
       {34, 2, 0,
        PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.lower_bound_)}},
      // optional string upper_bound = 5;
      {::_pbi::TcParser::FastBS1,
       {42, 3, 0,
        PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.upper_bound_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string minimum = 1;
      {PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional string maximum = 2;
      {PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional sint64 sum = 3;
      {PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.sum_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional string lower_bound = 4;
      {PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.lower_bound_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional string upper_bound = 5;
      {PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_.upper_bound_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr StringStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        maximum_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        lower_bound_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        upper_bound_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        sum_{::int64_t{0}} {}

template <typename>
constexpr StringStatistics::StringStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StringStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StringStatistics(arena);
}
constexpr auto StringStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StringStatistics), alignof(StringStatistics));
}
constexpr auto StringStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StringStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StringStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StringStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StringStatistics>(), &StringStatistics::ByteSizeLong,
              &StringStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StringStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StringStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StringStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StringStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StringStatistics::InternalGenerateClassData_(
            _default, &StringStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StringStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StringStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StringStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StringStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StringStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StringStatisticsGlobalsTypeInternal StringStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StringStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StringStatistics_globals_.GetClassData();
#else
  return StringStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StringPair::_Internal {
 public:
  using HasBits = decltype(::std::declval<StringPair>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StringPair, _impl_._has_bits_);
};

constexpr StringPair::ParseTableT_ StringPair::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StringPair, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StringPair>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional string value = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(StringPair, _impl_.value_)}},
      // optional string key = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StringPair, _impl_.key_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string key = 1;
      {PROTOBUF_FIELD_OFFSET(StringPair, _impl_.key_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional string value = 2;
      {PROTOBUF_FIELD_OFFSET(StringPair, _impl_.value_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr StringPair::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        key_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        value_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr StringPair::StringPair(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StringPair::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StringPair(arena);
}
constexpr auto StringPair::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StringPair), alignof(StringPair));
}
constexpr auto StringPair::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StringPair::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StringPair>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StringPair::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StringPair>(), &StringPair::ByteSizeLong,
              &StringPair::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StringPair, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[20],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StringPairGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StringPairGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StringPair_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StringPair::InternalGenerateClassData_(
            _default, &StringPair_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StringPair>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StringPairGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StringPair _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StringPair>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StringPairGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StringPairGlobalsTypeInternal StringPair_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StringPair_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StringPair_globals_.GetClassData();
#else
  return StringPair_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Stream::_Internal {
 public:
  using HasBits = decltype(::std::declval<Stream>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Stream, _impl_._has_bits_);
};

constexpr Stream::ParseTableT_ Stream::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Stream, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::Stream>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional .orc.proto.Stream.Kind kind = 1;
      {::_pbi::TcParser::FastEvS1,
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(Stream, _impl_.kind_)}},
      // optional uint32 column = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Stream, _impl_.column_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Stream, _impl_.column_)}},
      // optional uint64 length = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(Stream, _impl_.length_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(Stream, _impl_.length_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional .orc.proto.Stream.Kind kind = 1;
      {PROTOBUF_FIELD_OFFSET(Stream, _impl_.kind_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kEnum)},
      // optional uint32 column = 2;
      {PROTOBUF_FIELD_OFFSET(Stream, _impl_.column_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint64 length = 3;
      {PROTOBUF_FIELD_OFFSET(Stream, _impl_.length_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        {::_pbi::FieldAuxEnumData{}, ::orc::proto::Stream_Kind_internal_data_},
    }},
    {{
    }},
  };
}


inline constexpr Stream::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        kind_{static_cast< ::orc::proto::Stream_Kind >(0)},
        column_{0u},
        length_{::uint64_t{0u}} {}

template <typename>
constexpr Stream::Stream(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Stream::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Stream(arena);
}
constexpr auto Stream::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(Stream), alignof(Stream));
}
constexpr auto Stream::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Stream::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Stream>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Stream::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Stream>(), &Stream::ByteSizeLong,
              &Stream::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Stream, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[16],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StreamGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StreamGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Stream_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Stream::InternalGenerateClassData_(
            _default, &Stream_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Stream>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StreamGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Stream _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Stream>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StreamGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StreamGlobalsTypeInternal Stream_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Stream_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Stream_globals_.GetClassData();
#else
  return Stream_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class PostScript::_Internal {
 public:
  using HasBits = decltype(::std::declval<PostScript>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(PostScript, _impl_._has_bits_);
};

constexpr PostScript::ParseTableT_ PostScript::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(PostScript, _impl_._has_bits_),
      0, // no _extensions_
      8000, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967168,  // skipmap
      offsetof(ParseTableT_, field_entries),
      8,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::PostScript>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint64 footer_length = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(PostScript, _impl_.footer_length_), 2>(),
       {8, 2, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.footer_length_)}},
      // optional .orc.proto.CompressionKind compression = 2;
      {::_pbi::TcParser::FastEr0S1,
       {16, 4, 6,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.compression_)}},
      // optional uint64 compression_block_size = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(PostScript, _impl_.compression_block_size_), 3>(),
       {24, 3, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.compression_block_size_)}},
      // repeated uint32 version = 4 [packed = true];
      {::_pbi::TcParser::FastV32P1,
       {34, 0, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.version_)}},
      // optional uint64 metadata_length = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(PostScript, _impl_.metadata_length_), 6>(),
       {40, 6, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.metadata_length_)}},
      // optional uint32 writer_version = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(PostScript, _impl_.writer_version_), 5>(),
       {48, 5, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.writer_version_)}},
      // optional uint64 stripe_statistics_length = 7;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(PostScript, _impl_.stripe_statistics_length_), 7>(),
       {56, 7, 0,
        PROTOBUF_FIELD_OFFSET(PostScript, _impl_.stripe_statistics_length_)}},
    }}, {{
      8000, 0, 1,
      65534, 7,
      65535, 65535
    }}, {{
      // optional uint64 footer_length = 1;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.footer_length_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional .orc.proto.CompressionKind compression = 2;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.compression_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
      // optional uint64 compression_block_size = 3;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.compression_block_size_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // repeated uint32 version = 4 [packed = true];
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.version_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt32)},
      // optional uint64 metadata_length = 5;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.metadata_length_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint32 writer_version = 6;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.writer_version_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint64 stripe_statistics_length = 7;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.stripe_statistics_length_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional string magic = 8000;
      {PROTOBUF_FIELD_OFFSET(PostScript, _impl_.magic_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    {{
        {0, 6},
    }},
    {{
    }},
  };
}


inline constexpr PostScript::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        version_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::PostScript,
            PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.version_)>()
         }
        ,
        _version_cached_byte_size_{0},
        magic_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        footer_length_{::uint64_t{0u}},
        compression_block_size_{::uint64_t{0u}},
        compression_{static_cast< ::orc::proto::CompressionKind >(0)},
        writer_version_{0u},
        metadata_length_{::uint64_t{0u}},
        stripe_statistics_length_{::uint64_t{0u}} {}

template <typename>
constexpr PostScript::PostScript(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL PostScript::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) PostScript(arena);
}
constexpr auto PostScript::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(PostScript), alignof(PostScript));
}
constexpr auto PostScript::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &PostScript::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<PostScript>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &PostScript::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<PostScript>(), &PostScript::ByteSizeLong,
              &PostScript::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(PostScript, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[33],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct PostScriptGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr PostScriptGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 PostScript_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(PostScript::InternalGenerateClassData_(
            _default, &PostScript_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<PostScript>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~PostScriptGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) PostScript _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<PostScript>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(PostScriptGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST PostScriptGlobalsTypeInternal PostScript_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* PostScript_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return PostScript_globals_.GetClassData();
#else
  return PostScript_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class IntegerStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<IntegerStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_._has_bits_);
};

constexpr IntegerStatistics::ParseTableT_ IntegerStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::IntegerStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional sint64 minimum = 1;
      {::_pbi::TcParser::FastZ64S1,
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.minimum_)}},
      // optional sint64 maximum = 2;
      {::_pbi::TcParser::FastZ64S1,
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.maximum_)}},
      // optional sint64 sum = 3;
      {::_pbi::TcParser::FastZ64S1,
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.sum_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional sint64 minimum = 1;
      {PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional sint64 maximum = 2;
      {PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
      // optional sint64 sum = 3;
      {PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.sum_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr IntegerStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_{::int64_t{0}},
        maximum_{::int64_t{0}},
        sum_{::int64_t{0}} {}

template <typename>
constexpr IntegerStatistics::IntegerStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL IntegerStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) IntegerStatistics(arena);
}
constexpr auto IntegerStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(IntegerStatistics), alignof(IntegerStatistics));
}
constexpr auto IntegerStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &IntegerStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<IntegerStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &IntegerStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<IntegerStatistics>(), &IntegerStatistics::ByteSizeLong,
              &IntegerStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct IntegerStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr IntegerStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 IntegerStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(IntegerStatistics::InternalGenerateClassData_(
            _default, &IntegerStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<IntegerStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~IntegerStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) IntegerStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<IntegerStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(IntegerStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST IntegerStatisticsGlobalsTypeInternal IntegerStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* IntegerStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return IntegerStatistics_globals_.GetClassData();
#else
  return IntegerStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class EncryptionKey::_Internal {
 public:
  using HasBits = decltype(::std::declval<EncryptionKey>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_._has_bits_);
};

constexpr EncryptionKey::ParseTableT_ EncryptionKey::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::EncryptionKey>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string key_name = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.key_name_)}},
      // optional uint32 key_version = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(EncryptionKey, _impl_.key_version_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.key_version_)}},
      // optional .orc.proto.EncryptionAlgorithm algorithm = 3;
      {::_pbi::TcParser::FastEr0S1,
       {24, 2, 2,
        PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.algorithm_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string key_name = 1;
      {PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.key_name_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional uint32 key_version = 2;
      {PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.key_version_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional .orc.proto.EncryptionAlgorithm algorithm = 3;
      {PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.algorithm_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
    }},
    {{
        {0, 2},
    }},
    {{
    }},
  };
}


inline constexpr EncryptionKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        key_name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        key_version_{0u},
        algorithm_{static_cast< ::orc::proto::EncryptionAlgorithm >(0)} {}

template <typename>
constexpr EncryptionKey::EncryptionKey(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL EncryptionKey::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) EncryptionKey(arena);
}
constexpr auto EncryptionKey::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(EncryptionKey), alignof(EncryptionKey));
}
constexpr auto EncryptionKey::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &EncryptionKey::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<EncryptionKey>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &EncryptionKey::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<EncryptionKey>(), &EncryptionKey::ByteSizeLong,
              &EncryptionKey::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[29],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct EncryptionKeyGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr EncryptionKeyGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 EncryptionKey_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(EncryptionKey::InternalGenerateClassData_(
            _default, &EncryptionKey_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<EncryptionKey>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~EncryptionKeyGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) EncryptionKey _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<EncryptionKey>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(EncryptionKeyGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST EncryptionKeyGlobalsTypeInternal EncryptionKey_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* EncryptionKey_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return EncryptionKey_globals_.GetClassData();
#else
  return EncryptionKey_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DoubleStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<DoubleStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_._has_bits_);
};

constexpr DoubleStatistics::ParseTableT_ DoubleStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::DoubleStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional double minimum = 1;
      {::_pbi::TcParser::FastF64S1,
       {9, 0, 0,
        PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.minimum_)}},
      // optional double maximum = 2;
      {::_pbi::TcParser::FastF64S1,
       {17, 1, 0,
        PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.maximum_)}},
      // optional double sum = 3;
      {::_pbi::TcParser::FastF64S1,
       {25, 2, 0,
        PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.sum_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional double minimum = 1;
      {PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double maximum = 2;
      {PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double sum = 3;
      {PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.sum_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DoubleStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_{0},
        maximum_{0},
        sum_{0} {}

template <typename>
constexpr DoubleStatistics::DoubleStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DoubleStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DoubleStatistics(arena);
}
constexpr auto DoubleStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(DoubleStatistics), alignof(DoubleStatistics));
}
constexpr auto DoubleStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DoubleStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DoubleStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DoubleStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DoubleStatistics>(), &DoubleStatistics::ByteSizeLong,
              &DoubleStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DoubleStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DoubleStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DoubleStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DoubleStatistics::InternalGenerateClassData_(
            _default, &DoubleStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DoubleStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DoubleStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DoubleStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DoubleStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DoubleStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DoubleStatisticsGlobalsTypeInternal DoubleStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DoubleStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DoubleStatistics_globals_.GetClassData();
#else
  return DoubleStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DecimalStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<DecimalStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_._has_bits_);
};

constexpr DecimalStatistics::ParseTableT_ DecimalStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::DecimalStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string minimum = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.minimum_)}},
      // optional string maximum = 2;
      {::_pbi::TcParser::FastBS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.maximum_)}},
      // optional string sum = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.sum_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string minimum = 1;
      {PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional string maximum = 2;
      {PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional string sum = 3;
      {PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_.sum_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DecimalStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        maximum_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        sum_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr DecimalStatistics::DecimalStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DecimalStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DecimalStatistics(arena);
}
constexpr auto DecimalStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(DecimalStatistics), alignof(DecimalStatistics));
}
constexpr auto DecimalStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DecimalStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DecimalStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DecimalStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DecimalStatistics>(), &DecimalStatistics::ByteSizeLong,
              &DecimalStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DecimalStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[4],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DecimalStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DecimalStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DecimalStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DecimalStatistics::InternalGenerateClassData_(
            _default, &DecimalStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DecimalStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DecimalStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DecimalStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DecimalStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DecimalStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DecimalStatisticsGlobalsTypeInternal DecimalStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DecimalStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DecimalStatistics_globals_.GetClassData();
#else
  return DecimalStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DateStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<DateStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_._has_bits_);
};

constexpr DateStatistics::ParseTableT_ DateStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::DateStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional sint32 maximum = 2;
      {::_pbi::TcParser::FastZ32S1,
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.maximum_)}},
      // optional sint32 minimum = 1;
      {::_pbi::TcParser::FastZ32S1,
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.minimum_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional sint32 minimum = 1;
      {PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.minimum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt32)},
      // optional sint32 maximum = 2;
      {PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.maximum_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt32)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DateStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        minimum_{0},
        maximum_{0} {}

template <typename>
constexpr DateStatistics::DateStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DateStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DateStatistics(arena);
}
constexpr auto DateStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(DateStatistics), alignof(DateStatistics));
}
constexpr auto DateStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DateStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DateStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DateStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DateStatistics>(), &DateStatistics::ByteSizeLong,
              &DateStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[5],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DateStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DateStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DateStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DateStatistics::InternalGenerateClassData_(
            _default, &DateStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DateStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DateStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DateStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DateStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DateStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DateStatisticsGlobalsTypeInternal DateStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DateStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DateStatistics_globals_.GetClassData();
#else
  return DateStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class DataMask::_Internal {
 public:
  using HasBits = decltype(::std::declval<DataMask>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DataMask, _impl_._has_bits_);
};

constexpr DataMask::ParseTableT_ DataMask::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DataMask, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::DataMask>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional string name = 1;
      {::_pbi::TcParser::FastBS1,
       {10, 2, 0,
        PROTOBUF_FIELD_OFFSET(DataMask, _impl_.name_)}},
      // repeated string mask_parameters = 2;
      {::_pbi::TcParser::FastBR1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(DataMask, _impl_.mask_parameters_)}},
      // repeated uint32 columns = 3 [packed = true];
      {::_pbi::TcParser::FastV32P1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(DataMask, _impl_.columns_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional string name = 1;
      {PROTOBUF_FIELD_OFFSET(DataMask, _impl_.name_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // repeated string mask_parameters = 2;
      {PROTOBUF_FIELD_OFFSET(DataMask, _impl_.mask_parameters_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kBytes | ::_fl::kRepSString)},
      // repeated uint32 columns = 3 [packed = true];
      {PROTOBUF_FIELD_OFFSET(DataMask, _impl_.columns_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt32)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr DataMask::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mask_parameters_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::DataMask,
            PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.mask_parameters_)>()
         }
        ,
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::DataMask,
            PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.columns_)>()
         }
        ,
        _columns_cached_byte_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr DataMask::DataMask(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DataMask::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DataMask(arena);
}
constexpr auto DataMask::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(DataMask), alignof(DataMask));
}
constexpr auto DataMask::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DataMask::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DataMask>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DataMask::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DataMask>(), &DataMask::ByteSizeLong,
              &DataMask::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DataMask, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[28],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DataMaskGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DataMaskGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DataMask_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DataMask::InternalGenerateClassData_(
            _default, &DataMask_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DataMask>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DataMaskGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DataMask _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DataMask>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DataMaskGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DataMaskGlobalsTypeInternal DataMask_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DataMask_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DataMask_globals_.GetClassData();
#else
  return DataMask_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ColumnEncoding::_Internal {
 public:
  using HasBits = decltype(::std::declval<ColumnEncoding>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_._has_bits_);
};

constexpr ColumnEncoding::ParseTableT_ ColumnEncoding::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::ColumnEncoding>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional .orc.proto.ColumnEncoding.Kind kind = 1;
      {::_pbi::TcParser::FastEr0S1,
       {8, 0, 3,
        PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.kind_)}},
      // optional uint32 dictionary_size = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(ColumnEncoding, _impl_.dictionary_size_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.dictionary_size_)}},
      // optional uint32 bloom_encoding = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(ColumnEncoding, _impl_.bloom_encoding_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.bloom_encoding_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional .orc.proto.ColumnEncoding.Kind kind = 1;
      {PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.kind_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
      // optional uint32 dictionary_size = 2;
      {PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.dictionary_size_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint32 bloom_encoding = 3;
      {PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.bloom_encoding_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    }},
    {{
        {0, 3},
    }},
    {{
    }},
  };
}


inline constexpr ColumnEncoding::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        kind_{static_cast< ::orc::proto::ColumnEncoding_Kind >(0)},
        dictionary_size_{0u},
        bloom_encoding_{0u} {}

template <typename>
constexpr ColumnEncoding::ColumnEncoding(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ColumnEncoding::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ColumnEncoding(arena);
}
constexpr auto ColumnEncoding::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ColumnEncoding), alignof(ColumnEncoding));
}
constexpr auto ColumnEncoding::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ColumnEncoding::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ColumnEncoding>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ColumnEncoding::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ColumnEncoding>(), &ColumnEncoding::ByteSizeLong,
              &ColumnEncoding::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[17],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ColumnEncodingGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ColumnEncodingGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ColumnEncoding_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ColumnEncoding::InternalGenerateClassData_(
            _default, &ColumnEncoding_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ColumnEncoding>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ColumnEncodingGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ColumnEncoding _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ColumnEncoding>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ColumnEncodingGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ColumnEncodingGlobalsTypeInternal ColumnEncoding_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ColumnEncoding_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ColumnEncoding_globals_.GetClassData();
#else
  return ColumnEncoding_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CollectionStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<CollectionStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_._has_bits_);
};

constexpr CollectionStatistics::ParseTableT_ CollectionStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::CollectionStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint64 min_children = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(CollectionStatistics, _impl_.min_children_), 0>(),
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.min_children_)}},
      // optional uint64 max_children = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(CollectionStatistics, _impl_.max_children_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.max_children_)}},
      // optional uint64 total_children = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(CollectionStatistics, _impl_.total_children_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.total_children_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint64 min_children = 1;
      {PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.min_children_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 max_children = 2;
      {PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.max_children_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 total_children = 3;
      {PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.total_children_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr CollectionStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        min_children_{::uint64_t{0u}},
        max_children_{::uint64_t{0u}},
        total_children_{::uint64_t{0u}} {}

template <typename>
constexpr CollectionStatistics::CollectionStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CollectionStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CollectionStatistics(arena);
}
constexpr auto CollectionStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(CollectionStatistics), alignof(CollectionStatistics));
}
constexpr auto CollectionStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CollectionStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CollectionStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CollectionStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CollectionStatistics>(), &CollectionStatistics::ByteSizeLong,
              &CollectionStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[8],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CollectionStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CollectionStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CollectionStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CollectionStatistics::InternalGenerateClassData_(
            _default, &CollectionStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CollectionStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CollectionStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CollectionStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CollectionStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CollectionStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CollectionStatisticsGlobalsTypeInternal CollectionStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CollectionStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CollectionStatistics_globals_.GetClassData();
#else
  return CollectionStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class BucketStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<BucketStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BucketStatistics, _impl_._has_bits_);
};

constexpr BucketStatistics::ParseTableT_ BucketStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(BucketStatistics, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::BucketStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated uint64 count = 1 [packed = true];
      {::_pbi::TcParser::FastV64P1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(BucketStatistics, _impl_.count_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated uint64 count = 1 [packed = true];
      {PROTOBUF_FIELD_OFFSET(BucketStatistics, _impl_.count_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt64)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr BucketStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        count_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BucketStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BucketStatistics, _impl_.count_)>()
         }
        ,
        _count_cached_byte_size_{0} {}

template <typename>
constexpr BucketStatistics::BucketStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL BucketStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BucketStatistics(arena);
}
constexpr auto BucketStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(BucketStatistics), alignof(BucketStatistics));
}
constexpr auto BucketStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &BucketStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BucketStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BucketStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BucketStatistics>(), &BucketStatistics::ByteSizeLong,
              &BucketStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BucketStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[3],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct BucketStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr BucketStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 BucketStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(BucketStatistics::InternalGenerateClassData_(
            _default, &BucketStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<BucketStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~BucketStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) BucketStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<BucketStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(BucketStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST BucketStatisticsGlobalsTypeInternal BucketStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* BucketStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return BucketStatistics_globals_.GetClassData();
#else
  return BucketStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class BoundingBox::_Internal {
 public:
  using HasBits = decltype(::std::declval<BoundingBox>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_._has_bits_);
};

constexpr BoundingBox::ParseTableT_ BoundingBox::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_._has_bits_),
      0, // no _extensions_
      8, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967040,  // skipmap
      offsetof(ParseTableT_, field_entries),
      8,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::BoundingBox>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional double mmax = 8;
      {::_pbi::TcParser::FastF64S1,
       {65, 7, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.mmax_)}},
      // optional double xmin = 1;
      {::_pbi::TcParser::FastF64S1,
       {9, 0, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.xmin_)}},
      // optional double xmax = 2;
      {::_pbi::TcParser::FastF64S1,
       {17, 1, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.xmax_)}},
      // optional double ymin = 3;
      {::_pbi::TcParser::FastF64S1,
       {25, 2, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.ymin_)}},
      // optional double ymax = 4;
      {::_pbi::TcParser::FastF64S1,
       {33, 3, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.ymax_)}},
      // optional double zmin = 5;
      {::_pbi::TcParser::FastF64S1,
       {41, 4, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.zmin_)}},
      // optional double zmax = 6;
      {::_pbi::TcParser::FastF64S1,
       {49, 5, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.zmax_)}},
      // optional double mmin = 7;
      {::_pbi::TcParser::FastF64S1,
       {57, 6, 0,
        PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.mmin_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional double xmin = 1;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.xmin_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double xmax = 2;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.xmax_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double ymin = 3;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.ymin_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double ymax = 4;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.ymax_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double zmin = 5;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.zmin_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double zmax = 6;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.zmax_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double mmin = 7;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.mmin_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
      // optional double mmax = 8;
      {PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.mmax_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kDouble)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr BoundingBox::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        xmin_{0},
        xmax_{0},
        ymin_{0},
        ymax_{0},
        zmin_{0},
        zmax_{0},
        mmin_{0},
        mmax_{0} {}

template <typename>
constexpr BoundingBox::BoundingBox(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL BoundingBox::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BoundingBox(arena);
}
constexpr auto BoundingBox::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(BoundingBox), alignof(BoundingBox));
}
constexpr auto BoundingBox::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &BoundingBox::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BoundingBox>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BoundingBox::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BoundingBox>(), &BoundingBox::ByteSizeLong,
              &BoundingBox::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[9],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct BoundingBoxGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr BoundingBoxGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 BoundingBox_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(BoundingBox::InternalGenerateClassData_(
            _default, &BoundingBox_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<BoundingBox>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~BoundingBoxGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) BoundingBox _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<BoundingBox>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(BoundingBoxGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST BoundingBoxGlobalsTypeInternal BoundingBox_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* BoundingBox_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return BoundingBox_globals_.GetClassData();
#else
  return BoundingBox_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class BloomFilter::_Internal {
 public:
  using HasBits = decltype(::std::declval<BloomFilter>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_._has_bits_);
};

constexpr BloomFilter::ParseTableT_ BloomFilter::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::BloomFilter>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint32 num_hash_functions = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(BloomFilter, _impl_.num_hash_functions_), 2>(),
       {8, 2, 0,
        PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.num_hash_functions_)}},
      // repeated fixed64 bitset = 2;
      {::_pbi::TcParser::FastF64R1,
       {17, 0, 0,
        PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.bitset_)}},
      // optional bytes utf8bitset = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.utf8bitset_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint32 num_hash_functions = 1;
      {PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.num_hash_functions_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // repeated fixed64 bitset = 2;
      {PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.bitset_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kFixed64)},
      // optional bytes utf8bitset = 3;
      {PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_.utf8bitset_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr BloomFilter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        bitset_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BloomFilter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.bitset_)>()
         }
        ,
        utf8bitset_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        num_hash_functions_{0u} {}

template <typename>
constexpr BloomFilter::BloomFilter(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL BloomFilter::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BloomFilter(arena);
}
constexpr auto BloomFilter::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(BloomFilter), alignof(BloomFilter));
}
constexpr auto BloomFilter::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &BloomFilter::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BloomFilter>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BloomFilter::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BloomFilter>(), &BloomFilter::ByteSizeLong,
              &BloomFilter::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BloomFilter, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[14],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct BloomFilterGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr BloomFilterGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 BloomFilter_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(BloomFilter::InternalGenerateClassData_(
            _default, &BloomFilter_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<BloomFilter>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~BloomFilterGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) BloomFilter _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<BloomFilter>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(BloomFilterGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST BloomFilterGlobalsTypeInternal BloomFilter_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* BloomFilter_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return BloomFilter_globals_.GetClassData();
#else
  return BloomFilter_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class BinaryStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<BinaryStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BinaryStatistics, _impl_._has_bits_);
};

constexpr BinaryStatistics::ParseTableT_ BinaryStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(BinaryStatistics, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::BinaryStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional sint64 sum = 1;
      {::_pbi::TcParser::FastZ64S1,
       {8, 0, 0,
        PROTOBUF_FIELD_OFFSET(BinaryStatistics, _impl_.sum_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional sint64 sum = 1;
      {PROTOBUF_FIELD_OFFSET(BinaryStatistics, _impl_.sum_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kSInt64)},
    }},
    // no aux_entries
    {{
    }},
  };
}


inline constexpr BinaryStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        sum_{::int64_t{0}} {}

template <typename>
constexpr BinaryStatistics::BinaryStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL BinaryStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BinaryStatistics(arena);
}
constexpr auto BinaryStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(BinaryStatistics), alignof(BinaryStatistics));
}
constexpr auto BinaryStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &BinaryStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BinaryStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BinaryStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BinaryStatistics>(), &BinaryStatistics::ByteSizeLong,
              &BinaryStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BinaryStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[7],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct BinaryStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr BinaryStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 BinaryStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(BinaryStatistics::InternalGenerateClassData_(
            _default, &BinaryStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<BinaryStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~BinaryStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) BinaryStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<BinaryStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(BinaryStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST BinaryStatisticsGlobalsTypeInternal BinaryStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* BinaryStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return BinaryStatistics_globals_.GetClassData();
#else
  return BinaryStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Type::_Internal {
 public:
  using HasBits = decltype(::std::declval<Type>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Type, _impl_._has_bits_);
};

constexpr Type::ParseTableT_ Type::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Type, _impl_._has_bits_),
      0, // no _extensions_
      9, 120,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294966784,  // skipmap
      offsetof(ParseTableT_, field_entries),
      9,  // num_field_entries
      3,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::Type>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional .orc.proto.Type.Kind kind = 1;
      {::_pbi::TcParser::FastEr0S1,
       {8, 4, 20,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.kind_)}},
      // repeated uint32 subtypes = 2 [packed = true];
      {::_pbi::TcParser::FastV32P1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.subtypes_)}},
      // repeated string field_names = 3;
      {::_pbi::TcParser::FastBR1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.field_names_)}},
      // optional uint32 maximum_length = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Type, _impl_.maximum_length_), 5>(),
       {32, 5, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.maximum_length_)}},
      // optional uint32 precision = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Type, _impl_.precision_), 6>(),
       {40, 6, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.precision_)}},
      // optional uint32 scale = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Type, _impl_.scale_), 7>(),
       {48, 7, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.scale_)}},
      // repeated .orc.proto.StringPair attributes = 7;
      {::_pbi::TcParser::FastMtR1,
       {58, 2, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.attributes_)}},
      // optional string crs = 8;
      {::_pbi::TcParser::FastBS1,
       {66, 3, 0,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.crs_)}},
      // optional .orc.proto.Type.EdgeInterpolationAlgorithm algorithm = 9;
      {::_pbi::TcParser::FastEr0S1,
       {72, 8, 4,
        PROTOBUF_FIELD_OFFSET(Type, _impl_.algorithm_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional .orc.proto.Type.Kind kind = 1;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.kind_), _Internal::kHasBitsOffset + 4, 1, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
      // repeated uint32 subtypes = 2 [packed = true];
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.subtypes_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt32)},
      // repeated string field_names = 3;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.field_names_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcRepeated | ::_fl::kBytes | ::_fl::kRepSString)},
      // optional uint32 maximum_length = 4;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.maximum_length_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint32 precision = 5;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.precision_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint32 scale = 6;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.scale_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // repeated .orc.proto.StringPair attributes = 7;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.attributes_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional string crs = 8;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.crs_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // optional .orc.proto.Type.EdgeInterpolationAlgorithm algorithm = 9;
      {PROTOBUF_FIELD_OFFSET(Type, _impl_.algorithm_), _Internal::kHasBitsOffset + 8, 2, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::StringPair>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::StringPair_globals_},
        #endif
        {0, 20},
        {0, 4},
    }},
    {{
    }},
  };
}


inline constexpr Type::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        subtypes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.subtypes_)>()
         }
        ,
        _subtypes_cached_byte_size_{0},
        field_names_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.field_names_)>()
         }
        ,
        attributes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.attributes_)>()
         }
        ,
        crs_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        kind_{static_cast< ::orc::proto::Type_Kind >(0)},
        maximum_length_{0u},
        precision_{0u},
        scale_{0u},
        algorithm_{static_cast< ::orc::proto::Type_EdgeInterpolationAlgorithm >(0)} {}

template <typename>
constexpr Type::Type(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Type::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Type(arena);
}
constexpr auto Type::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Type), alignof(Type));
}
constexpr auto Type::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Type::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Type>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Type::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Type>(), &Type::ByteSizeLong,
              &Type::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Type, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[21],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct TypeGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr TypeGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Type_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Type::InternalGenerateClassData_(
            _default, &Type_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Type>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~TypeGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Type _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Type>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(TypeGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST TypeGlobalsTypeInternal Type_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Type_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Type_globals_.GetClassData();
#else
  return Type_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StripeEncryptionVariant::_Internal {
 public:
  using HasBits = decltype(::std::declval<StripeEncryptionVariant>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_._has_bits_);
};

constexpr StripeEncryptionVariant::ParseTableT_ StripeEncryptionVariant::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StripeEncryptionVariant>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.ColumnEncoding encoding = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 1, 1,
        PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_.encoding_)}},
      // repeated .orc.proto.Stream streams = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_.streams_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.Stream streams = 1;
      {PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_.streams_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.ColumnEncoding encoding = 2;
      {PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_.encoding_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Stream>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Stream_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnEncoding>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnEncoding_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr StripeEncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        streams_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeEncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.streams_)>()
         }
        ,
        encoding_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeEncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.encoding_)>()
         }
     {}

template <typename>
constexpr StripeEncryptionVariant::StripeEncryptionVariant(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StripeEncryptionVariant::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StripeEncryptionVariant(arena);
}
constexpr auto StripeEncryptionVariant::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StripeEncryptionVariant), alignof(StripeEncryptionVariant));
}
constexpr auto StripeEncryptionVariant::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StripeEncryptionVariant::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StripeEncryptionVariant>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StripeEncryptionVariant::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StripeEncryptionVariant>(), &StripeEncryptionVariant::ByteSizeLong,
              &StripeEncryptionVariant::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StripeEncryptionVariant, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[18],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StripeEncryptionVariantGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StripeEncryptionVariantGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StripeEncryptionVariant_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StripeEncryptionVariant::InternalGenerateClassData_(
            _default, &StripeEncryptionVariant_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StripeEncryptionVariant>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StripeEncryptionVariantGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StripeEncryptionVariant _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StripeEncryptionVariant>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StripeEncryptionVariantGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StripeEncryptionVariantGlobalsTypeInternal StripeEncryptionVariant_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StripeEncryptionVariant_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StripeEncryptionVariant_globals_.GetClassData();
#else
  return StripeEncryptionVariant_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class GeospatialStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<GeospatialStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_._has_bits_);
};

constexpr GeospatialStatistics::ParseTableT_ GeospatialStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::GeospatialStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated int32 geospatial_types = 2;
      {::_pbi::TcParser::FastV32R1,
       {16, 0, 0,
        PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_.geospatial_types_)}},
      // optional .orc.proto.BoundingBox bbox = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_.bbox_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional .orc.proto.BoundingBox bbox = 1;
      {PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_.bbox_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated int32 geospatial_types = 2;
      {PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_.geospatial_types_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kInt32)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::BoundingBox>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::BoundingBox_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr GeospatialStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        geospatial_types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::GeospatialStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_.geospatial_types_)>()
         }
        ,
        bbox_{nullptr} {}

template <typename>
constexpr GeospatialStatistics::GeospatialStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL GeospatialStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) GeospatialStatistics(arena);
}
constexpr auto GeospatialStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(GeospatialStatistics), alignof(GeospatialStatistics));
}
constexpr auto GeospatialStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &GeospatialStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<GeospatialStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &GeospatialStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<GeospatialStatistics>(), &GeospatialStatistics::ByteSizeLong,
              &GeospatialStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(GeospatialStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[10],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct GeospatialStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr GeospatialStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 GeospatialStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(GeospatialStatistics::InternalGenerateClassData_(
            _default, &GeospatialStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<GeospatialStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~GeospatialStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) GeospatialStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<GeospatialStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(GeospatialStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST GeospatialStatisticsGlobalsTypeInternal GeospatialStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* GeospatialStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return GeospatialStatistics_globals_.GetClassData();
#else
  return GeospatialStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class EncryptionVariant::_Internal {
 public:
  using HasBits = decltype(::std::declval<EncryptionVariant>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_._has_bits_);
};

constexpr EncryptionVariant::ParseTableT_ EncryptionVariant::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::EncryptionVariant>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint32 root = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(EncryptionVariant, _impl_.root_), 3>(),
       {8, 3, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.root_)}},
      // optional uint32 key = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(EncryptionVariant, _impl_.key_), 4>(),
       {16, 4, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.key_)}},
      // optional bytes encrypted_key = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.encrypted_key_)}},
      // repeated .orc.proto.Stream stripe_statistics = 4;
      {::_pbi::TcParser::FastMtR1,
       {34, 0, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.stripe_statistics_)}},
      // optional bytes file_statistics = 5;
      {::_pbi::TcParser::FastBS1,
       {42, 2, 0,
        PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.file_statistics_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint32 root = 1;
      {PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.root_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint32 key = 2;
      {PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.key_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional bytes encrypted_key = 3;
      {PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.encrypted_key_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // repeated .orc.proto.Stream stripe_statistics = 4;
      {PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.stripe_statistics_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional bytes file_statistics = 5;
      {PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.file_statistics_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Stream>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Stream_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr EncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        stripe_statistics_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::EncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.stripe_statistics_)>()
         }
        ,
        encrypted_key_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        file_statistics_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        root_{0u},
        key_{0u} {}

template <typename>
constexpr EncryptionVariant::EncryptionVariant(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL EncryptionVariant::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) EncryptionVariant(arena);
}
constexpr auto EncryptionVariant::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(EncryptionVariant), alignof(EncryptionVariant));
}
constexpr auto EncryptionVariant::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &EncryptionVariant::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<EncryptionVariant>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &EncryptionVariant::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<EncryptionVariant>(), &EncryptionVariant::ByteSizeLong,
              &EncryptionVariant::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[30],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct EncryptionVariantGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr EncryptionVariantGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 EncryptionVariant_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(EncryptionVariant::InternalGenerateClassData_(
            _default, &EncryptionVariant_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<EncryptionVariant>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~EncryptionVariantGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) EncryptionVariant _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<EncryptionVariant>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(EncryptionVariantGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST EncryptionVariantGlobalsTypeInternal EncryptionVariant_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* EncryptionVariant_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return EncryptionVariant_globals_.GetClassData();
#else
  return EncryptionVariant_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class BloomFilterIndex::_Internal {
 public:
  using HasBits = decltype(::std::declval<BloomFilterIndex>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(BloomFilterIndex, _impl_._has_bits_);
};

constexpr BloomFilterIndex::ParseTableT_ BloomFilterIndex::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(BloomFilterIndex, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::BloomFilterIndex>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.BloomFilter bloom_filter = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(BloomFilterIndex, _impl_.bloom_filter_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.BloomFilter bloom_filter = 1;
      {PROTOBUF_FIELD_OFFSET(BloomFilterIndex, _impl_.bloom_filter_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::BloomFilter>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::BloomFilter_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr BloomFilterIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        bloom_filter_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BloomFilterIndex,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilterIndex, _impl_.bloom_filter_)>()
         }
     {}

template <typename>
constexpr BloomFilterIndex::BloomFilterIndex(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL BloomFilterIndex::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) BloomFilterIndex(arena);
}
constexpr auto BloomFilterIndex::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(BloomFilterIndex), alignof(BloomFilterIndex));
}
constexpr auto BloomFilterIndex::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &BloomFilterIndex::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<BloomFilterIndex>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &BloomFilterIndex::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<BloomFilterIndex>(), &BloomFilterIndex::ByteSizeLong,
              &BloomFilterIndex::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(BloomFilterIndex, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[15],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct BloomFilterIndexGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr BloomFilterIndexGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 BloomFilterIndex_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(BloomFilterIndex::InternalGenerateClassData_(
            _default, &BloomFilterIndex_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<BloomFilterIndex>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~BloomFilterIndexGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) BloomFilterIndex _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<BloomFilterIndex>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(BloomFilterIndexGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST BloomFilterIndexGlobalsTypeInternal BloomFilterIndex_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* BloomFilterIndex_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return BloomFilterIndex_globals_.GetClassData();
#else
  return BloomFilterIndex_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StripeFooter::_Internal {
 public:
  using HasBits = decltype(::std::declval<StripeFooter>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_._has_bits_);
};

constexpr StripeFooter::ParseTableT_ StripeFooter::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      3,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StripeFooter>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.StripeEncryptionVariant encryption = 4;
      {::_pbi::TcParser::FastMtR1,
       {34, 2, 2,
        PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.encryption_)}},
      // repeated .orc.proto.Stream streams = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.streams_)}},
      // repeated .orc.proto.ColumnEncoding columns = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 1, 1,
        PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.columns_)}},
      // optional string writer_timezone = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 3, 0,
        PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.writer_timezone_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.Stream streams = 1;
      {PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.streams_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.ColumnEncoding columns = 2;
      {PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.columns_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional string writer_timezone = 3;
      {PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.writer_timezone_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // repeated .orc.proto.StripeEncryptionVariant encryption = 4;
      {PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_.encryption_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Stream>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Stream_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnEncoding>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnEncoding_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::StripeEncryptionVariant>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::StripeEncryptionVariant_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr StripeFooter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        streams_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.streams_)>()
         }
        ,
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.columns_)>()
         }
        ,
        encryption_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.encryption_)>()
         }
        ,
        writer_timezone_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr StripeFooter::StripeFooter(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StripeFooter::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StripeFooter(arena);
}
constexpr auto StripeFooter::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StripeFooter), alignof(StripeFooter));
}
constexpr auto StripeFooter::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StripeFooter::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StripeFooter>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StripeFooter::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StripeFooter>(), &StripeFooter::ByteSizeLong,
              &StripeFooter::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StripeFooter, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[19],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StripeFooterGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StripeFooterGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StripeFooter_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StripeFooter::InternalGenerateClassData_(
            _default, &StripeFooter_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StripeFooter>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StripeFooterGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StripeFooter _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StripeFooter>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StripeFooterGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StripeFooterGlobalsTypeInternal StripeFooter_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StripeFooter_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StripeFooter_globals_.GetClassData();
#else
  return StripeFooter_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Encryption::_Internal {
 public:
  using HasBits = decltype(::std::declval<Encryption>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Encryption, _impl_._has_bits_);
};

constexpr Encryption::ParseTableT_ Encryption::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Encryption, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      4,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::Encryption>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional .orc.proto.KeyProviderKind key_provider = 4;
      {::_pbi::TcParser::FastEr0S1,
       {32, 3, 4,
        PROTOBUF_FIELD_OFFSET(Encryption, _impl_.key_provider_)}},
      // repeated .orc.proto.DataMask mask = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Encryption, _impl_.mask_)}},
      // repeated .orc.proto.EncryptionKey key = 2;
      {::_pbi::TcParser::FastMtR1,
       {18, 1, 1,
        PROTOBUF_FIELD_OFFSET(Encryption, _impl_.key_)}},
      // repeated .orc.proto.EncryptionVariant variants = 3;
      {::_pbi::TcParser::FastMtR1,
       {26, 2, 2,
        PROTOBUF_FIELD_OFFSET(Encryption, _impl_.variants_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.DataMask mask = 1;
      {PROTOBUF_FIELD_OFFSET(Encryption, _impl_.mask_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.EncryptionKey key = 2;
      {PROTOBUF_FIELD_OFFSET(Encryption, _impl_.key_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.EncryptionVariant variants = 3;
      {PROTOBUF_FIELD_OFFSET(Encryption, _impl_.variants_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.KeyProviderKind key_provider = 4;
      {PROTOBUF_FIELD_OFFSET(Encryption, _impl_.key_provider_), _Internal::kHasBitsOffset + 3, 3, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::DataMask>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::DataMask_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::EncryptionKey>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::EncryptionKey_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::EncryptionVariant>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::EncryptionVariant_globals_},
        #endif
        {0, 4},
    }},
    {{
    }},
  };
}


inline constexpr Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        mask_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.mask_)>()
         }
        ,
        key_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.key_)>()
         }
        ,
        variants_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.variants_)>()
         }
        ,
        key_provider_{static_cast< ::orc::proto::KeyProviderKind >(0)} {}

template <typename>
constexpr Encryption::Encryption(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Encryption::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Encryption(arena);
}
constexpr auto Encryption::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Encryption), alignof(Encryption));
}
constexpr auto Encryption::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Encryption::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Encryption>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Encryption::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Encryption>(), &Encryption::ByteSizeLong,
              &Encryption::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Encryption, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[31],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct EncryptionGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr EncryptionGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Encryption_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Encryption::InternalGenerateClassData_(
            _default, &Encryption_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Encryption>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~EncryptionGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Encryption _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Encryption>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(EncryptionGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST EncryptionGlobalsTypeInternal Encryption_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Encryption_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Encryption_globals_.GetClassData();
#else
  return Encryption_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ColumnStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<ColumnStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_._has_bits_);
};

constexpr ColumnStatistics::ParseTableT_ ColumnStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_._has_bits_),
      0, // no _extensions_
      13, 120,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294959104,  // skipmap
      offsetof(ParseTableT_, field_entries),
      13,  // num_field_entries
      10,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint64 number_of_values = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ColumnStatistics, _impl_.number_of_values_), 10>(),
       {8, 10, 0,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.number_of_values_)}},
      // optional .orc.proto.IntegerStatistics int_statistics = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.int_statistics_)}},
      // optional .orc.proto.DoubleStatistics double_statistics = 3;
      {::_pbi::TcParser::FastMtS1,
       {26, 1, 1,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.double_statistics_)}},
      // optional .orc.proto.StringStatistics string_statistics = 4;
      {::_pbi::TcParser::FastMtS1,
       {34, 2, 2,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.string_statistics_)}},
      // optional .orc.proto.BucketStatistics bucket_statistics = 5;
      {::_pbi::TcParser::FastMtS1,
       {42, 3, 3,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.bucket_statistics_)}},
      // optional .orc.proto.DecimalStatistics decimal_statistics = 6;
      {::_pbi::TcParser::FastMtS1,
       {50, 4, 4,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.decimal_statistics_)}},
      // optional .orc.proto.DateStatistics date_statistics = 7;
      {::_pbi::TcParser::FastMtS1,
       {58, 5, 5,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.date_statistics_)}},
      // optional .orc.proto.BinaryStatistics binary_statistics = 8;
      {::_pbi::TcParser::FastMtS1,
       {66, 6, 6,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.binary_statistics_)}},
      // optional .orc.proto.TimestampStatistics timestamp_statistics = 9;
      {::_pbi::TcParser::FastMtS1,
       {74, 7, 7,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.timestamp_statistics_)}},
      // optional bool has_null = 10;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(ColumnStatistics, _impl_.has_null_), 12>(),
       {80, 12, 0,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.has_null_)}},
      // optional uint64 bytes_on_disk = 11;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ColumnStatistics, _impl_.bytes_on_disk_), 11>(),
       {88, 11, 0,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.bytes_on_disk_)}},
      // optional .orc.proto.CollectionStatistics collection_statistics = 12;
      {::_pbi::TcParser::FastMtS1,
       {98, 8, 8,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.collection_statistics_)}},
      // optional .orc.proto.GeospatialStatistics geospatial_statistics = 13;
      {::_pbi::TcParser::FastMtS1,
       {106, 9, 9,
        PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.geospatial_statistics_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint64 number_of_values = 1;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.number_of_values_), _Internal::kHasBitsOffset + 10, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional .orc.proto.IntegerStatistics int_statistics = 2;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.int_statistics_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.DoubleStatistics double_statistics = 3;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.double_statistics_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.StringStatistics string_statistics = 4;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.string_statistics_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.BucketStatistics bucket_statistics = 5;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.bucket_statistics_), _Internal::kHasBitsOffset + 3, 3, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.DecimalStatistics decimal_statistics = 6;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.decimal_statistics_), _Internal::kHasBitsOffset + 4, 4, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.DateStatistics date_statistics = 7;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.date_statistics_), _Internal::kHasBitsOffset + 5, 5, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.BinaryStatistics binary_statistics = 8;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.binary_statistics_), _Internal::kHasBitsOffset + 6, 6, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.TimestampStatistics timestamp_statistics = 9;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.timestamp_statistics_), _Internal::kHasBitsOffset + 7, 7, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional bool has_null = 10;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.has_null_), _Internal::kHasBitsOffset + 12, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // optional uint64 bytes_on_disk = 11;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.bytes_on_disk_), _Internal::kHasBitsOffset + 11, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional .orc.proto.CollectionStatistics collection_statistics = 12;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.collection_statistics_), _Internal::kHasBitsOffset + 8, 8, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.GeospatialStatistics geospatial_statistics = 13;
      {PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.geospatial_statistics_), _Internal::kHasBitsOffset + 9, 9, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::IntegerStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::IntegerStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::DoubleStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::DoubleStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::StringStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::StringStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::BucketStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::BucketStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::DecimalStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::DecimalStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::DateStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::DateStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::BinaryStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::BinaryStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::TimestampStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::TimestampStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::CollectionStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::CollectionStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::GeospatialStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::GeospatialStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr ColumnStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        int_statistics_{nullptr},
        double_statistics_{nullptr},
        string_statistics_{nullptr},
        bucket_statistics_{nullptr},
        decimal_statistics_{nullptr},
        date_statistics_{nullptr},
        binary_statistics_{nullptr},
        timestamp_statistics_{nullptr},
        collection_statistics_{nullptr},
        geospatial_statistics_{nullptr},
        number_of_values_{::uint64_t{0u}},
        bytes_on_disk_{::uint64_t{0u}},
        has_null_{false} {}

template <typename>
constexpr ColumnStatistics::ColumnStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ColumnStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ColumnStatistics(arena);
}
constexpr auto ColumnStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ColumnStatistics), alignof(ColumnStatistics));
}
constexpr auto ColumnStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ColumnStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ColumnStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ColumnStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ColumnStatistics>(), &ColumnStatistics::ByteSizeLong,
              &ColumnStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[11],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ColumnStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ColumnStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ColumnStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ColumnStatistics::InternalGenerateClassData_(
            _default, &ColumnStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ColumnStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ColumnStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ColumnStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ColumnStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ColumnStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ColumnStatisticsGlobalsTypeInternal ColumnStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ColumnStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ColumnStatistics_globals_.GetClassData();
#else
  return ColumnStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class StripeStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<StripeStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StripeStatistics, _impl_._has_bits_);
};

constexpr StripeStatistics::ParseTableT_ StripeStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StripeStatistics, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::StripeStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.ColumnStatistics col_stats = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StripeStatistics, _impl_.col_stats_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.ColumnStatistics col_stats = 1;
      {PROTOBUF_FIELD_OFFSET(StripeStatistics, _impl_.col_stats_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr StripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        col_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeStatistics, _impl_.col_stats_)>()
         }
     {}

template <typename>
constexpr StripeStatistics::StripeStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StripeStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StripeStatistics(arena);
}
constexpr auto StripeStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StripeStatistics), alignof(StripeStatistics));
}
constexpr auto StripeStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StripeStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StripeStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StripeStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StripeStatistics>(), &StripeStatistics::ByteSizeLong,
              &StripeStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StripeStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[24],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StripeStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StripeStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StripeStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StripeStatistics::InternalGenerateClassData_(
            _default, &StripeStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StripeStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StripeStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StripeStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StripeStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StripeStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StripeStatisticsGlobalsTypeInternal StripeStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StripeStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StripeStatistics_globals_.GetClassData();
#else
  return StripeStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class RowIndexEntry::_Internal {
 public:
  using HasBits = decltype(::std::declval<RowIndexEntry>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_._has_bits_);
};

constexpr RowIndexEntry::ParseTableT_ RowIndexEntry::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::RowIndexEntry>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional .orc.proto.ColumnStatistics statistics = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_.statistics_)}},
      // repeated uint64 positions = 1 [packed = true];
      {::_pbi::TcParser::FastV64P1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_.positions_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated uint64 positions = 1 [packed = true];
      {PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_.positions_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedUInt64)},
      // optional .orc.proto.ColumnStatistics statistics = 2;
      {PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_.statistics_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr RowIndexEntry::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        positions_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::RowIndexEntry,
            PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_.positions_)>()
         }
        ,
        _positions_cached_byte_size_{0},
        statistics_{nullptr} {}

template <typename>
constexpr RowIndexEntry::RowIndexEntry(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL RowIndexEntry::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) RowIndexEntry(arena);
}
constexpr auto RowIndexEntry::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(RowIndexEntry), alignof(RowIndexEntry));
}
constexpr auto RowIndexEntry::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &RowIndexEntry::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<RowIndexEntry>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &RowIndexEntry::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<RowIndexEntry>(), &RowIndexEntry::ByteSizeLong,
              &RowIndexEntry::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(RowIndexEntry, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[12],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct RowIndexEntryGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr RowIndexEntryGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 RowIndexEntry_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(RowIndexEntry::InternalGenerateClassData_(
            _default, &RowIndexEntry_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<RowIndexEntry>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~RowIndexEntryGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) RowIndexEntry _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<RowIndexEntry>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(RowIndexEntryGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST RowIndexEntryGlobalsTypeInternal RowIndexEntry_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* RowIndexEntry_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return RowIndexEntry_globals_.GetClassData();
#else
  return RowIndexEntry_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Footer::_Internal {
 public:
  using HasBits = decltype(::std::declval<Footer>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Footer, _impl_._has_bits_);
};

constexpr Footer::ParseTableT_ Footer::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Footer, _impl_._has_bits_),
      0, // no _extensions_
      12, 120,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294963200,  // skipmap
      offsetof(ParseTableT_, field_entries),
      12,  // num_field_entries
      6,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::Footer>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // optional uint64 header_length = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(Footer, _impl_.header_length_), 6>(),
       {8, 6, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.header_length_)}},
      // optional uint64 content_length = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(Footer, _impl_.content_length_), 7>(),
       {16, 7, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.content_length_)}},
      // repeated .orc.proto.StripeInformation stripes = 3;
      {::_pbi::TcParser::FastMtR1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.stripes_)}},
      // repeated .orc.proto.Type types = 4;
      {::_pbi::TcParser::FastMtR1,
       {34, 1, 1,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.types_)}},
      // repeated .orc.proto.UserMetadataItem metadata = 5;
      {::_pbi::TcParser::FastMtR1,
       {42, 2, 2,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.metadata_)}},
      // optional uint64 number_of_rows = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(Footer, _impl_.number_of_rows_), 8>(),
       {48, 8, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.number_of_rows_)}},
      // repeated .orc.proto.ColumnStatistics statistics = 7;
      {::_pbi::TcParser::FastMtR1,
       {58, 3, 3,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.statistics_)}},
      // optional uint32 row_index_stride = 8;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Footer, _impl_.row_index_stride_), 9>(),
       {64, 9, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.row_index_stride_)}},
      // optional uint32 writer = 9;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Footer, _impl_.writer_), 10>(),
       {72, 10, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.writer_)}},
      // optional .orc.proto.Encryption encryption = 10;
      {::_pbi::TcParser::FastMtS1,
       {82, 5, 4,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.encryption_)}},
      // optional .orc.proto.CalendarKind calendar = 11;
      {::_pbi::TcParser::FastEr0S1,
       {88, 11, 2,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.calendar_)}},
      // optional string software_version = 12;
      {::_pbi::TcParser::FastBS1,
       {98, 4, 0,
        PROTOBUF_FIELD_OFFSET(Footer, _impl_.software_version_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // optional uint64 header_length = 1;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.header_length_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 content_length = 2;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.content_length_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // repeated .orc.proto.StripeInformation stripes = 3;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.stripes_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.Type types = 4;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.types_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // repeated .orc.proto.UserMetadataItem metadata = 5;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.metadata_), _Internal::kHasBitsOffset + 2, 2, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional uint64 number_of_rows = 6;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.number_of_rows_), _Internal::kHasBitsOffset + 8, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // repeated .orc.proto.ColumnStatistics statistics = 7;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.statistics_), _Internal::kHasBitsOffset + 3, 3, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional uint32 row_index_stride = 8;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.row_index_stride_), _Internal::kHasBitsOffset + 9, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional uint32 writer = 9;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.writer_), _Internal::kHasBitsOffset + 10, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // optional .orc.proto.Encryption encryption = 10;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.encryption_), _Internal::kHasBitsOffset + 5, 4, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.CalendarKind calendar = 11;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.calendar_), _Internal::kHasBitsOffset + 11, 5, (0 | ::_fl::kFcOptional | ::_fl::kEnumRange)},
      // optional string software_version = 12;
      {PROTOBUF_FIELD_OFFSET(Footer, _impl_.software_version_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::StripeInformation>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::StripeInformation_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Type>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Type_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::UserMetadataItem>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::UserMetadataItem_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnStatistics_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Encryption>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Encryption_globals_},
        #endif
        {0, 2},
    }},
    {{
    }},
  };
}


inline constexpr Footer::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        stripes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.stripes_)>()
         }
        ,
        types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.types_)>()
         }
        ,
        metadata_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.metadata_)>()
         }
        ,
        statistics_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.statistics_)>()
         }
        ,
        software_version_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        encryption_{nullptr},
        header_length_{::uint64_t{0u}},
        content_length_{::uint64_t{0u}},
        number_of_rows_{::uint64_t{0u}},
        row_index_stride_{0u},
        writer_{0u},
        calendar_{static_cast< ::orc::proto::CalendarKind >(0)} {}

template <typename>
constexpr Footer::Footer(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Footer::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Footer(arena);
}
constexpr auto Footer::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Footer), alignof(Footer));
}
constexpr auto Footer::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Footer::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Footer>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Footer::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Footer>(), &Footer::ByteSizeLong,
              &Footer::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Footer, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[32],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FooterGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FooterGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Footer_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Footer::InternalGenerateClassData_(
            _default, &Footer_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Footer>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FooterGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Footer _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Footer>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FooterGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FooterGlobalsTypeInternal Footer_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Footer_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Footer_globals_.GetClassData();
#else
  return Footer_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class FileStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<FileStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FileStatistics, _impl_._has_bits_);
};

constexpr FileStatistics::ParseTableT_ FileStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(FileStatistics, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::FileStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.ColumnStatistics column = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(FileStatistics, _impl_.column_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.ColumnStatistics column = 1;
      {PROTOBUF_FIELD_OFFSET(FileStatistics, _impl_.column_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr FileStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        column_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::FileStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::FileStatistics, _impl_.column_)>()
         }
     {}

template <typename>
constexpr FileStatistics::FileStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL FileStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FileStatistics(arena);
}
constexpr auto FileStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(FileStatistics), alignof(FileStatistics));
}
constexpr auto FileStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &FileStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FileStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FileStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FileStatistics>(), &FileStatistics::ByteSizeLong,
              &FileStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FileStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[27],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FileStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FileStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 FileStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(FileStatistics::InternalGenerateClassData_(
            _default, &FileStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<FileStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FileStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) FileStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<FileStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FileStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FileStatisticsGlobalsTypeInternal FileStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* FileStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return FileStatistics_globals_.GetClassData();
#else
  return FileStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ColumnarStripeStatistics::_Internal {
 public:
  using HasBits = decltype(::std::declval<ColumnarStripeStatistics>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ColumnarStripeStatistics, _impl_._has_bits_);
};

constexpr ColumnarStripeStatistics::ParseTableT_ ColumnarStripeStatistics::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ColumnarStripeStatistics, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::ColumnarStripeStatistics>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.ColumnStatistics col_stats = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ColumnarStripeStatistics, _impl_.col_stats_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.ColumnStatistics col_stats = 1;
      {PROTOBUF_FIELD_OFFSET(ColumnarStripeStatistics, _impl_.col_stats_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::ColumnStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::ColumnStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr ColumnarStripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        col_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::ColumnarStripeStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnarStripeStatistics, _impl_.col_stats_)>()
         }
     {}

template <typename>
constexpr ColumnarStripeStatistics::ColumnarStripeStatistics(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ColumnarStripeStatistics::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ColumnarStripeStatistics(arena);
}
constexpr auto ColumnarStripeStatistics::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ColumnarStripeStatistics), alignof(ColumnarStripeStatistics));
}
constexpr auto ColumnarStripeStatistics::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ColumnarStripeStatistics::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ColumnarStripeStatistics>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ColumnarStripeStatistics::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ColumnarStripeStatistics>(), &ColumnarStripeStatistics::ByteSizeLong,
              &ColumnarStripeStatistics::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ColumnarStripeStatistics, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[26],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ColumnarStripeStatisticsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ColumnarStripeStatisticsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ColumnarStripeStatistics_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ColumnarStripeStatistics::InternalGenerateClassData_(
            _default, &ColumnarStripeStatistics_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ColumnarStripeStatistics>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ColumnarStripeStatisticsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ColumnarStripeStatistics _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ColumnarStripeStatistics>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ColumnarStripeStatisticsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ColumnarStripeStatisticsGlobalsTypeInternal ColumnarStripeStatistics_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ColumnarStripeStatistics_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ColumnarStripeStatistics_globals_.GetClassData();
#else
  return ColumnarStripeStatistics_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class RowIndex::_Internal {
 public:
  using HasBits = decltype(::std::declval<RowIndex>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(RowIndex, _impl_._has_bits_);
};

constexpr RowIndex::ParseTableT_ RowIndex::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(RowIndex, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::RowIndex>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.RowIndexEntry entry = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(RowIndex, _impl_.entry_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.RowIndexEntry entry = 1;
      {PROTOBUF_FIELD_OFFSET(RowIndex, _impl_.entry_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::RowIndexEntry>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::RowIndexEntry_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr RowIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        entry_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::RowIndex,
            PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndex, _impl_.entry_)>()
         }
     {}

template <typename>
constexpr RowIndex::RowIndex(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL RowIndex::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) RowIndex(arena);
}
constexpr auto RowIndex::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(RowIndex), alignof(RowIndex));
}
constexpr auto RowIndex::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &RowIndex::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<RowIndex>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &RowIndex::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<RowIndex>(), &RowIndex::ByteSizeLong,
              &RowIndex::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(RowIndex, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[13],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct RowIndexGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr RowIndexGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 RowIndex_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(RowIndex::InternalGenerateClassData_(
            _default, &RowIndex_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<RowIndex>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~RowIndexGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) RowIndex _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<RowIndex>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(RowIndexGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST RowIndexGlobalsTypeInternal RowIndex_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* RowIndex_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return RowIndex_globals_.GetClassData();
#else
  return RowIndex_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Metadata::_Internal {
 public:
  using HasBits = decltype(::std::declval<Metadata>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Metadata, _impl_._has_bits_);
};

constexpr Metadata::ParseTableT_ Metadata::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Metadata, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::Metadata>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .orc.proto.StripeStatistics stripe_stats = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Metadata, _impl_.stripe_stats_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .orc.proto.StripeStatistics stripe_stats = 1;
      {PROTOBUF_FIELD_OFFSET(Metadata, _impl_.stripe_stats_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::StripeStatistics>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::StripeStatistics_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr Metadata::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        stripe_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Metadata,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Metadata, _impl_.stripe_stats_)>()
         }
     {}

template <typename>
constexpr Metadata::Metadata(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Metadata::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Metadata(arena);
}
constexpr auto Metadata::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Metadata), alignof(Metadata));
}
constexpr auto Metadata::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Metadata::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Metadata>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Metadata::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Metadata>(), &Metadata::ByteSizeLong,
              &Metadata::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Metadata, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[25],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct MetadataGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr MetadataGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Metadata_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Metadata::InternalGenerateClassData_(
            _default, &Metadata_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Metadata>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~MetadataGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Metadata _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Metadata>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(MetadataGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST MetadataGlobalsTypeInternal Metadata_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Metadata_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Metadata_globals_.GetClassData();
#else
  return Metadata_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class FileTail::_Internal {
 public:
  using HasBits = decltype(::std::declval<FileTail>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(FileTail, _impl_._has_bits_);
};

constexpr FileTail::ParseTableT_ FileTail::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(FileTail, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::orc::proto::FileTail>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // optional uint64 postscript_length = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(FileTail, _impl_.postscript_length_), 3>(),
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_length_)}},
      // optional .orc.proto.PostScript postscript = 1;
      {::_pbi::TcParser::FastMtS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_)}},
      // optional .orc.proto.Footer footer = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 1, 1,
        PROTOBUF_FIELD_OFFSET(FileTail, _impl_.footer_)}},
      // optional uint64 file_length = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(FileTail, _impl_.file_length_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(FileTail, _impl_.file_length_)}},
    }}, {{
      65535, 65535
    }}, {{
      // optional .orc.proto.PostScript postscript = 1;
      {PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional .orc.proto.Footer footer = 2;
      {PROTOBUF_FIELD_OFFSET(FileTail, _impl_.footer_), _Internal::kHasBitsOffset + 1, 1, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // optional uint64 file_length = 3;
      {PROTOBUF_FIELD_OFFSET(FileTail, _impl_.file_length_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // optional uint64 postscript_length = 4;
      {PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_length_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::PostScript>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::PostScript_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::orc::proto::Footer>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::orc::proto::Footer_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr FileTail::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        postscript_{nullptr},
        footer_{nullptr},
        file_length_{::uint64_t{0u}},
        postscript_length_{::uint64_t{0u}} {}

template <typename>
constexpr FileTail::FileTail(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL FileTail::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) FileTail(arena);
}
constexpr auto FileTail::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(FileTail), alignof(FileTail));
}
constexpr auto FileTail::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &FileTail::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<FileTail>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &FileTail::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<FileTail>(), &FileTail::ByteSizeLong,
              &FileTail::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(FileTail, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[34],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_orc_5fproto_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FileTailGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FileTailGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 FileTail_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(FileTail::InternalGenerateClassData_(
            _default, &FileTail_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<FileTail>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FileTailGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) FileTail _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<FileTail>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FileTailGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FileTailGlobalsTypeInternal FileTail_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* FileTail_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return FileTail_globals_.GetClassData();
#else
  return FileTail_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace proto
}  // namespace orc
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_orc_5fproto_2eproto[8];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_orc_5fproto_2eproto = nullptr;
const ::uint32_t
    TableStruct_orc_5fproto_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::IntegerStatistics, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::IntegerStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::IntegerStatistics, _impl_.maximum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::IntegerStatistics, _impl_.sum_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::DoubleStatistics, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::DoubleStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DoubleStatistics, _impl_.maximum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DoubleStatistics, _impl_.sum_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_.maximum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_.sum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_.lower_bound_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringStatistics, _impl_.upper_bound_),
        0,
        1,
        4,
        2,
        3,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::BucketStatistics, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::BucketStatistics, _impl_.count_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::DecimalStatistics, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::DecimalStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DecimalStatistics, _impl_.maximum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DecimalStatistics, _impl_.sum_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::DateStatistics, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::DateStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DateStatistics, _impl_.maximum_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.minimum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.maximum_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.minimum_utc_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.maximum_utc_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.minimum_nanos_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::TimestampStatistics, _impl_.maximum_nanos_),
        0,
        1,
        2,
        3,
        4,
        5,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::BinaryStatistics, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::BinaryStatistics, _impl_.sum_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::CollectionStatistics, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::CollectionStatistics, _impl_.min_children_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::CollectionStatistics, _impl_.max_children_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::CollectionStatistics, _impl_.total_children_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_._has_bits_),
        11, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.xmin_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.xmax_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.ymin_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.ymax_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.zmin_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.zmax_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.mmin_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BoundingBox, _impl_.mmax_),
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_.bbox_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_.geospatial_types_),
        1,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_._has_bits_),
        16, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.number_of_values_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.int_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.double_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.string_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.bucket_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.decimal_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.date_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.binary_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.timestamp_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.has_null_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.bytes_on_disk_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.collection_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnStatistics, _impl_.geospatial_statistics_),
        10,
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        12,
        11,
        8,
        9,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_.positions_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_.statistics_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndex, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndex, _impl_.entry_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.num_hash_functions_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.bitset_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.utf8bitset_),
        2,
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilterIndex, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilterIndex, _impl_.bloom_filter_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::Stream, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::Stream, _impl_.kind_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Stream, _impl_.column_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Stream, _impl_.length_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnEncoding, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnEncoding, _impl_.kind_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnEncoding, _impl_.dictionary_size_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnEncoding, _impl_.bloom_encoding_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.streams_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.encoding_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.streams_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.columns_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.writer_timezone_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.encryption_),
        0,
        1,
        3,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringPair, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringPair, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StringPair, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_._has_bits_),
        12, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.kind_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.subtypes_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.field_names_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.maximum_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.precision_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.scale_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.attributes_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.crs_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.algorithm_),
        4,
        0,
        1,
        5,
        6,
        7,
        2,
        3,
        8,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_._has_bits_),
        10, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.offset_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.index_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.data_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.footer_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.number_of_rows_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.encrypt_stripe_id_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.encrypted_local_keys_),
        1,
        2,
        3,
        4,
        5,
        6,
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::UserMetadataItem, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::UserMetadataItem, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::UserMetadataItem, _impl_.value_),
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeStatistics, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::StripeStatistics, _impl_.col_stats_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::Metadata, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::Metadata, _impl_.stripe_stats_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnarStripeStatistics, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnarStripeStatistics, _impl_.col_stats_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileStatistics, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileStatistics, _impl_.column_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.mask_parameters_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.columns_),
        2,
        0,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionKey, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionKey, _impl_.key_name_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionKey, _impl_.key_version_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionKey, _impl_.algorithm_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.root_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.encrypted_key_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.stripe_statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.file_statistics_),
        3,
        4,
        1,
        0,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.mask_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.key_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.variants_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.key_provider_),
        0,
        1,
        2,
        3,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_._has_bits_),
        15, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.header_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.content_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.stripes_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.types_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.metadata_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.number_of_rows_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.statistics_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.row_index_stride_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.writer_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.encryption_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.calendar_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.software_version_),
        6,
        7,
        0,
        1,
        2,
        8,
        3,
        9,
        10,
        5,
        11,
        4,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_._has_bits_),
        11, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.footer_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.compression_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.compression_block_size_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.version_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.metadata_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.writer_version_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.stripe_statistics_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.magic_),
        2,
        4,
        3,
        0,
        6,
        5,
        7,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileTail, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileTail, _impl_.postscript_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileTail, _impl_.footer_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileTail, _impl_.file_length_),
        PROTOBUF_FIELD_OFFSET(::orc::proto::FileTail, _impl_.postscript_length_),
        0,
        1,
        2,
        3,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::orc::proto::IntegerStatistics)},
        {9, sizeof(::orc::proto::DoubleStatistics)},
        {18, sizeof(::orc::proto::StringStatistics)},
        {31, sizeof(::orc::proto::BucketStatistics)},
        {36, sizeof(::orc::proto::DecimalStatistics)},
        {45, sizeof(::orc::proto::DateStatistics)},
        {52, sizeof(::orc::proto::TimestampStatistics)},
        {67, sizeof(::orc::proto::BinaryStatistics)},
        {72, sizeof(::orc::proto::CollectionStatistics)},
        {81, sizeof(::orc::proto::BoundingBox)},
        {100, sizeof(::orc::proto::GeospatialStatistics)},
        {107, sizeof(::orc::proto::ColumnStatistics)},
        {136, sizeof(::orc::proto::RowIndexEntry)},
        {143, sizeof(::orc::proto::RowIndex)},
        {148, sizeof(::orc::proto::BloomFilter)},
        {157, sizeof(::orc::proto::BloomFilterIndex)},
        {162, sizeof(::orc::proto::Stream)},
        {171, sizeof(::orc::proto::ColumnEncoding)},
        {180, sizeof(::orc::proto::StripeEncryptionVariant)},
        {187, sizeof(::orc::proto::StripeFooter)},
        {198, sizeof(::orc::proto::StringPair)},
        {205, sizeof(::orc::proto::Type)},
        {226, sizeof(::orc::proto::StripeInformation)},
        {243, sizeof(::orc::proto::UserMetadataItem)},
        {250, sizeof(::orc::proto::StripeStatistics)},
        {255, sizeof(::orc::proto::Metadata)},
        {260, sizeof(::orc::proto::ColumnarStripeStatistics)},
        {265, sizeof(::orc::proto::FileStatistics)},
        {270, sizeof(::orc::proto::DataMask)},
        {279, sizeof(::orc::proto::EncryptionKey)},
        {288, sizeof(::orc::proto::EncryptionVariant)},
        {301, sizeof(::orc::proto::Encryption)},
        {312, sizeof(::orc::proto::Footer)},
        {339, sizeof(::orc::proto::PostScript)},
        {358, sizeof(::orc::proto::FileTail)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::orc::proto::IntegerStatistics_globals_,
        &::orc::proto::DoubleStatistics_globals_,
        &::orc::proto::StringStatistics_globals_,
        &::orc::proto::BucketStatistics_globals_,
        &::orc::proto::DecimalStatistics_globals_,
        &::orc::proto::DateStatistics_globals_,
        &::orc::proto::TimestampStatistics_globals_,
        &::orc::proto::BinaryStatistics_globals_,
        &::orc::proto::CollectionStatistics_globals_,
        &::orc::proto::BoundingBox_globals_,
        &::orc::proto::GeospatialStatistics_globals_,
        &::orc::proto::ColumnStatistics_globals_,
        &::orc::proto::RowIndexEntry_globals_,
        &::orc::proto::RowIndex_globals_,
        &::orc::proto::BloomFilter_globals_,
        &::orc::proto::BloomFilterIndex_globals_,
        &::orc::proto::Stream_globals_,
        &::orc::proto::ColumnEncoding_globals_,
        &::orc::proto::StripeEncryptionVariant_globals_,
        &::orc::proto::StripeFooter_globals_,
        &::orc::proto::StringPair_globals_,
        &::orc::proto::Type_globals_,
        &::orc::proto::StripeInformation_globals_,
        &::orc::proto::UserMetadataItem_globals_,
        &::orc::proto::StripeStatistics_globals_,
        &::orc::proto::Metadata_globals_,
        &::orc::proto::ColumnarStripeStatistics_globals_,
        &::orc::proto::FileStatistics_globals_,
        &::orc::proto::DataMask_globals_,
        &::orc::proto::EncryptionKey_globals_,
        &::orc::proto::EncryptionVariant_globals_,
        &::orc::proto::Encryption_globals_,
        &::orc::proto::Footer_globals_,
        &::orc::proto::PostScript_globals_,
        &::orc::proto::FileTail_globals_,
};
const char descriptor_table_protodef_orc_5fproto_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\017orc_proto.proto\022\torc.proto\"B\n\021IntegerS"
    "tatistics\022\017\n\007minimum\030\001 \001(\022\022\017\n\007maximum\030\002 "
    "\001(\022\022\013\n\003sum\030\003 \001(\022\"A\n\020DoubleStatistics\022\017\n\007"
    "minimum\030\001 \001(\001\022\017\n\007maximum\030\002 \001(\001\022\013\n\003sum\030\003 "
    "\001(\001\"k\n\020StringStatistics\022\017\n\007minimum\030\001 \001(\t"
    "\022\017\n\007maximum\030\002 \001(\t\022\013\n\003sum\030\003 \001(\022\022\023\n\013lower_"
    "bound\030\004 \001(\t\022\023\n\013upper_bound\030\005 \001(\t\"%\n\020Buck"
    "etStatistics\022\021\n\005count\030\001 \003(\004B\002\020\001\"B\n\021Decim"
    "alStatistics\022\017\n\007minimum\030\001 \001(\t\022\017\n\007maximum"
    "\030\002 \001(\t\022\013\n\003sum\030\003 \001(\t\"2\n\016DateStatistics\022\017\n"
    "\007minimum\030\001 \001(\021\022\017\n\007maximum\030\002 \001(\021\"\217\001\n\023Time"
    "stampStatistics\022\017\n\007minimum\030\001 \001(\022\022\017\n\007maxi"
    "mum\030\002 \001(\022\022\023\n\013minimum_utc\030\003 \001(\022\022\023\n\013maximu"
    "m_utc\030\004 \001(\022\022\025\n\rminimum_nanos\030\005 \001(\005\022\025\n\rma"
    "ximum_nanos\030\006 \001(\005\"\037\n\020BinaryStatistics\022\013\n"
    "\003sum\030\001 \001(\022\"Z\n\024CollectionStatistics\022\024\n\014mi"
    "n_children\030\001 \001(\004\022\024\n\014max_children\030\002 \001(\004\022\026"
    "\n\016total_children\030\003 \001(\004\"}\n\013BoundingBox\022\014\n"
    "\004xmin\030\001 \001(\001\022\014\n\004xmax\030\002 \001(\001\022\014\n\004ymin\030\003 \001(\001\022"
    "\014\n\004ymax\030\004 \001(\001\022\014\n\004zmin\030\005 \001(\001\022\014\n\004zmax\030\006 \001("
    "\001\022\014\n\004mmin\030\007 \001(\001\022\014\n\004mmax\030\010 \001(\001\"V\n\024Geospat"
    "ialStatistics\022$\n\004bbox\030\001 \001(\0132\026.orc.proto."
    "BoundingBox\022\030\n\020geospatial_types\030\002 \003(\005\"\227\005"
    "\n\020ColumnStatistics\022\030\n\020number_of_values\030\001"
    " \001(\004\0224\n\016int_statistics\030\002 \001(\0132\034.orc.proto"
    ".IntegerStatistics\0226\n\021double_statistics\030"
    "\003 \001(\0132\033.orc.proto.DoubleStatistics\0226\n\021st"
    "ring_statistics\030\004 \001(\0132\033.orc.proto.String"
    "Statistics\0226\n\021bucket_statistics\030\005 \001(\0132\033."
    "orc.proto.BucketStatistics\0228\n\022decimal_st"
    "atistics\030\006 \001(\0132\034.orc.proto.DecimalStatis"
    "tics\0222\n\017date_statistics\030\007 \001(\0132\031.orc.prot"
    "o.DateStatistics\0226\n\021binary_statistics\030\010 "
    "\001(\0132\033.orc.proto.BinaryStatistics\022<\n\024time"
    "stamp_statistics\030\t \001(\0132\036.orc.proto.Times"
    "tampStatistics\022\020\n\010has_null\030\n \001(\010\022\025\n\rbyte"
    "s_on_disk\030\013 \001(\004\022>\n\025collection_statistics"
    "\030\014 \001(\0132\037.orc.proto.CollectionStatistics\022"
    ">\n\025geospatial_statistics\030\r \001(\0132\037.orc.pro"
    "to.GeospatialStatistics\"W\n\rRowIndexEntry"
    "\022\025\n\tpositions\030\001 \003(\004B\002\020\001\022/\n\nstatistics\030\002 "
    "\001(\0132\033.orc.proto.ColumnStatistics\"3\n\010RowI"
    "ndex\022\'\n\005entry\030\001 \003(\0132\030.orc.proto.RowIndex"
    "Entry\"M\n\013BloomFilter\022\032\n\022num_hash_functio"
    "ns\030\001 \001(\r\022\016\n\006bitset\030\002 \003(\006\022\022\n\nutf8bitset\030\003"
    " \001(\014\"@\n\020BloomFilterIndex\022,\n\014bloom_filter"
    "\030\001 \003(\0132\026.orc.proto.BloomFilter\"\301\002\n\006Strea"
    "m\022$\n\004kind\030\001 \001(\0162\026.orc.proto.Stream.Kind\022"
    "\016\n\006column\030\002 \001(\r\022\016\n\006length\030\003 \001(\004\"\360\001\n\004Kind"
    "\022\013\n\007PRESENT\020\000\022\010\n\004DATA\020\001\022\n\n\006LENGTH\020\002\022\023\n\017D"
    "ICTIONARY_DATA\020\003\022\024\n\020DICTIONARY_COUNT\020\004\022\r"
    "\n\tSECONDARY\020\005\022\r\n\tROW_INDEX\020\006\022\020\n\014BLOOM_FI"
    "LTER\020\007\022\025\n\021BLOOM_FILTER_UTF8\020\010\022\023\n\017ENCRYPT"
    "ED_INDEX\020\t\022\022\n\016ENCRYPTED_DATA\020\n\022\025\n\021STRIPE"
    "_STATISTICS\020d\022\023\n\017FILE_STATISTICS\020e\"\265\001\n\016C"
    "olumnEncoding\022,\n\004kind\030\001 \001(\0162\036.orc.proto."
    "ColumnEncoding.Kind\022\027\n\017dictionary_size\030\002"
    " \001(\r\022\026\n\016bloom_encoding\030\003 \001(\r\"D\n\004Kind\022\n\n\006"
    "DIRECT\020\000\022\016\n\nDICTIONARY\020\001\022\r\n\tDIRECT_V2\020\002\022"
    "\021\n\rDICTIONARY_V2\020\003\"j\n\027StripeEncryptionVa"
    "riant\022\"\n\007streams\030\001 \003(\0132\021.orc.proto.Strea"
    "m\022+\n\010encoding\030\002 \003(\0132\031.orc.proto.ColumnEn"
    "coding\"\257\001\n\014StripeFooter\022\"\n\007streams\030\001 \003(\013"
    "2\021.orc.proto.Stream\022*\n\007columns\030\002 \003(\0132\031.o"
    "rc.proto.ColumnEncoding\022\027\n\017writer_timezo"
    "ne\030\003 \001(\t\0226\n\nencryption\030\004 \003(\0132\".orc.proto"
    ".StripeEncryptionVariant\"(\n\nStringPair\022\013"
    "\n\003key\030\001 \001(\t\022\r\n\005value\030\002 \001(\t\"\356\004\n\004Type\022\"\n\004k"
    "ind\030\001 \001(\0162\024.orc.proto.Type.Kind\022\024\n\010subty"
    "pes\030\002 \003(\rB\002\020\001\022\023\n\013field_names\030\003 \003(\t\022\026\n\016ma"
    "ximum_length\030\004 \001(\r\022\021\n\tprecision\030\005 \001(\r\022\r\n"
    "\005scale\030\006 \001(\r\022)\n\nattributes\030\007 \003(\0132\025.orc.p"
    "roto.StringPair\022\013\n\003crs\030\010 \001(\t\022=\n\talgorith"
    "m\030\t \001(\0162*.orc.proto.Type.EdgeInterpolati"
    "onAlgorithm\"\205\002\n\004Kind\022\013\n\007BOOLEAN\020\000\022\010\n\004BYT"
    "E\020\001\022\t\n\005SHORT\020\002\022\007\n\003INT\020\003\022\010\n\004LONG\020\004\022\t\n\005FLO"
    "AT\020\005\022\n\n\006DOUBLE\020\006\022\n\n\006STRING\020\007\022\n\n\006BINARY\020\010"
    "\022\r\n\tTIMESTAMP\020\t\022\010\n\004LIST\020\n\022\007\n\003MAP\020\013\022\n\n\006ST"
    "RUCT\020\014\022\t\n\005UNION\020\r\022\013\n\007DECIMAL\020\016\022\010\n\004DATE\020\017"
    "\022\013\n\007VARCHAR\020\020\022\010\n\004CHAR\020\021\022\025\n\021TIMESTAMP_INS"
    "TANT\020\022\022\014\n\010GEOMETRY\020\023\022\r\n\tGEOGRAPHY\020\024\"^\n\032E"
    "dgeInterpolationAlgorithm\022\r\n\tSPHERICAL\020\000"
    "\022\014\n\010VINCENTY\020\001\022\n\n\006THOMAS\020\002\022\013\n\007ANDOYER\020\003\022"
    "\n\n\006KARNEY\020\004\"\266\001\n\021StripeInformation\022\016\n\006off"
    "set\030\001 \001(\004\022\024\n\014index_length\030\002 \001(\004\022\023\n\013data_"
    "length\030\003 \001(\004\022\025\n\rfooter_length\030\004 \001(\004\022\026\n\016n"
    "umber_of_rows\030\005 \001(\004\022\031\n\021encrypt_stripe_id"
    "\030\006 \001(\004\022\034\n\024encrypted_local_keys\030\007 \003(\014\"/\n\020"
    "UserMetadataItem\022\014\n\004name\030\001 \001(\t\022\r\n\005value\030"
    "\002 \001(\014\"B\n\020StripeStatistics\022.\n\tcol_stats\030\001"
    " \003(\0132\033.orc.proto.ColumnStatistics\"=\n\010Met"
    "adata\0221\n\014stripe_stats\030\001 \003(\0132\033.orc.proto."
    "StripeStatistics\"J\n\030ColumnarStripeStatis"
    "tics\022.\n\tcol_stats\030\001 \003(\0132\033.orc.proto.Colu"
    "mnStatistics\"=\n\016FileStatistics\022+\n\006column"
    "\030\001 \003(\0132\033.orc.proto.ColumnStatistics\"F\n\010D"
    "ataMask\022\014\n\004name\030\001 \001(\t\022\027\n\017mask_parameters"
    "\030\002 \003(\t\022\023\n\007columns\030\003 \003(\rB\002\020\001\"i\n\rEncryptio"
    "nKey\022\020\n\010key_name\030\001 \001(\t\022\023\n\013key_version\030\002 "
    "\001(\r\0221\n\talgorithm\030\003 \001(\0162\036.orc.proto.Encry"
    "ptionAlgorithm\"\214\001\n\021EncryptionVariant\022\014\n\004"
    "root\030\001 \001(\r\022\013\n\003key\030\002 \001(\r\022\025\n\rencrypted_key"
    "\030\003 \001(\014\022,\n\021stripe_statistics\030\004 \003(\0132\021.orc."
    "proto.Stream\022\027\n\017file_statistics\030\005 \001(\014\"\270\001"
    "\n\nEncryption\022!\n\004mask\030\001 \003(\0132\023.orc.proto.D"
    "ataMask\022%\n\003key\030\002 \003(\0132\030.orc.proto.Encrypt"
    "ionKey\022.\n\010variants\030\003 \003(\0132\034.orc.proto.Enc"
    "ryptionVariant\0220\n\014key_provider\030\004 \001(\0162\032.o"
    "rc.proto.KeyProviderKind\"\230\003\n\006Footer\022\025\n\rh"
    "eader_length\030\001 \001(\004\022\026\n\016content_length\030\002 \001"
    "(\004\022-\n\007stripes\030\003 \003(\0132\034.orc.proto.StripeIn"
    "formation\022\036\n\005types\030\004 \003(\0132\017.orc.proto.Typ"
    "e\022-\n\010metadata\030\005 \003(\0132\033.orc.proto.UserMeta"
    "dataItem\022\026\n\016number_of_rows\030\006 \001(\004\022/\n\nstat"
    "istics\030\007 \003(\0132\033.orc.proto.ColumnStatistic"
    "s\022\030\n\020row_index_stride\030\010 \001(\r\022\016\n\006writer\030\t "
    "\001(\r\022)\n\nencryption\030\n \001(\0132\025.orc.proto.Encr"
    "yption\022)\n\010calendar\030\013 \001(\0162\027.orc.proto.Cal"
    "endarKind\022\030\n\020software_version\030\014 \001(\t\"\354\001\n\n"
    "PostScript\022\025\n\rfooter_length\030\001 \001(\004\022/\n\013com"
    "pression\030\002 \001(\0162\032.orc.proto.CompressionKi"
    "nd\022\036\n\026compression_block_size\030\003 \001(\004\022\023\n\007ve"
    "rsion\030\004 \003(\rB\002\020\001\022\027\n\017metadata_length\030\005 \001(\004"
    "\022\026\n\016writer_version\030\006 \001(\r\022 \n\030stripe_stati"
    "stics_length\030\007 \001(\004\022\016\n\005magic\030\300> \001(\t\"\210\001\n\010F"
    "ileTail\022)\n\npostscript\030\001 \001(\0132\025.orc.proto."
    "PostScript\022!\n\006footer\030\002 \001(\0132\021.orc.proto.F"
    "ooter\022\023\n\013file_length\030\003 \001(\004\022\031\n\021postscript"
    "_length\030\004 \001(\004*O\n\023EncryptionAlgorithm\022\026\n\022"
    "UNKNOWN_ENCRYPTION\020\000\022\017\n\013AES_CTR_128\020\001\022\017\n"
    "\013AES_CTR_256\020\002*G\n\017KeyProviderKind\022\013\n\007UNK"
    "NOWN\020\000\022\n\n\006HADOOP\020\001\022\007\n\003AWS\020\002\022\007\n\003GCP\020\003\022\t\n\005"
    "AZURE\020\004*S\n\014CalendarKind\022\024\n\020UNKNOWN_CALEN"
    "DAR\020\000\022\024\n\020JULIAN_GREGORIAN\020\001\022\027\n\023PROLEPTIC"
    "_GREGORIAN\020\002*Y\n\017CompressionKind\022\010\n\004NONE\020"
    "\000\022\010\n\004ZLIB\020\001\022\n\n\006SNAPPY\020\002\022\007\n\003LZO\020\003\022\007\n\003LZ4\020"
    "\004\022\010\n\004ZSTD\020\005\022\n\n\006BROTLI\020\006B\020\n\016org.apache.or"
    "c"
};
static ::absl::once_flag descriptor_table_orc_5fproto_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_orc_5fproto_2eproto = {
    false,
    false,
    5481,
    descriptor_table_protodef_orc_5fproto_2eproto,
    "orc_proto.proto",
    &descriptor_table_orc_5fproto_2eproto_once,
    nullptr,
    0,
    35,
    schemas,
    file_message_globals,
    TableStruct_orc_5fproto_2eproto::offsets,
    file_level_enum_descriptors_orc_5fproto_2eproto,
    file_level_service_descriptors_orc_5fproto_2eproto,
};
namespace orc {
namespace proto {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Stream_Kind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t Stream_Kind_internal_data_[] = {
    720896u, 131072u, 101u, 100u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
ColumnEncoding_Kind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[1];
}
PROTOBUF_CONSTINIT const uint32_t ColumnEncoding_Kind_internal_data_[] = {
    262144u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Type_Kind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[2];
}
PROTOBUF_CONSTINIT const uint32_t Type_Kind_internal_data_[] = {
    1376256u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
Type_EdgeInterpolationAlgorithm_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[3];
}
PROTOBUF_CONSTINIT const uint32_t Type_EdgeInterpolationAlgorithm_internal_data_[] = {
    327680u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
EncryptionAlgorithm_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[4];
}
PROTOBUF_CONSTINIT const uint32_t EncryptionAlgorithm_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
KeyProviderKind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[5];
}
PROTOBUF_CONSTINIT const uint32_t KeyProviderKind_internal_data_[] = {
    327680u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CalendarKind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[6];
}
PROTOBUF_CONSTINIT const uint32_t CalendarKind_internal_data_[] = {
    196608u, 0u, };
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
CompressionKind_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_orc_5fproto_2eproto);
  return file_level_enum_descriptors_orc_5fproto_2eproto[7];
}
PROTOBUF_CONSTINIT const uint32_t CompressionKind_internal_data_[] = {
    458752u, 0u, };
// ===================================================================

IntegerStatistics::IntegerStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, IntegerStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.IntegerStatistics)
}
IntegerStatistics::IntegerStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const IntegerStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, IntegerStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE IntegerStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void IntegerStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, minimum_),
           0,
           offsetof(Impl_, sum_) -
               offsetof(Impl_, minimum_) +
               sizeof(Impl_::sum_));
}
IntegerStatistics::~IntegerStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.IntegerStatistics)
  SharedDtor(*this);
}
inline void IntegerStatistics::SharedDtor(MessageLite& self) {
  IntegerStatistics& this_ = static_cast<IntegerStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull IntegerStatistics_class_data_ =
        IntegerStatistics::InternalGenerateClassData_(IntegerStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
IntegerStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&IntegerStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(IntegerStatistics_class_data_.tc_table);
  return IntegerStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
IntegerStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&IntegerStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&IntegerStatistics_globals_));
  return IntegerStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const IntegerStatistics::ParseTableT_
    IntegerStatistics::_table_ =
        IntegerStatistics::InternalGenerateParseTable_(IntegerStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void IntegerStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.IntegerStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    ::memset(&_impl_.minimum_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.sum_) -
        reinterpret_cast<char*>(&_impl_.minimum_)) + sizeof(_impl_.sum_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL IntegerStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const IntegerStatistics& this_ = static_cast<const IntegerStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL IntegerStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const IntegerStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.IntegerStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional sint64 minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        1, this_._internal_minimum(), target);
  }

  // optional sint64 maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        2, this_._internal_maximum(), target);
  }

  // optional sint64 sum = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        3, this_._internal_sum(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.IntegerStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t IntegerStatistics::ByteSizeLong(const MessageLite& base) {
  const IntegerStatistics& this_ = static_cast<const IntegerStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t IntegerStatistics::ByteSizeLong() const {
  const IntegerStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.IntegerStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional sint64 minimum = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_minimum());
    }
    // optional sint64 maximum = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_maximum());
    }
    // optional sint64 sum = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_sum());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void IntegerStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<IntegerStatistics*>(&to_msg);
  auto& from = static_cast<const IntegerStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.IntegerStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.minimum_ = from._impl_.minimum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.maximum_ = from._impl_.maximum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.sum_ = from._impl_.sum_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void IntegerStatistics::CopyFrom(const IntegerStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.IntegerStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void IntegerStatistics::InternalSwap(IntegerStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.sum_)
      + sizeof(IntegerStatistics::_impl_.sum_)
      - PROTOBUF_FIELD_OFFSET(IntegerStatistics, _impl_.minimum_)>(
          reinterpret_cast<char*>(&_impl_.minimum_),
          reinterpret_cast<char*>(&other->_impl_.minimum_));
}

::google::protobuf::Metadata IntegerStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DoubleStatistics::DoubleStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoubleStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.DoubleStatistics)
}
DoubleStatistics::DoubleStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const DoubleStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DoubleStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE DoubleStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void DoubleStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, minimum_),
           0,
           offsetof(Impl_, sum_) -
               offsetof(Impl_, minimum_) +
               sizeof(Impl_::sum_));
}
DoubleStatistics::~DoubleStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.DoubleStatistics)
  SharedDtor(*this);
}
inline void DoubleStatistics::SharedDtor(MessageLite& self) {
  DoubleStatistics& this_ = static_cast<DoubleStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DoubleStatistics_class_data_ =
        DoubleStatistics::InternalGenerateClassData_(DoubleStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoubleStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoubleStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DoubleStatistics_class_data_.tc_table);
  return DoubleStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DoubleStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DoubleStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DoubleStatistics_globals_));
  return DoubleStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DoubleStatistics::ParseTableT_
    DoubleStatistics::_table_ =
        DoubleStatistics::InternalGenerateParseTable_(DoubleStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DoubleStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.DoubleStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    ::memset(&_impl_.minimum_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.sum_) -
        reinterpret_cast<char*>(&_impl_.minimum_)) + sizeof(_impl_.sum_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DoubleStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DoubleStatistics& this_ = static_cast<const DoubleStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DoubleStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DoubleStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.DoubleStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional double minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        1, this_._internal_minimum(), target);
  }

  // optional double maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        2, this_._internal_maximum(), target);
  }

  // optional double sum = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        3, this_._internal_sum(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.DoubleStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DoubleStatistics::ByteSizeLong(const MessageLite& base) {
  const DoubleStatistics& this_ = static_cast<const DoubleStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DoubleStatistics::ByteSizeLong() const {
  const DoubleStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.DoubleStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  total_size += ::absl::popcount(0x00000007U & cached_has_bits) * 9;
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DoubleStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DoubleStatistics*>(&to_msg);
  auto& from = static_cast<const DoubleStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.DoubleStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.minimum_ = from._impl_.minimum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.maximum_ = from._impl_.maximum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.sum_ = from._impl_.sum_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DoubleStatistics::CopyFrom(const DoubleStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.DoubleStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DoubleStatistics::InternalSwap(DoubleStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.sum_)
      + sizeof(DoubleStatistics::_impl_.sum_)
      - PROTOBUF_FIELD_OFFSET(DoubleStatistics, _impl_.minimum_)>(
          reinterpret_cast<char*>(&_impl_.minimum_),
          reinterpret_cast<char*>(&other->_impl_.minimum_));
}

::google::protobuf::Metadata DoubleStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StringStatistics::StringStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StringStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StringStatistics)
}
PROTOBUF_NDEBUG_INLINE StringStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StringStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        minimum_(arena, from.minimum_),
        maximum_(arena, from.maximum_),
        lower_bound_(arena, from.lower_bound_),
        upper_bound_(arena, from.upper_bound_) {}

StringStatistics::StringStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StringStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StringStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StringStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.sum_ = from._impl_.sum_;

  // @@protoc_insertion_point(copy_constructor:orc.proto.StringStatistics)
}
PROTOBUF_NDEBUG_INLINE StringStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        minimum_(arena),
        maximum_(arena),
        lower_bound_(arena),
        upper_bound_(arena) {}

inline void StringStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.sum_ = {};
}
StringStatistics::~StringStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.StringStatistics)
  SharedDtor(*this);
}
inline void StringStatistics::SharedDtor(MessageLite& self) {
  StringStatistics& this_ = static_cast<StringStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.minimum_.Destroy();
  this_._impl_.maximum_.Destroy();
  this_._impl_.lower_bound_.Destroy();
  this_._impl_.upper_bound_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StringStatistics_class_data_ =
        StringStatistics::InternalGenerateClassData_(StringStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StringStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StringStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StringStatistics_class_data_.tc_table);
  return StringStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StringStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StringStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StringStatistics_globals_));
  return StringStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StringStatistics::ParseTableT_
    StringStatistics::_table_ =
        StringStatistics::InternalGenerateParseTable_(StringStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StringStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StringStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.minimum_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.maximum_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.lower_bound_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.upper_bound_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.sum_ = ::int64_t{0};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StringStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StringStatistics& this_ = static_cast<const StringStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StringStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StringStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StringStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_minimum();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_maximum();
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional sint64 sum = 3;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        3, this_._internal_sum(), target);
  }

  // optional string lower_bound = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_lower_bound();
    target = stream->WriteStringMaybeAliased(4, _s, target);
  }

  // optional string upper_bound = 5;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_upper_bound();
    target = stream->WriteStringMaybeAliased(5, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StringStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StringStatistics::ByteSizeLong(const MessageLite& base) {
  const StringStatistics& this_ = static_cast<const StringStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StringStatistics::ByteSizeLong() const {
  const StringStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StringStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // optional string minimum = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_minimum());
    }
    // optional string maximum = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_maximum());
    }
    // optional string lower_bound = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_lower_bound());
    }
    // optional string upper_bound = 5;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_upper_bound());
    }
    // optional sint64 sum = 3;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_sum());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StringStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StringStatistics*>(&to_msg);
  auto& from = static_cast<const StringStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StringStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_minimum(from._internal_minimum());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_maximum(from._internal_maximum());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_lower_bound(from._internal_lower_bound());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_upper_bound(from._internal_upper_bound());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.sum_ = from._impl_.sum_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StringStatistics::CopyFrom(const StringStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StringStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StringStatistics::InternalSwap(StringStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.minimum_, &other->_impl_.minimum_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.maximum_, &other->_impl_.maximum_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.lower_bound_, &other->_impl_.lower_bound_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.upper_bound_, &other->_impl_.upper_bound_, arena);
  swap(_impl_.sum_, other->_impl_.sum_);
}

::google::protobuf::Metadata StringStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

BucketStatistics::BucketStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BucketStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.BucketStatistics)
}
PROTOBUF_NDEBUG_INLINE BucketStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::BucketStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        count_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::BucketStatistics,
              PROTOBUF_FIELD_OFFSET(::orc::proto::BucketStatistics, _impl_.count_)>()
          , from.count_
        }
        ,
        _count_cached_byte_size_{0} {}

BucketStatistics::BucketStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const BucketStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BucketStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  BucketStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.BucketStatistics)
}
PROTOBUF_NDEBUG_INLINE BucketStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        count_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BucketStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BucketStatistics, _impl_.count_)>()
         }
        ,
        _count_cached_byte_size_{0} {}

inline void BucketStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
BucketStatistics::~BucketStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.BucketStatistics)
  SharedDtor(*this);
}
inline void BucketStatistics::SharedDtor(MessageLite& self) {
  BucketStatistics& this_ = static_cast<BucketStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BucketStatistics_class_data_ =
        BucketStatistics::InternalGenerateClassData_(BucketStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BucketStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BucketStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BucketStatistics_class_data_.tc_table);
  return BucketStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BucketStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BucketStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&BucketStatistics_globals_));
  return BucketStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const BucketStatistics::ParseTableT_
    BucketStatistics::_table_ =
        BucketStatistics::InternalGenerateParseTable_(BucketStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void BucketStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.BucketStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.count_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BucketStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BucketStatistics& this_ = static_cast<const BucketStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BucketStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BucketStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.BucketStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated uint64 count = 1 [packed = true];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    {
      int byte_size = this_._impl_._count_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt64Packed(
            1, this_._internal_count(), byte_size, target);
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.BucketStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BucketStatistics::ByteSizeLong(const MessageLite& base) {
  const BucketStatistics& this_ = static_cast<const BucketStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BucketStatistics::ByteSizeLong() const {
  const BucketStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.BucketStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated uint64 count = 1 [packed = true];
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt64SizeWithPackedTagSize(
              this_._internal_count(), 1,
              this_._impl_._count_cached_byte_size_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BucketStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<BucketStatistics*>(&to_msg);
  auto& from = static_cast<const BucketStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.BucketStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_count()->MergeFrom(from._internal_count());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BucketStatistics::CopyFrom(const BucketStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.BucketStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BucketStatistics::InternalSwap(BucketStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.count_.InternalSwap(&other->_impl_.count_);
}

::google::protobuf::Metadata BucketStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DecimalStatistics::DecimalStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DecimalStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.DecimalStatistics)
}
PROTOBUF_NDEBUG_INLINE DecimalStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::DecimalStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        minimum_(arena, from.minimum_),
        maximum_(arena, from.maximum_),
        sum_(arena, from.sum_) {}

DecimalStatistics::DecimalStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const DecimalStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DecimalStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  DecimalStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.DecimalStatistics)
}
PROTOBUF_NDEBUG_INLINE DecimalStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        minimum_(arena),
        maximum_(arena),
        sum_(arena) {}

inline void DecimalStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
DecimalStatistics::~DecimalStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.DecimalStatistics)
  SharedDtor(*this);
}
inline void DecimalStatistics::SharedDtor(MessageLite& self) {
  DecimalStatistics& this_ = static_cast<DecimalStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.minimum_.Destroy();
  this_._impl_.maximum_.Destroy();
  this_._impl_.sum_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DecimalStatistics_class_data_ =
        DecimalStatistics::InternalGenerateClassData_(DecimalStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DecimalStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DecimalStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DecimalStatistics_class_data_.tc_table);
  return DecimalStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DecimalStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DecimalStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DecimalStatistics_globals_));
  return DecimalStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DecimalStatistics::ParseTableT_
    DecimalStatistics::_table_ =
        DecimalStatistics::InternalGenerateParseTable_(DecimalStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DecimalStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.DecimalStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.minimum_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.maximum_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.sum_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DecimalStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DecimalStatistics& this_ = static_cast<const DecimalStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DecimalStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DecimalStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.DecimalStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_minimum();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_maximum();
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional string sum = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_sum();
    target = stream->WriteStringMaybeAliased(3, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.DecimalStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DecimalStatistics::ByteSizeLong(const MessageLite& base) {
  const DecimalStatistics& this_ = static_cast<const DecimalStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DecimalStatistics::ByteSizeLong() const {
  const DecimalStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.DecimalStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional string minimum = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_minimum());
    }
    // optional string maximum = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_maximum());
    }
    // optional string sum = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_sum());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DecimalStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DecimalStatistics*>(&to_msg);
  auto& from = static_cast<const DecimalStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.DecimalStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_minimum(from._internal_minimum());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_maximum(from._internal_maximum());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_sum(from._internal_sum());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DecimalStatistics::CopyFrom(const DecimalStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.DecimalStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DecimalStatistics::InternalSwap(DecimalStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.minimum_, &other->_impl_.minimum_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.maximum_, &other->_impl_.maximum_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.sum_, &other->_impl_.sum_, arena);
}

::google::protobuf::Metadata DecimalStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DateStatistics::DateStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DateStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.DateStatistics)
}
DateStatistics::DateStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const DateStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DateStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE DateStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void DateStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, minimum_),
           0,
           offsetof(Impl_, maximum_) -
               offsetof(Impl_, minimum_) +
               sizeof(Impl_::maximum_));
}
DateStatistics::~DateStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.DateStatistics)
  SharedDtor(*this);
}
inline void DateStatistics::SharedDtor(MessageLite& self) {
  DateStatistics& this_ = static_cast<DateStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DateStatistics_class_data_ =
        DateStatistics::InternalGenerateClassData_(DateStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DateStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DateStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DateStatistics_class_data_.tc_table);
  return DateStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DateStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DateStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DateStatistics_globals_));
  return DateStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DateStatistics::ParseTableT_
    DateStatistics::_table_ =
        DateStatistics::InternalGenerateParseTable_(DateStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DateStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.DateStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    ::memset(&_impl_.minimum_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.maximum_) -
        reinterpret_cast<char*>(&_impl_.minimum_)) + sizeof(_impl_.maximum_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DateStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DateStatistics& this_ = static_cast<const DateStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DateStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DateStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.DateStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional sint32 minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt32ToArray(
        1, this_._internal_minimum(), target);
  }

  // optional sint32 maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt32ToArray(
        2, this_._internal_maximum(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.DateStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DateStatistics::ByteSizeLong(const MessageLite& base) {
  const DateStatistics& this_ = static_cast<const DateStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DateStatistics::ByteSizeLong() const {
  const DateStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.DateStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional sint32 minimum = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::SInt32SizePlusOne(
          this_._internal_minimum());
    }
    // optional sint32 maximum = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::SInt32SizePlusOne(
          this_._internal_maximum());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DateStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DateStatistics*>(&to_msg);
  auto& from = static_cast<const DateStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.DateStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.minimum_ = from._impl_.minimum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.maximum_ = from._impl_.maximum_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DateStatistics::CopyFrom(const DateStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.DateStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DateStatistics::InternalSwap(DateStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.maximum_)
      + sizeof(DateStatistics::_impl_.maximum_)
      - PROTOBUF_FIELD_OFFSET(DateStatistics, _impl_.minimum_)>(
          reinterpret_cast<char*>(&_impl_.minimum_),
          reinterpret_cast<char*>(&other->_impl_.minimum_));
}

::google::protobuf::Metadata DateStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

TimestampStatistics::TimestampStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, TimestampStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.TimestampStatistics)
}
TimestampStatistics::TimestampStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const TimestampStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, TimestampStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE TimestampStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void TimestampStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, minimum_),
           0,
           offsetof(Impl_, maximum_nanos_) -
               offsetof(Impl_, minimum_) +
               sizeof(Impl_::maximum_nanos_));
}
TimestampStatistics::~TimestampStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.TimestampStatistics)
  SharedDtor(*this);
}
inline void TimestampStatistics::SharedDtor(MessageLite& self) {
  TimestampStatistics& this_ = static_cast<TimestampStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull TimestampStatistics_class_data_ =
        TimestampStatistics::InternalGenerateClassData_(TimestampStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
TimestampStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&TimestampStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(TimestampStatistics_class_data_.tc_table);
  return TimestampStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
TimestampStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&TimestampStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&TimestampStatistics_globals_));
  return TimestampStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const TimestampStatistics::ParseTableT_
    TimestampStatistics::_table_ =
        TimestampStatistics::InternalGenerateParseTable_(TimestampStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void TimestampStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.TimestampStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    ::memset(&_impl_.minimum_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.maximum_nanos_) -
        reinterpret_cast<char*>(&_impl_.minimum_)) + sizeof(_impl_.maximum_nanos_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL TimestampStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const TimestampStatistics& this_ = static_cast<const TimestampStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL TimestampStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const TimestampStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.TimestampStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional sint64 minimum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        1, this_._internal_minimum(), target);
  }

  // optional sint64 maximum = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        2, this_._internal_maximum(), target);
  }

  // optional sint64 minimum_utc = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        3, this_._internal_minimum_utc(), target);
  }

  // optional sint64 maximum_utc = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        4, this_._internal_maximum_utc(), target);
  }

  // optional int32 minimum_nanos = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target =
        ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<5>(
            stream, this_._internal_minimum_nanos(), target);
  }

  // optional int32 maximum_nanos = 6;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target =
        ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<6>(
            stream, this_._internal_maximum_nanos(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.TimestampStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t TimestampStatistics::ByteSizeLong(const MessageLite& base) {
  const TimestampStatistics& this_ = static_cast<const TimestampStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t TimestampStatistics::ByteSizeLong() const {
  const TimestampStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.TimestampStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // optional sint64 minimum = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_minimum());
    }
    // optional sint64 maximum = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_maximum());
    }
    // optional sint64 minimum_utc = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_minimum_utc());
    }
    // optional sint64 maximum_utc = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_maximum_utc());
    }
    // optional int32 minimum_nanos = 5;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
          this_._internal_minimum_nanos());
    }
    // optional int32 maximum_nanos = 6;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
          this_._internal_maximum_nanos());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void TimestampStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<TimestampStatistics*>(&to_msg);
  auto& from = static_cast<const TimestampStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.TimestampStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.minimum_ = from._impl_.minimum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.maximum_ = from._impl_.maximum_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.minimum_utc_ = from._impl_.minimum_utc_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.maximum_utc_ = from._impl_.maximum_utc_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.minimum_nanos_ = from._impl_.minimum_nanos_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _this->_impl_.maximum_nanos_ = from._impl_.maximum_nanos_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void TimestampStatistics::CopyFrom(const TimestampStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.TimestampStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void TimestampStatistics::InternalSwap(TimestampStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.maximum_nanos_)
      + sizeof(TimestampStatistics::_impl_.maximum_nanos_)
      - PROTOBUF_FIELD_OFFSET(TimestampStatistics, _impl_.minimum_)>(
          reinterpret_cast<char*>(&_impl_.minimum_),
          reinterpret_cast<char*>(&other->_impl_.minimum_));
}

::google::protobuf::Metadata TimestampStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

BinaryStatistics::BinaryStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BinaryStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.BinaryStatistics)
}
BinaryStatistics::BinaryStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const BinaryStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BinaryStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE BinaryStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void BinaryStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.sum_ = {};
}
BinaryStatistics::~BinaryStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.BinaryStatistics)
  SharedDtor(*this);
}
inline void BinaryStatistics::SharedDtor(MessageLite& self) {
  BinaryStatistics& this_ = static_cast<BinaryStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BinaryStatistics_class_data_ =
        BinaryStatistics::InternalGenerateClassData_(BinaryStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BinaryStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BinaryStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BinaryStatistics_class_data_.tc_table);
  return BinaryStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BinaryStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BinaryStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&BinaryStatistics_globals_));
  return BinaryStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const BinaryStatistics::ParseTableT_
    BinaryStatistics::_table_ =
        BinaryStatistics::InternalGenerateParseTable_(BinaryStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void BinaryStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.BinaryStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.sum_ = ::int64_t{0};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BinaryStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BinaryStatistics& this_ = static_cast<const BinaryStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BinaryStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BinaryStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.BinaryStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional sint64 sum = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteSInt64ToArray(
        1, this_._internal_sum(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.BinaryStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BinaryStatistics::ByteSizeLong(const MessageLite& base) {
  const BinaryStatistics& this_ = static_cast<const BinaryStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BinaryStatistics::ByteSizeLong() const {
  const BinaryStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.BinaryStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

   {
    // optional sint64 sum = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::SInt64SizePlusOne(
          this_._internal_sum());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BinaryStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<BinaryStatistics*>(&to_msg);
  auto& from = static_cast<const BinaryStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.BinaryStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_impl_.sum_ = from._impl_.sum_;
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BinaryStatistics::CopyFrom(const BinaryStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.BinaryStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BinaryStatistics::InternalSwap(BinaryStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.sum_, other->_impl_.sum_);
}

::google::protobuf::Metadata BinaryStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CollectionStatistics::CollectionStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CollectionStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.CollectionStatistics)
}
CollectionStatistics::CollectionStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CollectionStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CollectionStatistics_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE CollectionStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void CollectionStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, min_children_),
           0,
           offsetof(Impl_, total_children_) -
               offsetof(Impl_, min_children_) +
               sizeof(Impl_::total_children_));
}
CollectionStatistics::~CollectionStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.CollectionStatistics)
  SharedDtor(*this);
}
inline void CollectionStatistics::SharedDtor(MessageLite& self) {
  CollectionStatistics& this_ = static_cast<CollectionStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CollectionStatistics_class_data_ =
        CollectionStatistics::InternalGenerateClassData_(CollectionStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CollectionStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CollectionStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CollectionStatistics_class_data_.tc_table);
  return CollectionStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CollectionStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CollectionStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CollectionStatistics_globals_));
  return CollectionStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CollectionStatistics::ParseTableT_
    CollectionStatistics::_table_ =
        CollectionStatistics::InternalGenerateParseTable_(CollectionStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CollectionStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.CollectionStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    ::memset(&_impl_.min_children_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.total_children_) -
        reinterpret_cast<char*>(&_impl_.min_children_)) + sizeof(_impl_.total_children_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CollectionStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CollectionStatistics& this_ = static_cast<const CollectionStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CollectionStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CollectionStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.CollectionStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint64 min_children = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this_._internal_min_children(), target);
  }

  // optional uint64 max_children = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this_._internal_max_children(), target);
  }

  // optional uint64 total_children = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this_._internal_total_children(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.CollectionStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CollectionStatistics::ByteSizeLong(const MessageLite& base) {
  const CollectionStatistics& this_ = static_cast<const CollectionStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CollectionStatistics::ByteSizeLong() const {
  const CollectionStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.CollectionStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional uint64 min_children = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_min_children());
    }
    // optional uint64 max_children = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_max_children());
    }
    // optional uint64 total_children = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_total_children());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CollectionStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CollectionStatistics*>(&to_msg);
  auto& from = static_cast<const CollectionStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.CollectionStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.min_children_ = from._impl_.min_children_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.max_children_ = from._impl_.max_children_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.total_children_ = from._impl_.total_children_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CollectionStatistics::CopyFrom(const CollectionStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.CollectionStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CollectionStatistics::InternalSwap(CollectionStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.total_children_)
      + sizeof(CollectionStatistics::_impl_.total_children_)
      - PROTOBUF_FIELD_OFFSET(CollectionStatistics, _impl_.min_children_)>(
          reinterpret_cast<char*>(&_impl_.min_children_),
          reinterpret_cast<char*>(&other->_impl_.min_children_));
}

::google::protobuf::Metadata CollectionStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

BoundingBox::BoundingBox(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BoundingBox_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.BoundingBox)
}
BoundingBox::BoundingBox(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const BoundingBox& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BoundingBox_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE BoundingBox::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void BoundingBox::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, xmin_),
           0,
           offsetof(Impl_, mmax_) -
               offsetof(Impl_, xmin_) +
               sizeof(Impl_::mmax_));
}
BoundingBox::~BoundingBox() {
  // @@protoc_insertion_point(destructor:orc.proto.BoundingBox)
  SharedDtor(*this);
}
inline void BoundingBox::SharedDtor(MessageLite& self) {
  BoundingBox& this_ = static_cast<BoundingBox&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BoundingBox_class_data_ =
        BoundingBox::InternalGenerateClassData_(BoundingBox_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BoundingBox::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BoundingBox_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BoundingBox_class_data_.tc_table);
  return BoundingBox_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BoundingBox::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BoundingBox_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&BoundingBox_globals_));
  return BoundingBox_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const BoundingBox::ParseTableT_
    BoundingBox::_table_ =
        BoundingBox::InternalGenerateParseTable_(BoundingBox_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void BoundingBox::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.BoundingBox)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    ::memset(&_impl_.xmin_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.mmax_) -
        reinterpret_cast<char*>(&_impl_.xmin_)) + sizeof(_impl_.mmax_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BoundingBox::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BoundingBox& this_ = static_cast<const BoundingBox&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BoundingBox::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BoundingBox& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.BoundingBox)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional double xmin = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        1, this_._internal_xmin(), target);
  }

  // optional double xmax = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        2, this_._internal_xmax(), target);
  }

  // optional double ymin = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        3, this_._internal_ymin(), target);
  }

  // optional double ymax = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        4, this_._internal_ymax(), target);
  }

  // optional double zmin = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        5, this_._internal_zmin(), target);
  }

  // optional double zmax = 6;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        6, this_._internal_zmax(), target);
  }

  // optional double mmin = 7;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        7, this_._internal_mmin(), target);
  }

  // optional double mmax = 8;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteDoubleToArray(
        8, this_._internal_mmax(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.BoundingBox)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BoundingBox::ByteSizeLong(const MessageLite& base) {
  const BoundingBox& this_ = static_cast<const BoundingBox&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BoundingBox::ByteSizeLong() const {
  const BoundingBox& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.BoundingBox)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  total_size += ::absl::popcount(0x000000ffU & cached_has_bits) * 9;
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BoundingBox::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<BoundingBox*>(&to_msg);
  auto& from = static_cast<const BoundingBox&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.BoundingBox)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.xmin_ = from._impl_.xmin_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.xmax_ = from._impl_.xmax_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.ymin_ = from._impl_.ymin_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.ymax_ = from._impl_.ymax_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.zmin_ = from._impl_.zmin_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _this->_impl_.zmax_ = from._impl_.zmax_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.mmin_ = from._impl_.mmin_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _this->_impl_.mmax_ = from._impl_.mmax_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BoundingBox::CopyFrom(const BoundingBox& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.BoundingBox)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BoundingBox::InternalSwap(BoundingBox* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.mmax_)
      + sizeof(BoundingBox::_impl_.mmax_)
      - PROTOBUF_FIELD_OFFSET(BoundingBox, _impl_.xmin_)>(
          reinterpret_cast<char*>(&_impl_.xmin_),
          reinterpret_cast<char*>(&other->_impl_.xmin_));
}

::google::protobuf::Metadata BoundingBox::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

GeospatialStatistics::GeospatialStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, GeospatialStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.GeospatialStatistics)
}
PROTOBUF_NDEBUG_INLINE GeospatialStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::GeospatialStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        geospatial_types_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::GeospatialStatistics,
              PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_.geospatial_types_)>()
          , from.geospatial_types_
        }
     {}

GeospatialStatistics::GeospatialStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const GeospatialStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, GeospatialStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  GeospatialStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.bbox_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bbox_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:orc.proto.GeospatialStatistics)
}
PROTOBUF_NDEBUG_INLINE GeospatialStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        geospatial_types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::GeospatialStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::GeospatialStatistics, _impl_.geospatial_types_)>()
         }
     {}

inline void GeospatialStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.bbox_ = {};
}
GeospatialStatistics::~GeospatialStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.GeospatialStatistics)
  SharedDtor(*this);
}
inline void GeospatialStatistics::SharedDtor(MessageLite& self) {
  GeospatialStatistics& this_ = static_cast<GeospatialStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.bbox_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull GeospatialStatistics_class_data_ =
        GeospatialStatistics::InternalGenerateClassData_(GeospatialStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
GeospatialStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&GeospatialStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(GeospatialStatistics_class_data_.tc_table);
  return GeospatialStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
GeospatialStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&GeospatialStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&GeospatialStatistics_globals_));
  return GeospatialStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const GeospatialStatistics::ParseTableT_
    GeospatialStatistics::_table_ =
        GeospatialStatistics::InternalGenerateParseTable_(GeospatialStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void GeospatialStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.GeospatialStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.geospatial_types_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.bbox_ != nullptr);
      _impl_.bbox_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL GeospatialStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const GeospatialStatistics& this_ = static_cast<const GeospatialStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL GeospatialStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const GeospatialStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.GeospatialStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional .orc.proto.BoundingBox bbox = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.bbox_, this_._impl_.bbox_->GetCachedSize(), target,
        stream);
  }

  // repeated int32 geospatial_types = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_geospatial_types_size(); i < n; ++i) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteInt32ToArray(
          2, this_._internal_geospatial_types().Get(i), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.GeospatialStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t GeospatialStatistics::ByteSizeLong(const MessageLite& base) {
  const GeospatialStatistics& this_ = static_cast<const GeospatialStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t GeospatialStatistics::ByteSizeLong() const {
  const GeospatialStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.GeospatialStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated int32 geospatial_types = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ::size_t data_size = ::_pbi::WireFormatLite::Int32Size(
          this_._internal_geospatial_types());
      ::size_t tag_size = ::size_t{1} *
          ::_pbi::FromIntSize(this_._internal_geospatial_types_size());
      total_size += tag_size + data_size;
    }
    // optional .orc.proto.BoundingBox bbox = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.bbox_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void GeospatialStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<GeospatialStatistics*>(&to_msg);
  auto& from = static_cast<const GeospatialStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.GeospatialStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_geospatial_types()->MergeFrom(from._internal_geospatial_types());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.bbox_ != nullptr);
      if (_this->_impl_.bbox_ == nullptr) {
        _this->_impl_.bbox_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bbox_);
      } else {
        _this->_impl_.bbox_->MergeFrom(*from._impl_.bbox_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void GeospatialStatistics::CopyFrom(const GeospatialStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.GeospatialStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void GeospatialStatistics::InternalSwap(GeospatialStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.geospatial_types_.InternalSwap(&other->_impl_.geospatial_types_);
  swap(_impl_.bbox_, other->_impl_.bbox_);
}

::google::protobuf::Metadata GeospatialStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ColumnStatistics::ColumnStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.ColumnStatistics)
}
PROTOBUF_NDEBUG_INLINE ColumnStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::ColumnStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

ColumnStatistics::ColumnStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ColumnStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ColumnStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.int_statistics_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.int_statistics_)
                : nullptr;
  _impl_.double_statistics_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.double_statistics_)
                : nullptr;
  _impl_.string_statistics_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.string_statistics_)
                : nullptr;
  _impl_.bucket_statistics_ = (CheckHasBit(cached_has_bits, 0x00000008U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bucket_statistics_)
                : nullptr;
  _impl_.decimal_statistics_ = (CheckHasBit(cached_has_bits, 0x00000010U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.decimal_statistics_)
                : nullptr;
  _impl_.date_statistics_ = (CheckHasBit(cached_has_bits, 0x00000020U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.date_statistics_)
                : nullptr;
  _impl_.binary_statistics_ = (CheckHasBit(cached_has_bits, 0x00000040U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.binary_statistics_)
                : nullptr;
  _impl_.timestamp_statistics_ = (CheckHasBit(cached_has_bits, 0x00000080U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_statistics_)
                : nullptr;
  _impl_.collection_statistics_ = (CheckHasBit(cached_has_bits, 0x00000100U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.collection_statistics_)
                : nullptr;
  _impl_.geospatial_statistics_ = (CheckHasBit(cached_has_bits, 0x00000200U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.geospatial_statistics_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, number_of_values_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, number_of_values_),
           offsetof(Impl_, has_null_) -
               offsetof(Impl_, number_of_values_) +
               sizeof(Impl_::has_null_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.ColumnStatistics)
}
PROTOBUF_NDEBUG_INLINE ColumnStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void ColumnStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, int_statistics_),
           0,
           offsetof(Impl_, has_null_) -
               offsetof(Impl_, int_statistics_) +
               sizeof(Impl_::has_null_));
}
ColumnStatistics::~ColumnStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.ColumnStatistics)
  SharedDtor(*this);
}
inline void ColumnStatistics::SharedDtor(MessageLite& self) {
  ColumnStatistics& this_ = static_cast<ColumnStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.int_statistics_;
  delete this_._impl_.double_statistics_;
  delete this_._impl_.string_statistics_;
  delete this_._impl_.bucket_statistics_;
  delete this_._impl_.decimal_statistics_;
  delete this_._impl_.date_statistics_;
  delete this_._impl_.binary_statistics_;
  delete this_._impl_.timestamp_statistics_;
  delete this_._impl_.collection_statistics_;
  delete this_._impl_.geospatial_statistics_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ColumnStatistics_class_data_ =
        ColumnStatistics::InternalGenerateClassData_(ColumnStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ColumnStatistics_class_data_.tc_table);
  return ColumnStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ColumnStatistics_globals_));
  return ColumnStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ColumnStatistics::ParseTableT_
    ColumnStatistics::_table_ =
        ColumnStatistics::InternalGenerateParseTable_(ColumnStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ColumnStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.ColumnStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(_impl_.int_statistics_ != nullptr);
      _impl_.int_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.double_statistics_ != nullptr);
      _impl_.double_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.string_statistics_ != nullptr);
      _impl_.string_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(_impl_.bucket_statistics_ != nullptr);
      _impl_.bucket_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(_impl_.decimal_statistics_ != nullptr);
      _impl_.decimal_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      ABSL_DCHECK(_impl_.date_statistics_ != nullptr);
      _impl_.date_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      ABSL_DCHECK(_impl_.binary_statistics_ != nullptr);
      _impl_.binary_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      ABSL_DCHECK(_impl_.timestamp_statistics_ != nullptr);
      _impl_.timestamp_statistics_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      ABSL_DCHECK(_impl_.collection_statistics_ != nullptr);
      _impl_.collection_statistics_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      ABSL_DCHECK(_impl_.geospatial_statistics_ != nullptr);
      _impl_.geospatial_statistics_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00001c00U)) {
    ::memset(&_impl_.number_of_values_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.has_null_) -
        reinterpret_cast<char*>(&_impl_.number_of_values_)) + sizeof(_impl_.has_null_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ColumnStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ColumnStatistics& this_ = static_cast<const ColumnStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ColumnStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ColumnStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.ColumnStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint64 number_of_values = 1;
  if (CheckHasBit(cached_has_bits, 0x00000400U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this_._internal_number_of_values(), target);
  }

  // optional .orc.proto.IntegerStatistics int_statistics = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.int_statistics_, this_._impl_.int_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.DoubleStatistics double_statistics = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.double_statistics_, this_._impl_.double_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.StringStatistics string_statistics = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        4, *this_._impl_.string_statistics_, this_._impl_.string_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.BucketStatistics bucket_statistics = 5;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        5, *this_._impl_.bucket_statistics_, this_._impl_.bucket_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.DecimalStatistics decimal_statistics = 6;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        6, *this_._impl_.decimal_statistics_, this_._impl_.decimal_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.DateStatistics date_statistics = 7;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        7, *this_._impl_.date_statistics_, this_._impl_.date_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.BinaryStatistics binary_statistics = 8;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        8, *this_._impl_.binary_statistics_, this_._impl_.binary_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.TimestampStatistics timestamp_statistics = 9;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        9, *this_._impl_.timestamp_statistics_, this_._impl_.timestamp_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional bool has_null = 10;
  if (CheckHasBit(cached_has_bits, 0x00001000U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        10, this_._internal_has_null(), target);
  }

  // optional uint64 bytes_on_disk = 11;
  if (CheckHasBit(cached_has_bits, 0x00000800U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        11, this_._internal_bytes_on_disk(), target);
  }

  // optional .orc.proto.CollectionStatistics collection_statistics = 12;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        12, *this_._impl_.collection_statistics_, this_._impl_.collection_statistics_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.GeospatialStatistics geospatial_statistics = 13;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        13, *this_._impl_.geospatial_statistics_, this_._impl_.geospatial_statistics_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.ColumnStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ColumnStatistics::ByteSizeLong(const MessageLite& base) {
  const ColumnStatistics& this_ = static_cast<const ColumnStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ColumnStatistics::ByteSizeLong() const {
  const ColumnStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.ColumnStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  total_size += static_cast<bool>(0x00001000U & cached_has_bits) * 2;
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // optional .orc.proto.IntegerStatistics int_statistics = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.int_statistics_);
    }
    // optional .orc.proto.DoubleStatistics double_statistics = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.double_statistics_);
    }
    // optional .orc.proto.StringStatistics string_statistics = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.string_statistics_);
    }
    // optional .orc.proto.BucketStatistics bucket_statistics = 5;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.bucket_statistics_);
    }
    // optional .orc.proto.DecimalStatistics decimal_statistics = 6;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.decimal_statistics_);
    }
    // optional .orc.proto.DateStatistics date_statistics = 7;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.date_statistics_);
    }
    // optional .orc.proto.BinaryStatistics binary_statistics = 8;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.binary_statistics_);
    }
    // optional .orc.proto.TimestampStatistics timestamp_statistics = 9;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.timestamp_statistics_);
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000f00U)) {
    // optional .orc.proto.CollectionStatistics collection_statistics = 12;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.collection_statistics_);
    }
    // optional .orc.proto.GeospatialStatistics geospatial_statistics = 13;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.geospatial_statistics_);
    }
    // optional uint64 number_of_values = 1;
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_number_of_values());
    }
    // optional uint64 bytes_on_disk = 11;
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_bytes_on_disk());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ColumnStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ColumnStatistics*>(&to_msg);
  auto& from = static_cast<const ColumnStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.ColumnStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.int_statistics_ != nullptr);
      if (_this->_impl_.int_statistics_ == nullptr) {
        _this->_impl_.int_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.int_statistics_);
      } else {
        _this->_impl_.int_statistics_->MergeFrom(*from._impl_.int_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.double_statistics_ != nullptr);
      if (_this->_impl_.double_statistics_ == nullptr) {
        _this->_impl_.double_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.double_statistics_);
      } else {
        _this->_impl_.double_statistics_->MergeFrom(*from._impl_.double_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.string_statistics_ != nullptr);
      if (_this->_impl_.string_statistics_ == nullptr) {
        _this->_impl_.string_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.string_statistics_);
      } else {
        _this->_impl_.string_statistics_->MergeFrom(*from._impl_.string_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(from._impl_.bucket_statistics_ != nullptr);
      if (_this->_impl_.bucket_statistics_ == nullptr) {
        _this->_impl_.bucket_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.bucket_statistics_);
      } else {
        _this->_impl_.bucket_statistics_->MergeFrom(*from._impl_.bucket_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(from._impl_.decimal_statistics_ != nullptr);
      if (_this->_impl_.decimal_statistics_ == nullptr) {
        _this->_impl_.decimal_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.decimal_statistics_);
      } else {
        _this->_impl_.decimal_statistics_->MergeFrom(*from._impl_.decimal_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      ABSL_DCHECK(from._impl_.date_statistics_ != nullptr);
      if (_this->_impl_.date_statistics_ == nullptr) {
        _this->_impl_.date_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.date_statistics_);
      } else {
        _this->_impl_.date_statistics_->MergeFrom(*from._impl_.date_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      ABSL_DCHECK(from._impl_.binary_statistics_ != nullptr);
      if (_this->_impl_.binary_statistics_ == nullptr) {
        _this->_impl_.binary_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.binary_statistics_);
      } else {
        _this->_impl_.binary_statistics_->MergeFrom(*from._impl_.binary_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      ABSL_DCHECK(from._impl_.timestamp_statistics_ != nullptr);
      if (_this->_impl_.timestamp_statistics_ == nullptr) {
        _this->_impl_.timestamp_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_statistics_);
      } else {
        _this->_impl_.timestamp_statistics_->MergeFrom(*from._impl_.timestamp_statistics_);
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00001f00U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      ABSL_DCHECK(from._impl_.collection_statistics_ != nullptr);
      if (_this->_impl_.collection_statistics_ == nullptr) {
        _this->_impl_.collection_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.collection_statistics_);
      } else {
        _this->_impl_.collection_statistics_->MergeFrom(*from._impl_.collection_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      ABSL_DCHECK(from._impl_.geospatial_statistics_ != nullptr);
      if (_this->_impl_.geospatial_statistics_ == nullptr) {
        _this->_impl_.geospatial_statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.geospatial_statistics_);
      } else {
        _this->_impl_.geospatial_statistics_->MergeFrom(*from._impl_.geospatial_statistics_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      _this->_impl_.number_of_values_ = from._impl_.number_of_values_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      _this->_impl_.bytes_on_disk_ = from._impl_.bytes_on_disk_;
    }
    if (CheckHasBit(cached_has_bits, 0x00001000U)) {
      _this->_impl_.has_null_ = from._impl_.has_null_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ColumnStatistics::CopyFrom(const ColumnStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.ColumnStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ColumnStatistics::InternalSwap(ColumnStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.has_null_)
      + sizeof(ColumnStatistics::_impl_.has_null_)
      - PROTOBUF_FIELD_OFFSET(ColumnStatistics, _impl_.int_statistics_)>(
          reinterpret_cast<char*>(&_impl_.int_statistics_),
          reinterpret_cast<char*>(&other->_impl_.int_statistics_));
}

::google::protobuf::Metadata ColumnStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

RowIndexEntry::RowIndexEntry(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RowIndexEntry_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.RowIndexEntry)
}
PROTOBUF_NDEBUG_INLINE RowIndexEntry::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::RowIndexEntry& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        positions_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::RowIndexEntry,
              PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_.positions_)>()
          , from.positions_
        }
        ,
        _positions_cached_byte_size_{0} {}

RowIndexEntry::RowIndexEntry(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const RowIndexEntry& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RowIndexEntry_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  RowIndexEntry* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.statistics_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.statistics_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:orc.proto.RowIndexEntry)
}
PROTOBUF_NDEBUG_INLINE RowIndexEntry::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        positions_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::RowIndexEntry,
            PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndexEntry, _impl_.positions_)>()
         }
        ,
        _positions_cached_byte_size_{0} {}

inline void RowIndexEntry::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.statistics_ = {};
}
RowIndexEntry::~RowIndexEntry() {
  // @@protoc_insertion_point(destructor:orc.proto.RowIndexEntry)
  SharedDtor(*this);
}
inline void RowIndexEntry::SharedDtor(MessageLite& self) {
  RowIndexEntry& this_ = static_cast<RowIndexEntry&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.statistics_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull RowIndexEntry_class_data_ =
        RowIndexEntry::InternalGenerateClassData_(RowIndexEntry_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
RowIndexEntry::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&RowIndexEntry_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(RowIndexEntry_class_data_.tc_table);
  return RowIndexEntry_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
RowIndexEntry::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&RowIndexEntry_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&RowIndexEntry_globals_));
  return RowIndexEntry_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const RowIndexEntry::ParseTableT_
    RowIndexEntry::_table_ =
        RowIndexEntry::InternalGenerateParseTable_(RowIndexEntry_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void RowIndexEntry::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.RowIndexEntry)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.positions_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.statistics_ != nullptr);
      _impl_.statistics_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL RowIndexEntry::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const RowIndexEntry& this_ = static_cast<const RowIndexEntry&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL RowIndexEntry::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const RowIndexEntry& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.RowIndexEntry)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated uint64 positions = 1 [packed = true];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    {
      int byte_size = this_._impl_._positions_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt64Packed(
            1, this_._internal_positions(), byte_size, target);
      }
    }
  }

  // optional .orc.proto.ColumnStatistics statistics = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.statistics_, this_._impl_.statistics_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.RowIndexEntry)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t RowIndexEntry::ByteSizeLong(const MessageLite& base) {
  const RowIndexEntry& this_ = static_cast<const RowIndexEntry&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t RowIndexEntry::ByteSizeLong() const {
  const RowIndexEntry& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.RowIndexEntry)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated uint64 positions = 1 [packed = true];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt64SizeWithPackedTagSize(
              this_._internal_positions(), 1,
              this_._impl_._positions_cached_byte_size_);
    }
    // optional .orc.proto.ColumnStatistics statistics = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.statistics_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void RowIndexEntry::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<RowIndexEntry*>(&to_msg);
  auto& from = static_cast<const RowIndexEntry&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.RowIndexEntry)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_positions()->MergeFrom(from._internal_positions());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.statistics_ != nullptr);
      if (_this->_impl_.statistics_ == nullptr) {
        _this->_impl_.statistics_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.statistics_);
      } else {
        _this->_impl_.statistics_->MergeFrom(*from._impl_.statistics_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void RowIndexEntry::CopyFrom(const RowIndexEntry& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.RowIndexEntry)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void RowIndexEntry::InternalSwap(RowIndexEntry* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.positions_.InternalSwap(&other->_impl_.positions_);
  swap(_impl_.statistics_, other->_impl_.statistics_);
}

::google::protobuf::Metadata RowIndexEntry::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

RowIndex::RowIndex(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RowIndex_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.RowIndex)
}
PROTOBUF_NDEBUG_INLINE RowIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::RowIndex& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        entry_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::RowIndex,
              PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndex, _impl_.entry_)>()
          , from.entry_
        }
     {}

RowIndex::RowIndex(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const RowIndex& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, RowIndex_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  RowIndex* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.RowIndex)
}
PROTOBUF_NDEBUG_INLINE RowIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        entry_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::RowIndex,
            PROTOBUF_FIELD_OFFSET(::orc::proto::RowIndex, _impl_.entry_)>()
         }
     {}

inline void RowIndex::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
RowIndex::~RowIndex() {
  // @@protoc_insertion_point(destructor:orc.proto.RowIndex)
  SharedDtor(*this);
}
inline void RowIndex::SharedDtor(MessageLite& self) {
  RowIndex& this_ = static_cast<RowIndex&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull RowIndex_class_data_ =
        RowIndex::InternalGenerateClassData_(RowIndex_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
RowIndex::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&RowIndex_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(RowIndex_class_data_.tc_table);
  return RowIndex_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
RowIndex::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&RowIndex_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&RowIndex_globals_));
  return RowIndex_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const RowIndex::ParseTableT_
    RowIndex::_table_ =
        RowIndex::InternalGenerateParseTable_(RowIndex_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void RowIndex::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.RowIndex)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.entry_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL RowIndex::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const RowIndex& this_ = static_cast<const RowIndex&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL RowIndex::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const RowIndex& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.RowIndex)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.RowIndexEntry entry = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_entry_size());
         i < n; i++) {
      const auto& repfield = this_._internal_entry().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.RowIndex)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t RowIndex::ByteSizeLong(const MessageLite& base) {
  const RowIndex& this_ = static_cast<const RowIndex&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t RowIndex::ByteSizeLong() const {
  const RowIndex& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.RowIndex)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.RowIndexEntry entry = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_entry_size();
      for (const auto& msg : this_._internal_entry()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void RowIndex::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<RowIndex*>(&to_msg);
  auto& from = static_cast<const RowIndex&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.RowIndex)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_entry()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_entry());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void RowIndex::CopyFrom(const RowIndex& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.RowIndex)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void RowIndex::InternalSwap(RowIndex* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.entry_.InternalSwap(&other->_impl_.entry_);
}

::google::protobuf::Metadata RowIndex::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

BloomFilter::BloomFilter(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BloomFilter_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.BloomFilter)
}
PROTOBUF_NDEBUG_INLINE BloomFilter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::BloomFilter& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        bitset_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::BloomFilter,
              PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.bitset_)>()
          , from.bitset_
        }
        ,
        utf8bitset_(arena, from.utf8bitset_) {}

BloomFilter::BloomFilter(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const BloomFilter& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BloomFilter_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  BloomFilter* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.num_hash_functions_ = from._impl_.num_hash_functions_;

  // @@protoc_insertion_point(copy_constructor:orc.proto.BloomFilter)
}
PROTOBUF_NDEBUG_INLINE BloomFilter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        bitset_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BloomFilter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilter, _impl_.bitset_)>()
         }
        ,
        utf8bitset_(arena) {}

inline void BloomFilter::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.num_hash_functions_ = {};
}
BloomFilter::~BloomFilter() {
  // @@protoc_insertion_point(destructor:orc.proto.BloomFilter)
  SharedDtor(*this);
}
inline void BloomFilter::SharedDtor(MessageLite& self) {
  BloomFilter& this_ = static_cast<BloomFilter&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.utf8bitset_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BloomFilter_class_data_ =
        BloomFilter::InternalGenerateClassData_(BloomFilter_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BloomFilter::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BloomFilter_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BloomFilter_class_data_.tc_table);
  return BloomFilter_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BloomFilter::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BloomFilter_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&BloomFilter_globals_));
  return BloomFilter_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const BloomFilter::ParseTableT_
    BloomFilter::_table_ =
        BloomFilter::InternalGenerateParseTable_(BloomFilter_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void BloomFilter::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.BloomFilter)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.bitset_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.utf8bitset_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.num_hash_functions_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BloomFilter::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BloomFilter& this_ = static_cast<const BloomFilter&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BloomFilter::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BloomFilter& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.BloomFilter)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint32 num_hash_functions = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this_._internal_num_hash_functions(), target);
  }

  // repeated fixed64 bitset = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_bitset_size(); i < n; ++i) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteFixed64ToArray(
          2, this_._internal_bitset().Get(i), target);
    }
  }

  // optional bytes utf8bitset = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_utf8bitset();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.BloomFilter)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BloomFilter::ByteSizeLong(const MessageLite& base) {
  const BloomFilter& this_ = static_cast<const BloomFilter&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BloomFilter::ByteSizeLong() const {
  const BloomFilter& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.BloomFilter)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated fixed64 bitset = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ::size_t data_size = ::size_t{8} *
          ::_pbi::FromIntSize(this_._internal_bitset_size());
      ::size_t tag_size = ::size_t{1} *
          ::_pbi::FromIntSize(this_._internal_bitset_size());
      total_size += tag_size + data_size;
    }
    // optional bytes utf8bitset = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_utf8bitset());
    }
    // optional uint32 num_hash_functions = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_num_hash_functions());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BloomFilter::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<BloomFilter*>(&to_msg);
  auto& from = static_cast<const BloomFilter&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.BloomFilter)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_bitset()->MergeFrom(from._internal_bitset());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_utf8bitset(from._internal_utf8bitset());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.num_hash_functions_ = from._impl_.num_hash_functions_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BloomFilter::CopyFrom(const BloomFilter& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.BloomFilter)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BloomFilter::InternalSwap(BloomFilter* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.bitset_.InternalSwap(&other->_impl_.bitset_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.utf8bitset_, &other->_impl_.utf8bitset_, arena);
  swap(_impl_.num_hash_functions_, other->_impl_.num_hash_functions_);
}

::google::protobuf::Metadata BloomFilter::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

BloomFilterIndex::BloomFilterIndex(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BloomFilterIndex_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.BloomFilterIndex)
}
PROTOBUF_NDEBUG_INLINE BloomFilterIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::BloomFilterIndex& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        bloom_filter_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::BloomFilterIndex,
              PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilterIndex, _impl_.bloom_filter_)>()
          , from.bloom_filter_
        }
     {}

BloomFilterIndex::BloomFilterIndex(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const BloomFilterIndex& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, BloomFilterIndex_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  BloomFilterIndex* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.BloomFilterIndex)
}
PROTOBUF_NDEBUG_INLINE BloomFilterIndex::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        bloom_filter_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::BloomFilterIndex,
            PROTOBUF_FIELD_OFFSET(::orc::proto::BloomFilterIndex, _impl_.bloom_filter_)>()
         }
     {}

inline void BloomFilterIndex::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
BloomFilterIndex::~BloomFilterIndex() {
  // @@protoc_insertion_point(destructor:orc.proto.BloomFilterIndex)
  SharedDtor(*this);
}
inline void BloomFilterIndex::SharedDtor(MessageLite& self) {
  BloomFilterIndex& this_ = static_cast<BloomFilterIndex&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull BloomFilterIndex_class_data_ =
        BloomFilterIndex::InternalGenerateClassData_(BloomFilterIndex_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BloomFilterIndex::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BloomFilterIndex_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(BloomFilterIndex_class_data_.tc_table);
  return BloomFilterIndex_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
BloomFilterIndex::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&BloomFilterIndex_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&BloomFilterIndex_globals_));
  return BloomFilterIndex_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const BloomFilterIndex::ParseTableT_
    BloomFilterIndex::_table_ =
        BloomFilterIndex::InternalGenerateParseTable_(BloomFilterIndex_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void BloomFilterIndex::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.BloomFilterIndex)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.bloom_filter_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL BloomFilterIndex::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const BloomFilterIndex& this_ = static_cast<const BloomFilterIndex&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL BloomFilterIndex::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const BloomFilterIndex& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.BloomFilterIndex)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.BloomFilter bloom_filter = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_bloom_filter_size());
         i < n; i++) {
      const auto& repfield = this_._internal_bloom_filter().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.BloomFilterIndex)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t BloomFilterIndex::ByteSizeLong(const MessageLite& base) {
  const BloomFilterIndex& this_ = static_cast<const BloomFilterIndex&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t BloomFilterIndex::ByteSizeLong() const {
  const BloomFilterIndex& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.BloomFilterIndex)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.BloomFilter bloom_filter = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_bloom_filter_size();
      for (const auto& msg : this_._internal_bloom_filter()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void BloomFilterIndex::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<BloomFilterIndex*>(&to_msg);
  auto& from = static_cast<const BloomFilterIndex&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.BloomFilterIndex)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_bloom_filter()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_bloom_filter());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void BloomFilterIndex::CopyFrom(const BloomFilterIndex& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.BloomFilterIndex)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void BloomFilterIndex::InternalSwap(BloomFilterIndex* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.bloom_filter_.InternalSwap(&other->_impl_.bloom_filter_);
}

::google::protobuf::Metadata BloomFilterIndex::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Stream::Stream(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Stream_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.Stream)
}
Stream::Stream(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Stream& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Stream_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE Stream::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void Stream::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, kind_),
           0,
           offsetof(Impl_, length_) -
               offsetof(Impl_, kind_) +
               sizeof(Impl_::length_));
}
Stream::~Stream() {
  // @@protoc_insertion_point(destructor:orc.proto.Stream)
  SharedDtor(*this);
}
inline void Stream::SharedDtor(MessageLite& self) {
  Stream& this_ = static_cast<Stream&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Stream_class_data_ =
        Stream::InternalGenerateClassData_(Stream_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Stream::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Stream_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Stream_class_data_.tc_table);
  return Stream_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Stream::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Stream_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Stream_globals_));
  return Stream_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Stream::ParseTableT_
    Stream::_table_ =
        Stream::InternalGenerateParseTable_(Stream_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Stream::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.Stream)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    ::memset(&_impl_.kind_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.length_) -
        reinterpret_cast<char*>(&_impl_.kind_)) + sizeof(_impl_.length_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Stream::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Stream& this_ = static_cast<const Stream&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Stream::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Stream& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.Stream)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional .orc.proto.Stream.Kind kind = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        1, this_._internal_kind(), target);
  }

  // optional uint32 column = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this_._internal_column(), target);
  }

  // optional uint64 length = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this_._internal_length(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.Stream)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Stream::ByteSizeLong(const MessageLite& base) {
  const Stream& this_ = static_cast<const Stream&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Stream::ByteSizeLong() const {
  const Stream& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.Stream)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional .orc.proto.Stream.Kind kind = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_kind());
    }
    // optional uint32 column = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_column());
    }
    // optional uint64 length = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_length());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Stream::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Stream*>(&to_msg);
  auto& from = static_cast<const Stream&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.Stream)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.kind_ = from._impl_.kind_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.column_ = from._impl_.column_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.length_ = from._impl_.length_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Stream::CopyFrom(const Stream& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.Stream)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Stream::InternalSwap(Stream* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Stream, _impl_.length_)
      + sizeof(Stream::_impl_.length_)
      - PROTOBUF_FIELD_OFFSET(Stream, _impl_.kind_)>(
          reinterpret_cast<char*>(&_impl_.kind_),
          reinterpret_cast<char*>(&other->_impl_.kind_));
}

::google::protobuf::Metadata Stream::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ColumnEncoding::ColumnEncoding(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnEncoding_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.ColumnEncoding)
}
ColumnEncoding::ColumnEncoding(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ColumnEncoding& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnEncoding_get_class_data()),
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena),
#endif  // PROTOBUF_CUSTOM_VTABLE
      _impl_(from._impl_) {
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}
PROTOBUF_NDEBUG_INLINE ColumnEncoding::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void ColumnEncoding::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, kind_),
           0,
           offsetof(Impl_, bloom_encoding_) -
               offsetof(Impl_, kind_) +
               sizeof(Impl_::bloom_encoding_));
}
ColumnEncoding::~ColumnEncoding() {
  // @@protoc_insertion_point(destructor:orc.proto.ColumnEncoding)
  SharedDtor(*this);
}
inline void ColumnEncoding::SharedDtor(MessageLite& self) {
  ColumnEncoding& this_ = static_cast<ColumnEncoding&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ColumnEncoding_class_data_ =
        ColumnEncoding::InternalGenerateClassData_(ColumnEncoding_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnEncoding::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnEncoding_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ColumnEncoding_class_data_.tc_table);
  return ColumnEncoding_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnEncoding::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnEncoding_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ColumnEncoding_globals_));
  return ColumnEncoding_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ColumnEncoding::ParseTableT_
    ColumnEncoding::_table_ =
        ColumnEncoding::InternalGenerateParseTable_(ColumnEncoding_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ColumnEncoding::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.ColumnEncoding)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    ::memset(&_impl_.kind_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bloom_encoding_) -
        reinterpret_cast<char*>(&_impl_.kind_)) + sizeof(_impl_.bloom_encoding_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ColumnEncoding::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ColumnEncoding& this_ = static_cast<const ColumnEncoding&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ColumnEncoding::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ColumnEncoding& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.ColumnEncoding)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional .orc.proto.ColumnEncoding.Kind kind = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        1, this_._internal_kind(), target);
  }

  // optional uint32 dictionary_size = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this_._internal_dictionary_size(), target);
  }

  // optional uint32 bloom_encoding = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this_._internal_bloom_encoding(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.ColumnEncoding)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ColumnEncoding::ByteSizeLong(const MessageLite& base) {
  const ColumnEncoding& this_ = static_cast<const ColumnEncoding&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ColumnEncoding::ByteSizeLong() const {
  const ColumnEncoding& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.ColumnEncoding)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional .orc.proto.ColumnEncoding.Kind kind = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_kind());
    }
    // optional uint32 dictionary_size = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_dictionary_size());
    }
    // optional uint32 bloom_encoding = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_bloom_encoding());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ColumnEncoding::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ColumnEncoding*>(&to_msg);
  auto& from = static_cast<const ColumnEncoding&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.ColumnEncoding)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_impl_.kind_ = from._impl_.kind_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.dictionary_size_ = from._impl_.dictionary_size_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.bloom_encoding_ = from._impl_.bloom_encoding_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ColumnEncoding::CopyFrom(const ColumnEncoding& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.ColumnEncoding)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ColumnEncoding::InternalSwap(ColumnEncoding* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.bloom_encoding_)
      + sizeof(ColumnEncoding::_impl_.bloom_encoding_)
      - PROTOBUF_FIELD_OFFSET(ColumnEncoding, _impl_.kind_)>(
          reinterpret_cast<char*>(&_impl_.kind_),
          reinterpret_cast<char*>(&other->_impl_.kind_));
}

::google::protobuf::Metadata ColumnEncoding::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StripeEncryptionVariant::StripeEncryptionVariant(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeEncryptionVariant_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StripeEncryptionVariant)
}
PROTOBUF_NDEBUG_INLINE StripeEncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StripeEncryptionVariant& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        streams_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeEncryptionVariant,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.streams_)>()
          , from.streams_
        }
        ,
        encoding_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeEncryptionVariant,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.encoding_)>()
          , from.encoding_
        }
     {}

StripeEncryptionVariant::StripeEncryptionVariant(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StripeEncryptionVariant& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeEncryptionVariant_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StripeEncryptionVariant* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.StripeEncryptionVariant)
}
PROTOBUF_NDEBUG_INLINE StripeEncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        streams_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeEncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.streams_)>()
         }
        ,
        encoding_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeEncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeEncryptionVariant, _impl_.encoding_)>()
         }
     {}

inline void StripeEncryptionVariant::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
StripeEncryptionVariant::~StripeEncryptionVariant() {
  // @@protoc_insertion_point(destructor:orc.proto.StripeEncryptionVariant)
  SharedDtor(*this);
}
inline void StripeEncryptionVariant::SharedDtor(MessageLite& self) {
  StripeEncryptionVariant& this_ = static_cast<StripeEncryptionVariant&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StripeEncryptionVariant_class_data_ =
        StripeEncryptionVariant::InternalGenerateClassData_(StripeEncryptionVariant_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeEncryptionVariant::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeEncryptionVariant_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StripeEncryptionVariant_class_data_.tc_table);
  return StripeEncryptionVariant_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeEncryptionVariant::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeEncryptionVariant_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StripeEncryptionVariant_globals_));
  return StripeEncryptionVariant_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StripeEncryptionVariant::ParseTableT_
    StripeEncryptionVariant::_table_ =
        StripeEncryptionVariant::InternalGenerateParseTable_(StripeEncryptionVariant_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StripeEncryptionVariant::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StripeEncryptionVariant)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.streams_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.encoding_.Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StripeEncryptionVariant::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StripeEncryptionVariant& this_ = static_cast<const StripeEncryptionVariant&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StripeEncryptionVariant::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StripeEncryptionVariant& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StripeEncryptionVariant)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.Stream streams = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_streams_size());
         i < n; i++) {
      const auto& repfield = this_._internal_streams().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.ColumnEncoding encoding = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_encoding_size());
         i < n; i++) {
      const auto& repfield = this_._internal_encoding().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StripeEncryptionVariant)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StripeEncryptionVariant::ByteSizeLong(const MessageLite& base) {
  const StripeEncryptionVariant& this_ = static_cast<const StripeEncryptionVariant&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StripeEncryptionVariant::ByteSizeLong() const {
  const StripeEncryptionVariant& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StripeEncryptionVariant)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // repeated .orc.proto.Stream streams = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_streams_size();
      for (const auto& msg : this_._internal_streams()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.ColumnEncoding encoding = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_encoding_size();
      for (const auto& msg : this_._internal_encoding()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StripeEncryptionVariant::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StripeEncryptionVariant*>(&to_msg);
  auto& from = static_cast<const StripeEncryptionVariant&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StripeEncryptionVariant)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_streams()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_streams());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_encoding()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_encoding());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StripeEncryptionVariant::CopyFrom(const StripeEncryptionVariant& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StripeEncryptionVariant)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StripeEncryptionVariant::InternalSwap(StripeEncryptionVariant* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.streams_.InternalSwap(&other->_impl_.streams_);
  _impl_.encoding_.InternalSwap(&other->_impl_.encoding_);
}

::google::protobuf::Metadata StripeEncryptionVariant::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StripeFooter::StripeFooter(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeFooter_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StripeFooter)
}
PROTOBUF_NDEBUG_INLINE StripeFooter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StripeFooter& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        streams_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeFooter,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.streams_)>()
          , from.streams_
        }
        ,
        columns_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeFooter,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.columns_)>()
          , from.columns_
        }
        ,
        encryption_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeFooter,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.encryption_)>()
          , from.encryption_
        }
        ,
        writer_timezone_(arena, from.writer_timezone_) {}

StripeFooter::StripeFooter(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StripeFooter& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeFooter_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StripeFooter* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.StripeFooter)
}
PROTOBUF_NDEBUG_INLINE StripeFooter::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        streams_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.streams_)>()
         }
        ,
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.columns_)>()
         }
        ,
        encryption_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeFooter,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeFooter, _impl_.encryption_)>()
         }
        ,
        writer_timezone_(arena) {}

inline void StripeFooter::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
StripeFooter::~StripeFooter() {
  // @@protoc_insertion_point(destructor:orc.proto.StripeFooter)
  SharedDtor(*this);
}
inline void StripeFooter::SharedDtor(MessageLite& self) {
  StripeFooter& this_ = static_cast<StripeFooter&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.writer_timezone_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StripeFooter_class_data_ =
        StripeFooter::InternalGenerateClassData_(StripeFooter_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeFooter::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeFooter_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StripeFooter_class_data_.tc_table);
  return StripeFooter_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeFooter::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeFooter_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StripeFooter_globals_));
  return StripeFooter_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StripeFooter::ParseTableT_
    StripeFooter::_table_ =
        StripeFooter::InternalGenerateParseTable_(StripeFooter_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StripeFooter::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StripeFooter)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.streams_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.columns_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.encryption_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.writer_timezone_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StripeFooter::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StripeFooter& this_ = static_cast<const StripeFooter&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StripeFooter::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StripeFooter& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StripeFooter)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.Stream streams = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_streams_size());
         i < n; i++) {
      const auto& repfield = this_._internal_streams().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.ColumnEncoding columns = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_columns_size());
         i < n; i++) {
      const auto& repfield = this_._internal_columns().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional string writer_timezone = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_writer_timezone();
    target = stream->WriteStringMaybeAliased(3, _s, target);
  }

  // repeated .orc.proto.StripeEncryptionVariant encryption = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_encryption_size());
         i < n; i++) {
      const auto& repfield = this_._internal_encryption().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              4, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StripeFooter)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StripeFooter::ByteSizeLong(const MessageLite& base) {
  const StripeFooter& this_ = static_cast<const StripeFooter&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StripeFooter::ByteSizeLong() const {
  const StripeFooter& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StripeFooter)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // repeated .orc.proto.Stream streams = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_streams_size();
      for (const auto& msg : this_._internal_streams()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.ColumnEncoding columns = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_columns_size();
      for (const auto& msg : this_._internal_columns()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.StripeEncryptionVariant encryption = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_encryption_size();
      for (const auto& msg : this_._internal_encryption()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // optional string writer_timezone = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_writer_timezone());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StripeFooter::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StripeFooter*>(&to_msg);
  auto& from = static_cast<const StripeFooter&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StripeFooter)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_streams()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_streams());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_columns()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_columns());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_encryption()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_encryption());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_writer_timezone(from._internal_writer_timezone());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StripeFooter::CopyFrom(const StripeFooter& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StripeFooter)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StripeFooter::InternalSwap(StripeFooter* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.streams_.InternalSwap(&other->_impl_.streams_);
  _impl_.columns_.InternalSwap(&other->_impl_.columns_);
  _impl_.encryption_.InternalSwap(&other->_impl_.encryption_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.writer_timezone_, &other->_impl_.writer_timezone_, arena);
}

::google::protobuf::Metadata StripeFooter::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StringPair::StringPair(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StringPair_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StringPair)
}
PROTOBUF_NDEBUG_INLINE StringPair::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StringPair& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        key_(arena, from.key_),
        value_(arena, from.value_) {}

StringPair::StringPair(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StringPair& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StringPair_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StringPair* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.StringPair)
}
PROTOBUF_NDEBUG_INLINE StringPair::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        key_(arena),
        value_(arena) {}

inline void StringPair::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
StringPair::~StringPair() {
  // @@protoc_insertion_point(destructor:orc.proto.StringPair)
  SharedDtor(*this);
}
inline void StringPair::SharedDtor(MessageLite& self) {
  StringPair& this_ = static_cast<StringPair&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.key_.Destroy();
  this_._impl_.value_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StringPair_class_data_ =
        StringPair::InternalGenerateClassData_(StringPair_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StringPair::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StringPair_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StringPair_class_data_.tc_table);
  return StringPair_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StringPair::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StringPair_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StringPair_globals_));
  return StringPair_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StringPair::ParseTableT_
    StringPair::_table_ =
        StringPair::InternalGenerateParseTable_(StringPair_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StringPair::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StringPair)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.key_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.value_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StringPair::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StringPair& this_ = static_cast<const StringPair&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StringPair::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StringPair& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StringPair)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string key = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_key();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string value = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_value();
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StringPair)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StringPair::ByteSizeLong(const MessageLite& base) {
  const StringPair& this_ = static_cast<const StringPair&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StringPair::ByteSizeLong() const {
  const StringPair& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StringPair)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional string key = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_key());
    }
    // optional string value = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_value());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StringPair::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StringPair*>(&to_msg);
  auto& from = static_cast<const StringPair&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StringPair)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_key(from._internal_key());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_value(from._internal_value());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StringPair::CopyFrom(const StringPair& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StringPair)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StringPair::InternalSwap(StringPair* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.key_, &other->_impl_.key_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.value_, &other->_impl_.value_, arena);
}

::google::protobuf::Metadata StringPair::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Type::Type(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Type_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.Type)
}
PROTOBUF_NDEBUG_INLINE Type::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::Type& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        subtypes_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Type,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.subtypes_)>()
          , from.subtypes_
        }
        ,
        _subtypes_cached_byte_size_{0},
        field_names_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Type,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.field_names_)>()
          , from.field_names_
        }
        ,
        attributes_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Type,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.attributes_)>()
          , from.attributes_
        }
        ,
        crs_(arena, from.crs_) {}

Type::Type(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Type& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Type_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Type* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, kind_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, kind_),
           offsetof(Impl_, algorithm_) -
               offsetof(Impl_, kind_) +
               sizeof(Impl_::algorithm_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.Type)
}
PROTOBUF_NDEBUG_INLINE Type::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        subtypes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.subtypes_)>()
         }
        ,
        _subtypes_cached_byte_size_{0},
        field_names_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.field_names_)>()
         }
        ,
        attributes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Type,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Type, _impl_.attributes_)>()
         }
        ,
        crs_(arena) {}

inline void Type::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, kind_),
           0,
           offsetof(Impl_, algorithm_) -
               offsetof(Impl_, kind_) +
               sizeof(Impl_::algorithm_));
}
Type::~Type() {
  // @@protoc_insertion_point(destructor:orc.proto.Type)
  SharedDtor(*this);
}
inline void Type::SharedDtor(MessageLite& self) {
  Type& this_ = static_cast<Type&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.crs_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Type_class_data_ =
        Type::InternalGenerateClassData_(Type_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Type::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Type_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Type_class_data_.tc_table);
  return Type_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Type::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Type_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Type_globals_));
  return Type_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Type::ParseTableT_
    Type::_table_ =
        Type::InternalGenerateParseTable_(Type_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Type::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.Type)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.subtypes_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.field_names_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.attributes_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.crs_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x000000f0U)) {
    ::memset(&_impl_.kind_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.scale_) -
        reinterpret_cast<char*>(&_impl_.kind_)) + sizeof(_impl_.scale_));
  }
  _impl_.algorithm_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Type::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Type& this_ = static_cast<const Type&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Type::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Type& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.Type)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional .orc.proto.Type.Kind kind = 1;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        1, this_._internal_kind(), target);
  }

  // repeated uint32 subtypes = 2 [packed = true];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    {
      int byte_size = this_._impl_._subtypes_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt32Packed(
            2, this_._internal_subtypes(), byte_size, target);
      }
    }
  }

  // repeated string field_names = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (int i = 0, n = this_._internal_field_names_size(); i < n; ++i) {
      const auto& s = this_._internal_field_names().Get(i);
      target = stream->WriteString(3, s, target);
    }
  }

  // optional uint32 maximum_length = 4;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this_._internal_maximum_length(), target);
  }

  // optional uint32 precision = 5;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this_._internal_precision(), target);
  }

  // optional uint32 scale = 6;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this_._internal_scale(), target);
  }

  // repeated .orc.proto.StringPair attributes = 7;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_attributes_size());
         i < n; i++) {
      const auto& repfield = this_._internal_attributes().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              7, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional string crs = 8;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    const ::std::string& _s = this_._internal_crs();
    target = stream->WriteStringMaybeAliased(8, _s, target);
  }

  // optional .orc.proto.Type.EdgeInterpolationAlgorithm algorithm = 9;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        9, this_._internal_algorithm(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.Type)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Type::ByteSizeLong(const MessageLite& base) {
  const Type& this_ = static_cast<const Type&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Type::ByteSizeLong() const {
  const Type& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.Type)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated uint32 subtypes = 2 [packed = true];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt32SizeWithPackedTagSize(
              this_._internal_subtypes(), 1,
              this_._impl_._subtypes_cached_byte_size_);
    }
    // repeated string field_names = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_field_names().size());
      for (int i = 0, n = this_._internal_field_names().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_field_names().Get(i));
      }
    }
    // repeated .orc.proto.StringPair attributes = 7;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_attributes_size();
      for (const auto& msg : this_._internal_attributes()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // optional string crs = 8;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_crs());
    }
    // optional .orc.proto.Type.Kind kind = 1;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_kind());
    }
    // optional uint32 maximum_length = 4;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_maximum_length());
    }
    // optional uint32 precision = 5;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_precision());
    }
    // optional uint32 scale = 6;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_scale());
    }
  }
   {
    // optional .orc.proto.Type.EdgeInterpolationAlgorithm algorithm = 9;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_algorithm());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Type::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Type*>(&to_msg);
  auto& from = static_cast<const Type&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.Type)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_subtypes()->MergeFrom(from._internal_subtypes());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_field_names()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_field_names());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_attributes()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_attributes());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_set_crs(from._internal_crs());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.kind_ = from._impl_.kind_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _this->_impl_.maximum_length_ = from._impl_.maximum_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.precision_ = from._impl_.precision_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _this->_impl_.scale_ = from._impl_.scale_;
    }
  }
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    _this->_impl_.algorithm_ = from._impl_.algorithm_;
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Type::CopyFrom(const Type& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.Type)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Type::InternalSwap(Type* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.subtypes_.InternalSwap(&other->_impl_.subtypes_);
  _impl_.field_names_.InternalSwap(&other->_impl_.field_names_);
  _impl_.attributes_.InternalSwap(&other->_impl_.attributes_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.crs_, &other->_impl_.crs_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Type, _impl_.algorithm_)
      + sizeof(Type::_impl_.algorithm_)
      - PROTOBUF_FIELD_OFFSET(Type, _impl_.kind_)>(
          reinterpret_cast<char*>(&_impl_.kind_),
          reinterpret_cast<char*>(&other->_impl_.kind_));
}

::google::protobuf::Metadata Type::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StripeInformation::StripeInformation(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeInformation_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StripeInformation)
}
PROTOBUF_NDEBUG_INLINE StripeInformation::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StripeInformation& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        encrypted_local_keys_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeInformation,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.encrypted_local_keys_)>()
          , from.encrypted_local_keys_
        }
     {}

StripeInformation::StripeInformation(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StripeInformation& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeInformation_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StripeInformation* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, offset_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, offset_),
           offsetof(Impl_, encrypt_stripe_id_) -
               offsetof(Impl_, offset_) +
               sizeof(Impl_::encrypt_stripe_id_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.StripeInformation)
}
PROTOBUF_NDEBUG_INLINE StripeInformation::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        encrypted_local_keys_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeInformation,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeInformation, _impl_.encrypted_local_keys_)>()
         }
     {}

inline void StripeInformation::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, offset_),
           0,
           offsetof(Impl_, encrypt_stripe_id_) -
               offsetof(Impl_, offset_) +
               sizeof(Impl_::encrypt_stripe_id_));
}
StripeInformation::~StripeInformation() {
  // @@protoc_insertion_point(destructor:orc.proto.StripeInformation)
  SharedDtor(*this);
}
inline void StripeInformation::SharedDtor(MessageLite& self) {
  StripeInformation& this_ = static_cast<StripeInformation&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StripeInformation_class_data_ =
        StripeInformation::InternalGenerateClassData_(StripeInformation_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeInformation::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeInformation_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StripeInformation_class_data_.tc_table);
  return StripeInformation_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeInformation::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeInformation_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StripeInformation_globals_));
  return StripeInformation_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StripeInformation::ParseTableT_
    StripeInformation::_table_ =
        StripeInformation::InternalGenerateParseTable_(StripeInformation_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StripeInformation::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StripeInformation)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.encrypted_local_keys_.Clear();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000007eU)) {
    ::memset(&_impl_.offset_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.encrypt_stripe_id_) -
        reinterpret_cast<char*>(&_impl_.offset_)) + sizeof(_impl_.encrypt_stripe_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StripeInformation::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StripeInformation& this_ = static_cast<const StripeInformation&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StripeInformation::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StripeInformation& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StripeInformation)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint64 offset = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this_._internal_offset(), target);
  }

  // optional uint64 index_length = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this_._internal_index_length(), target);
  }

  // optional uint64 data_length = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this_._internal_data_length(), target);
  }

  // optional uint64 footer_length = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this_._internal_footer_length(), target);
  }

  // optional uint64 number_of_rows = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this_._internal_number_of_rows(), target);
  }

  // optional uint64 encrypt_stripe_id = 6;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        6, this_._internal_encrypt_stripe_id(), target);
  }

  // repeated bytes encrypted_local_keys = 7;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_encrypted_local_keys_size(); i < n; ++i) {
      const auto& s = this_._internal_encrypted_local_keys().Get(i);
      target = stream->WriteBytes(7, s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StripeInformation)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StripeInformation::ByteSizeLong(const MessageLite& base) {
  const StripeInformation& this_ = static_cast<const StripeInformation&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StripeInformation::ByteSizeLong() const {
  const StripeInformation& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StripeInformation)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    // repeated bytes encrypted_local_keys = 7;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_encrypted_local_keys().size());
      for (int i = 0, n = this_._internal_encrypted_local_keys().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::BytesSize(
            this_._internal_encrypted_local_keys().Get(i));
      }
    }
    // optional uint64 offset = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_offset());
    }
    // optional uint64 index_length = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_index_length());
    }
    // optional uint64 data_length = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_data_length());
    }
    // optional uint64 footer_length = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_footer_length());
    }
    // optional uint64 number_of_rows = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_number_of_rows());
    }
    // optional uint64 encrypt_stripe_id = 6;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_encrypt_stripe_id());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StripeInformation::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StripeInformation*>(&to_msg);
  auto& from = static_cast<const StripeInformation&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StripeInformation)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000007fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_encrypted_local_keys()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_encrypted_local_keys());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.index_length_ = from._impl_.index_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.data_length_ = from._impl_.data_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.footer_length_ = from._impl_.footer_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _this->_impl_.number_of_rows_ = from._impl_.number_of_rows_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.encrypt_stripe_id_ = from._impl_.encrypt_stripe_id_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StripeInformation::CopyFrom(const StripeInformation& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StripeInformation)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StripeInformation::InternalSwap(StripeInformation* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.encrypted_local_keys_.InternalSwap(&other->_impl_.encrypted_local_keys_);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.encrypt_stripe_id_)
      + sizeof(StripeInformation::_impl_.encrypt_stripe_id_)
      - PROTOBUF_FIELD_OFFSET(StripeInformation, _impl_.offset_)>(
          reinterpret_cast<char*>(&_impl_.offset_),
          reinterpret_cast<char*>(&other->_impl_.offset_));
}

::google::protobuf::Metadata StripeInformation::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

UserMetadataItem::UserMetadataItem(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UserMetadataItem_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.UserMetadataItem)
}
PROTOBUF_NDEBUG_INLINE UserMetadataItem::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::UserMetadataItem& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        name_(arena, from.name_),
        value_(arena, from.value_) {}

UserMetadataItem::UserMetadataItem(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const UserMetadataItem& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UserMetadataItem_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  UserMetadataItem* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.UserMetadataItem)
}
PROTOBUF_NDEBUG_INLINE UserMetadataItem::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        name_(arena),
        value_(arena) {}

inline void UserMetadataItem::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
UserMetadataItem::~UserMetadataItem() {
  // @@protoc_insertion_point(destructor:orc.proto.UserMetadataItem)
  SharedDtor(*this);
}
inline void UserMetadataItem::SharedDtor(MessageLite& self) {
  UserMetadataItem& this_ = static_cast<UserMetadataItem&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.name_.Destroy();
  this_._impl_.value_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull UserMetadataItem_class_data_ =
        UserMetadataItem::InternalGenerateClassData_(UserMetadataItem_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UserMetadataItem::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UserMetadataItem_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(UserMetadataItem_class_data_.tc_table);
  return UserMetadataItem_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UserMetadataItem::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UserMetadataItem_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&UserMetadataItem_globals_));
  return UserMetadataItem_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const UserMetadataItem::ParseTableT_
    UserMetadataItem::_table_ =
        UserMetadataItem::InternalGenerateParseTable_(UserMetadataItem_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void UserMetadataItem::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.UserMetadataItem)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.value_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL UserMetadataItem::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const UserMetadataItem& this_ = static_cast<const UserMetadataItem&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL UserMetadataItem::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const UserMetadataItem& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.UserMetadataItem)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_name();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional bytes value = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_value();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.UserMetadataItem)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t UserMetadataItem::ByteSizeLong(const MessageLite& base) {
  const UserMetadataItem& this_ = static_cast<const UserMetadataItem&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t UserMetadataItem::ByteSizeLong() const {
  const UserMetadataItem& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.UserMetadataItem)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // optional string name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_name());
    }
    // optional bytes value = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_value());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void UserMetadataItem::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<UserMetadataItem*>(&to_msg);
  auto& from = static_cast<const UserMetadataItem&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.UserMetadataItem)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_name(from._internal_name());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_value(from._internal_value());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void UserMetadataItem::CopyFrom(const UserMetadataItem& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.UserMetadataItem)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void UserMetadataItem::InternalSwap(UserMetadataItem* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.value_, &other->_impl_.value_, arena);
}

::google::protobuf::Metadata UserMetadataItem::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StripeStatistics::StripeStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.StripeStatistics)
}
PROTOBUF_NDEBUG_INLINE StripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::StripeStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        col_stats_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::StripeStatistics,
              PROTOBUF_FIELD_OFFSET(::orc::proto::StripeStatistics, _impl_.col_stats_)>()
          , from.col_stats_
        }
     {}

StripeStatistics::StripeStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StripeStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StripeStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StripeStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.StripeStatistics)
}
PROTOBUF_NDEBUG_INLINE StripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        col_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::StripeStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::StripeStatistics, _impl_.col_stats_)>()
         }
     {}

inline void StripeStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
StripeStatistics::~StripeStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.StripeStatistics)
  SharedDtor(*this);
}
inline void StripeStatistics::SharedDtor(MessageLite& self) {
  StripeStatistics& this_ = static_cast<StripeStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StripeStatistics_class_data_ =
        StripeStatistics::InternalGenerateClassData_(StripeStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StripeStatistics_class_data_.tc_table);
  return StripeStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StripeStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StripeStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StripeStatistics_globals_));
  return StripeStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StripeStatistics::ParseTableT_
    StripeStatistics::_table_ =
        StripeStatistics::InternalGenerateParseTable_(StripeStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StripeStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.StripeStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.col_stats_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StripeStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StripeStatistics& this_ = static_cast<const StripeStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StripeStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StripeStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.StripeStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.ColumnStatistics col_stats = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_col_stats_size());
         i < n; i++) {
      const auto& repfield = this_._internal_col_stats().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.StripeStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StripeStatistics::ByteSizeLong(const MessageLite& base) {
  const StripeStatistics& this_ = static_cast<const StripeStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StripeStatistics::ByteSizeLong() const {
  const StripeStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.StripeStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.ColumnStatistics col_stats = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_col_stats_size();
      for (const auto& msg : this_._internal_col_stats()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StripeStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StripeStatistics*>(&to_msg);
  auto& from = static_cast<const StripeStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.StripeStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_col_stats()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_col_stats());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StripeStatistics::CopyFrom(const StripeStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.StripeStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StripeStatistics::InternalSwap(StripeStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.col_stats_.InternalSwap(&other->_impl_.col_stats_);
}

::google::protobuf::Metadata StripeStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Metadata::Metadata(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Metadata_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.Metadata)
}
PROTOBUF_NDEBUG_INLINE Metadata::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::Metadata& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        stripe_stats_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Metadata,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Metadata, _impl_.stripe_stats_)>()
          , from.stripe_stats_
        }
     {}

Metadata::Metadata(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Metadata& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Metadata_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Metadata* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.Metadata)
}
PROTOBUF_NDEBUG_INLINE Metadata::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        stripe_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Metadata,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Metadata, _impl_.stripe_stats_)>()
         }
     {}

inline void Metadata::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Metadata::~Metadata() {
  // @@protoc_insertion_point(destructor:orc.proto.Metadata)
  SharedDtor(*this);
}
inline void Metadata::SharedDtor(MessageLite& self) {
  Metadata& this_ = static_cast<Metadata&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Metadata_class_data_ =
        Metadata::InternalGenerateClassData_(Metadata_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Metadata::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Metadata_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Metadata_class_data_.tc_table);
  return Metadata_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Metadata::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Metadata_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Metadata_globals_));
  return Metadata_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Metadata::ParseTableT_
    Metadata::_table_ =
        Metadata::InternalGenerateParseTable_(Metadata_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Metadata::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.Metadata)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.stripe_stats_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Metadata::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Metadata& this_ = static_cast<const Metadata&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Metadata::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Metadata& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.Metadata)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.StripeStatistics stripe_stats = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_stripe_stats_size());
         i < n; i++) {
      const auto& repfield = this_._internal_stripe_stats().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.Metadata)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Metadata::ByteSizeLong(const MessageLite& base) {
  const Metadata& this_ = static_cast<const Metadata&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Metadata::ByteSizeLong() const {
  const Metadata& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.Metadata)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.StripeStatistics stripe_stats = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_stripe_stats_size();
      for (const auto& msg : this_._internal_stripe_stats()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Metadata::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Metadata*>(&to_msg);
  auto& from = static_cast<const Metadata&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.Metadata)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_stripe_stats()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_stripe_stats());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Metadata::CopyFrom(const Metadata& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.Metadata)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Metadata::InternalSwap(Metadata* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.stripe_stats_.InternalSwap(&other->_impl_.stripe_stats_);
}

::google::protobuf::Metadata Metadata::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ColumnarStripeStatistics::ColumnarStripeStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnarStripeStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.ColumnarStripeStatistics)
}
PROTOBUF_NDEBUG_INLINE ColumnarStripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::ColumnarStripeStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        col_stats_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::ColumnarStripeStatistics,
              PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnarStripeStatistics, _impl_.col_stats_)>()
          , from.col_stats_
        }
     {}

ColumnarStripeStatistics::ColumnarStripeStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ColumnarStripeStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ColumnarStripeStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ColumnarStripeStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.ColumnarStripeStatistics)
}
PROTOBUF_NDEBUG_INLINE ColumnarStripeStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        col_stats_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::ColumnarStripeStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::ColumnarStripeStatistics, _impl_.col_stats_)>()
         }
     {}

inline void ColumnarStripeStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
ColumnarStripeStatistics::~ColumnarStripeStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.ColumnarStripeStatistics)
  SharedDtor(*this);
}
inline void ColumnarStripeStatistics::SharedDtor(MessageLite& self) {
  ColumnarStripeStatistics& this_ = static_cast<ColumnarStripeStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ColumnarStripeStatistics_class_data_ =
        ColumnarStripeStatistics::InternalGenerateClassData_(ColumnarStripeStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnarStripeStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnarStripeStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ColumnarStripeStatistics_class_data_.tc_table);
  return ColumnarStripeStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ColumnarStripeStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ColumnarStripeStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ColumnarStripeStatistics_globals_));
  return ColumnarStripeStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ColumnarStripeStatistics::ParseTableT_
    ColumnarStripeStatistics::_table_ =
        ColumnarStripeStatistics::InternalGenerateParseTable_(ColumnarStripeStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ColumnarStripeStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.ColumnarStripeStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.col_stats_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ColumnarStripeStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ColumnarStripeStatistics& this_ = static_cast<const ColumnarStripeStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ColumnarStripeStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ColumnarStripeStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.ColumnarStripeStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.ColumnStatistics col_stats = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_col_stats_size());
         i < n; i++) {
      const auto& repfield = this_._internal_col_stats().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.ColumnarStripeStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ColumnarStripeStatistics::ByteSizeLong(const MessageLite& base) {
  const ColumnarStripeStatistics& this_ = static_cast<const ColumnarStripeStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ColumnarStripeStatistics::ByteSizeLong() const {
  const ColumnarStripeStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.ColumnarStripeStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.ColumnStatistics col_stats = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_col_stats_size();
      for (const auto& msg : this_._internal_col_stats()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ColumnarStripeStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ColumnarStripeStatistics*>(&to_msg);
  auto& from = static_cast<const ColumnarStripeStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.ColumnarStripeStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_col_stats()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_col_stats());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ColumnarStripeStatistics::CopyFrom(const ColumnarStripeStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.ColumnarStripeStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ColumnarStripeStatistics::InternalSwap(ColumnarStripeStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.col_stats_.InternalSwap(&other->_impl_.col_stats_);
}

::google::protobuf::Metadata ColumnarStripeStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

FileStatistics::FileStatistics(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FileStatistics_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.FileStatistics)
}
PROTOBUF_NDEBUG_INLINE FileStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::FileStatistics& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        column_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::FileStatistics,
              PROTOBUF_FIELD_OFFSET(::orc::proto::FileStatistics, _impl_.column_)>()
          , from.column_
        }
     {}

FileStatistics::FileStatistics(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FileStatistics& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FileStatistics_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FileStatistics* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.FileStatistics)
}
PROTOBUF_NDEBUG_INLINE FileStatistics::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        column_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::FileStatistics,
            PROTOBUF_FIELD_OFFSET(::orc::proto::FileStatistics, _impl_.column_)>()
         }
     {}

inline void FileStatistics::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
FileStatistics::~FileStatistics() {
  // @@protoc_insertion_point(destructor:orc.proto.FileStatistics)
  SharedDtor(*this);
}
inline void FileStatistics::SharedDtor(MessageLite& self) {
  FileStatistics& this_ = static_cast<FileStatistics&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FileStatistics_class_data_ =
        FileStatistics::InternalGenerateClassData_(FileStatistics_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FileStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FileStatistics_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FileStatistics_class_data_.tc_table);
  return FileStatistics_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FileStatistics::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FileStatistics_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&FileStatistics_globals_));
  return FileStatistics_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const FileStatistics::ParseTableT_
    FileStatistics::_table_ =
        FileStatistics::InternalGenerateParseTable_(FileStatistics_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void FileStatistics::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.FileStatistics)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.column_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FileStatistics::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FileStatistics& this_ = static_cast<const FileStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FileStatistics::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FileStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.FileStatistics)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.ColumnStatistics column = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_column_size());
         i < n; i++) {
      const auto& repfield = this_._internal_column().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.FileStatistics)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FileStatistics::ByteSizeLong(const MessageLite& base) {
  const FileStatistics& this_ = static_cast<const FileStatistics&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FileStatistics::ByteSizeLong() const {
  const FileStatistics& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.FileStatistics)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .orc.proto.ColumnStatistics column = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_column_size();
      for (const auto& msg : this_._internal_column()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FileStatistics::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<FileStatistics*>(&to_msg);
  auto& from = static_cast<const FileStatistics&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.FileStatistics)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_column()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_column());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FileStatistics::CopyFrom(const FileStatistics& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.FileStatistics)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FileStatistics::InternalSwap(FileStatistics* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.column_.InternalSwap(&other->_impl_.column_);
}

::google::protobuf::Metadata FileStatistics::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

DataMask::DataMask(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DataMask_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.DataMask)
}
PROTOBUF_NDEBUG_INLINE DataMask::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::DataMask& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mask_parameters_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::DataMask,
              PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.mask_parameters_)>()
          , from.mask_parameters_
        }
        ,
        columns_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::DataMask,
              PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.columns_)>()
          , from.columns_
        }
        ,
        _columns_cached_byte_size_{0},
        name_(arena, from.name_) {}

DataMask::DataMask(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const DataMask& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DataMask_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  DataMask* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:orc.proto.DataMask)
}
PROTOBUF_NDEBUG_INLINE DataMask::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mask_parameters_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::DataMask,
            PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.mask_parameters_)>()
         }
        ,
        columns_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::DataMask,
            PROTOBUF_FIELD_OFFSET(::orc::proto::DataMask, _impl_.columns_)>()
         }
        ,
        _columns_cached_byte_size_{0},
        name_(arena) {}

inline void DataMask::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
DataMask::~DataMask() {
  // @@protoc_insertion_point(destructor:orc.proto.DataMask)
  SharedDtor(*this);
}
inline void DataMask::SharedDtor(MessageLite& self) {
  DataMask& this_ = static_cast<DataMask&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.name_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DataMask_class_data_ =
        DataMask::InternalGenerateClassData_(DataMask_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DataMask::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DataMask_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DataMask_class_data_.tc_table);
  return DataMask_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DataMask::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DataMask_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DataMask_globals_));
  return DataMask_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DataMask::ParseTableT_
    DataMask::_table_ =
        DataMask::InternalGenerateParseTable_(DataMask_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DataMask::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.DataMask)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.mask_parameters_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.columns_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DataMask::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DataMask& this_ = static_cast<const DataMask&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DataMask::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DataMask& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.DataMask)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_name();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // repeated string mask_parameters = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_mask_parameters_size(); i < n; ++i) {
      const auto& s = this_._internal_mask_parameters().Get(i);
      target = stream->WriteString(2, s, target);
    }
  }

  // repeated uint32 columns = 3 [packed = true];
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    {
      int byte_size = this_._impl_._columns_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt32Packed(
            3, this_._internal_columns(), byte_size, target);
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.DataMask)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DataMask::ByteSizeLong(const MessageLite& base) {
  const DataMask& this_ = static_cast<const DataMask&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DataMask::ByteSizeLong() const {
  const DataMask& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.DataMask)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // repeated string mask_parameters = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_mask_parameters().size());
      for (int i = 0, n = this_._internal_mask_parameters().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_mask_parameters().Get(i));
      }
    }
    // repeated uint32 columns = 3 [packed = true];
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt32SizeWithPackedTagSize(
              this_._internal_columns(), 1,
              this_._impl_._columns_cached_byte_size_);
    }
    // optional string name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_name());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DataMask::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DataMask*>(&to_msg);
  auto& from = static_cast<const DataMask&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.DataMask)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_mask_parameters()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_mask_parameters());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_columns()->MergeFrom(from._internal_columns());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_name(from._internal_name());
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DataMask::CopyFrom(const DataMask& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.DataMask)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DataMask::InternalSwap(DataMask* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.mask_parameters_.InternalSwap(&other->_impl_.mask_parameters_);
  _impl_.columns_.InternalSwap(&other->_impl_.columns_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
}

::google::protobuf::Metadata DataMask::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

EncryptionKey::EncryptionKey(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, EncryptionKey_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.EncryptionKey)
}
PROTOBUF_NDEBUG_INLINE EncryptionKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::EncryptionKey& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        key_name_(arena, from.key_name_) {}

EncryptionKey::EncryptionKey(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const EncryptionKey& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, EncryptionKey_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  EncryptionKey* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, key_version_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, key_version_),
           offsetof(Impl_, algorithm_) -
               offsetof(Impl_, key_version_) +
               sizeof(Impl_::algorithm_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.EncryptionKey)
}
PROTOBUF_NDEBUG_INLINE EncryptionKey::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        key_name_(arena) {}

inline void EncryptionKey::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, key_version_),
           0,
           offsetof(Impl_, algorithm_) -
               offsetof(Impl_, key_version_) +
               sizeof(Impl_::algorithm_));
}
EncryptionKey::~EncryptionKey() {
  // @@protoc_insertion_point(destructor:orc.proto.EncryptionKey)
  SharedDtor(*this);
}
inline void EncryptionKey::SharedDtor(MessageLite& self) {
  EncryptionKey& this_ = static_cast<EncryptionKey&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.key_name_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull EncryptionKey_class_data_ =
        EncryptionKey::InternalGenerateClassData_(EncryptionKey_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
EncryptionKey::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&EncryptionKey_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(EncryptionKey_class_data_.tc_table);
  return EncryptionKey_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
EncryptionKey::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&EncryptionKey_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&EncryptionKey_globals_));
  return EncryptionKey_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const EncryptionKey::ParseTableT_
    EncryptionKey::_table_ =
        EncryptionKey::InternalGenerateParseTable_(EncryptionKey_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void EncryptionKey::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.EncryptionKey)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.key_name_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    ::memset(&_impl_.key_version_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.algorithm_) -
        reinterpret_cast<char*>(&_impl_.key_version_)) + sizeof(_impl_.algorithm_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL EncryptionKey::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const EncryptionKey& this_ = static_cast<const EncryptionKey&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL EncryptionKey::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const EncryptionKey& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.EncryptionKey)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional string key_name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    const ::std::string& _s = this_._internal_key_name();
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional uint32 key_version = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this_._internal_key_version(), target);
  }

  // optional .orc.proto.EncryptionAlgorithm algorithm = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        3, this_._internal_algorithm(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.EncryptionKey)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t EncryptionKey::ByteSizeLong(const MessageLite& base) {
  const EncryptionKey& this_ = static_cast<const EncryptionKey&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t EncryptionKey::ByteSizeLong() const {
  const EncryptionKey& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.EncryptionKey)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // optional string key_name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_key_name());
    }
    // optional uint32 key_version = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_key_version());
    }
    // optional .orc.proto.EncryptionAlgorithm algorithm = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_algorithm());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void EncryptionKey::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<EncryptionKey*>(&to_msg);
  auto& from = static_cast<const EncryptionKey&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.EncryptionKey)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_set_key_name(from._internal_key_name());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_impl_.key_version_ = from._impl_.key_version_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.algorithm_ = from._impl_.algorithm_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void EncryptionKey::CopyFrom(const EncryptionKey& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.EncryptionKey)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void EncryptionKey::InternalSwap(EncryptionKey* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.key_name_, &other->_impl_.key_name_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.algorithm_)
      + sizeof(EncryptionKey::_impl_.algorithm_)
      - PROTOBUF_FIELD_OFFSET(EncryptionKey, _impl_.key_version_)>(
          reinterpret_cast<char*>(&_impl_.key_version_),
          reinterpret_cast<char*>(&other->_impl_.key_version_));
}

::google::protobuf::Metadata EncryptionKey::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

EncryptionVariant::EncryptionVariant(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, EncryptionVariant_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.EncryptionVariant)
}
PROTOBUF_NDEBUG_INLINE EncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::EncryptionVariant& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        stripe_statistics_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::EncryptionVariant,
              PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.stripe_statistics_)>()
          , from.stripe_statistics_
        }
        ,
        encrypted_key_(arena, from.encrypted_key_),
        file_statistics_(arena, from.file_statistics_) {}

EncryptionVariant::EncryptionVariant(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const EncryptionVariant& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, EncryptionVariant_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  EncryptionVariant* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, root_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, root_),
           offsetof(Impl_, key_) -
               offsetof(Impl_, root_) +
               sizeof(Impl_::key_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.EncryptionVariant)
}
PROTOBUF_NDEBUG_INLINE EncryptionVariant::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        stripe_statistics_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::EncryptionVariant,
            PROTOBUF_FIELD_OFFSET(::orc::proto::EncryptionVariant, _impl_.stripe_statistics_)>()
         }
        ,
        encrypted_key_(arena),
        file_statistics_(arena) {}

inline void EncryptionVariant::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, root_),
           0,
           offsetof(Impl_, key_) -
               offsetof(Impl_, root_) +
               sizeof(Impl_::key_));
}
EncryptionVariant::~EncryptionVariant() {
  // @@protoc_insertion_point(destructor:orc.proto.EncryptionVariant)
  SharedDtor(*this);
}
inline void EncryptionVariant::SharedDtor(MessageLite& self) {
  EncryptionVariant& this_ = static_cast<EncryptionVariant&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.encrypted_key_.Destroy();
  this_._impl_.file_statistics_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull EncryptionVariant_class_data_ =
        EncryptionVariant::InternalGenerateClassData_(EncryptionVariant_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
EncryptionVariant::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&EncryptionVariant_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(EncryptionVariant_class_data_.tc_table);
  return EncryptionVariant_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
EncryptionVariant::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&EncryptionVariant_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&EncryptionVariant_globals_));
  return EncryptionVariant_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const EncryptionVariant::ParseTableT_
    EncryptionVariant::_table_ =
        EncryptionVariant::InternalGenerateParseTable_(EncryptionVariant_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void EncryptionVariant::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.EncryptionVariant)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.stripe_statistics_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.encrypted_key_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.file_statistics_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000018U)) {
    ::memset(&_impl_.root_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.key_) -
        reinterpret_cast<char*>(&_impl_.root_)) + sizeof(_impl_.key_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL EncryptionVariant::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const EncryptionVariant& this_ = static_cast<const EncryptionVariant&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL EncryptionVariant::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const EncryptionVariant& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.EncryptionVariant)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint32 root = 1;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this_._internal_root(), target);
  }

  // optional uint32 key = 2;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this_._internal_key(), target);
  }

  // optional bytes encrypted_key = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_encrypted_key();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // repeated .orc.proto.Stream stripe_statistics = 4;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_stripe_statistics_size());
         i < n; i++) {
      const auto& repfield = this_._internal_stripe_statistics().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              4, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional bytes file_statistics = 5;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    const ::std::string& _s = this_._internal_file_statistics();
    target = stream->WriteBytesMaybeAliased(5, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.EncryptionVariant)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t EncryptionVariant::ByteSizeLong(const MessageLite& base) {
  const EncryptionVariant& this_ = static_cast<const EncryptionVariant&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t EncryptionVariant::ByteSizeLong() const {
  const EncryptionVariant& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.EncryptionVariant)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // repeated .orc.proto.Stream stripe_statistics = 4;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_stripe_statistics_size();
      for (const auto& msg : this_._internal_stripe_statistics()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // optional bytes encrypted_key = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_encrypted_key());
    }
    // optional bytes file_statistics = 5;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this_._internal_file_statistics());
    }
    // optional uint32 root = 1;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_root());
    }
    // optional uint32 key = 2;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_key());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void EncryptionVariant::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<EncryptionVariant*>(&to_msg);
  auto& from = static_cast<const EncryptionVariant&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.EncryptionVariant)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_stripe_statistics()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_stripe_statistics());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_encrypted_key(from._internal_encrypted_key());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_set_file_statistics(from._internal_file_statistics());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.root_ = from._impl_.root_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.key_ = from._impl_.key_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void EncryptionVariant::CopyFrom(const EncryptionVariant& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.EncryptionVariant)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void EncryptionVariant::InternalSwap(EncryptionVariant* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.stripe_statistics_.InternalSwap(&other->_impl_.stripe_statistics_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.encrypted_key_, &other->_impl_.encrypted_key_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.file_statistics_, &other->_impl_.file_statistics_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.key_)
      + sizeof(EncryptionVariant::_impl_.key_)
      - PROTOBUF_FIELD_OFFSET(EncryptionVariant, _impl_.root_)>(
          reinterpret_cast<char*>(&_impl_.root_),
          reinterpret_cast<char*>(&other->_impl_.root_));
}

::google::protobuf::Metadata EncryptionVariant::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Encryption::Encryption(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Encryption_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.Encryption)
}
PROTOBUF_NDEBUG_INLINE Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::Encryption& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        mask_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Encryption,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.mask_)>()
          , from.mask_
        }
        ,
        key_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Encryption,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.key_)>()
          , from.key_
        }
        ,
        variants_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Encryption,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.variants_)>()
          , from.variants_
        }
     {}

Encryption::Encryption(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Encryption& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Encryption_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Encryption* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.key_provider_ = from._impl_.key_provider_;

  // @@protoc_insertion_point(copy_constructor:orc.proto.Encryption)
}
PROTOBUF_NDEBUG_INLINE Encryption::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        mask_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.mask_)>()
         }
        ,
        key_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.key_)>()
         }
        ,
        variants_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Encryption,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Encryption, _impl_.variants_)>()
         }
     {}

inline void Encryption::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.key_provider_ = {};
}
Encryption::~Encryption() {
  // @@protoc_insertion_point(destructor:orc.proto.Encryption)
  SharedDtor(*this);
}
inline void Encryption::SharedDtor(MessageLite& self) {
  Encryption& this_ = static_cast<Encryption&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Encryption_class_data_ =
        Encryption::InternalGenerateClassData_(Encryption_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Encryption::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Encryption_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Encryption_class_data_.tc_table);
  return Encryption_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Encryption::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Encryption_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Encryption_globals_));
  return Encryption_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Encryption::ParseTableT_
    Encryption::_table_ =
        Encryption::InternalGenerateParseTable_(Encryption_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Encryption::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.Encryption)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.mask_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.key_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.variants_.Clear();
    }
  }
  _impl_.key_provider_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Encryption::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Encryption& this_ = static_cast<const Encryption&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Encryption::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Encryption& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.Encryption)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .orc.proto.DataMask mask = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_mask_size());
         i < n; i++) {
      const auto& repfield = this_._internal_mask().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.EncryptionKey key = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_key_size());
         i < n; i++) {
      const auto& repfield = this_._internal_key().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              2, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.EncryptionVariant variants = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_variants_size());
         i < n; i++) {
      const auto& repfield = this_._internal_variants().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              3, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional .orc.proto.KeyProviderKind key_provider = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        4, this_._internal_key_provider(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.Encryption)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Encryption::ByteSizeLong(const MessageLite& base) {
  const Encryption& this_ = static_cast<const Encryption&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Encryption::ByteSizeLong() const {
  const Encryption& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.Encryption)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // repeated .orc.proto.DataMask mask = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_mask_size();
      for (const auto& msg : this_._internal_mask()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.EncryptionKey key = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_key_size();
      for (const auto& msg : this_._internal_key()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.EncryptionVariant variants = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_variants_size();
      for (const auto& msg : this_._internal_variants()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // optional .orc.proto.KeyProviderKind key_provider = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_key_provider());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Encryption::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Encryption*>(&to_msg);
  auto& from = static_cast<const Encryption&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.Encryption)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_mask()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_mask());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_key()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_key());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_variants()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_variants());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.key_provider_ = from._impl_.key_provider_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Encryption::CopyFrom(const Encryption& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.Encryption)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Encryption::InternalSwap(Encryption* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.mask_.InternalSwap(&other->_impl_.mask_);
  _impl_.key_.InternalSwap(&other->_impl_.key_);
  _impl_.variants_.InternalSwap(&other->_impl_.variants_);
  swap(_impl_.key_provider_, other->_impl_.key_provider_);
}

::google::protobuf::Metadata Encryption::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Footer::Footer(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Footer_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.Footer)
}
PROTOBUF_NDEBUG_INLINE Footer::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::Footer& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        stripes_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Footer,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.stripes_)>()
          , from.stripes_
        }
        ,
        types_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Footer,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.types_)>()
          , from.types_
        }
        ,
        metadata_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Footer,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.metadata_)>()
          , from.metadata_
        }
        ,
        statistics_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::Footer,
              PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.statistics_)>()
          , from.statistics_
        }
        ,
        software_version_(arena, from.software_version_) {}

Footer::Footer(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Footer& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Footer_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Footer* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.encryption_ = (CheckHasBit(cached_has_bits, 0x00000020U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.encryption_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, header_length_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, header_length_),
           offsetof(Impl_, calendar_) -
               offsetof(Impl_, header_length_) +
               sizeof(Impl_::calendar_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.Footer)
}
PROTOBUF_NDEBUG_INLINE Footer::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        stripes_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.stripes_)>()
         }
        ,
        types_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.types_)>()
         }
        ,
        metadata_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.metadata_)>()
         }
        ,
        statistics_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::Footer,
            PROTOBUF_FIELD_OFFSET(::orc::proto::Footer, _impl_.statistics_)>()
         }
        ,
        software_version_(arena) {}

inline void Footer::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, encryption_),
           0,
           offsetof(Impl_, calendar_) -
               offsetof(Impl_, encryption_) +
               sizeof(Impl_::calendar_));
}
Footer::~Footer() {
  // @@protoc_insertion_point(destructor:orc.proto.Footer)
  SharedDtor(*this);
}
inline void Footer::SharedDtor(MessageLite& self) {
  Footer& this_ = static_cast<Footer&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.software_version_.Destroy();
  delete this_._impl_.encryption_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Footer_class_data_ =
        Footer::InternalGenerateClassData_(Footer_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Footer::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Footer_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Footer_class_data_.tc_table);
  return Footer_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Footer::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Footer_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Footer_globals_));
  return Footer_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Footer::ParseTableT_
    Footer::_table_ =
        Footer::InternalGenerateParseTable_(Footer_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Footer::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.Footer)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.stripes_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.types_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.metadata_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.statistics_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.software_version_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      ABSL_DCHECK(_impl_.encryption_ != nullptr);
      _impl_.encryption_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x000000c0U)) {
    ::memset(&_impl_.header_length_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.content_length_) -
        reinterpret_cast<char*>(&_impl_.header_length_)) + sizeof(_impl_.content_length_));
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000f00U)) {
    ::memset(&_impl_.number_of_rows_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.calendar_) -
        reinterpret_cast<char*>(&_impl_.number_of_rows_)) + sizeof(_impl_.calendar_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Footer::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Footer& this_ = static_cast<const Footer&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Footer::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Footer& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.Footer)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint64 header_length = 1;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this_._internal_header_length(), target);
  }

  // optional uint64 content_length = 2;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this_._internal_content_length(), target);
  }

  // repeated .orc.proto.StripeInformation stripes = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_stripes_size());
         i < n; i++) {
      const auto& repfield = this_._internal_stripes().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              3, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.Type types = 4;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_types_size());
         i < n; i++) {
      const auto& repfield = this_._internal_types().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              4, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // repeated .orc.proto.UserMetadataItem metadata = 5;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_metadata_size());
         i < n; i++) {
      const auto& repfield = this_._internal_metadata().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              5, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional uint64 number_of_rows = 6;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        6, this_._internal_number_of_rows(), target);
  }

  // repeated .orc.proto.ColumnStatistics statistics = 7;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_statistics_size());
         i < n; i++) {
      const auto& repfield = this_._internal_statistics().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              7, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // optional uint32 row_index_stride = 8;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        8, this_._internal_row_index_stride(), target);
  }

  // optional uint32 writer = 9;
  if (CheckHasBit(cached_has_bits, 0x00000400U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        9, this_._internal_writer(), target);
  }

  // optional .orc.proto.Encryption encryption = 10;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        10, *this_._impl_.encryption_, this_._impl_.encryption_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.CalendarKind calendar = 11;
  if (CheckHasBit(cached_has_bits, 0x00000800U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        11, this_._internal_calendar(), target);
  }

  // optional string software_version = 12;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    const ::std::string& _s = this_._internal_software_version();
    target = stream->WriteStringMaybeAliased(12, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.Footer)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Footer::ByteSizeLong(const MessageLite& base) {
  const Footer& this_ = static_cast<const Footer&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Footer::ByteSizeLong() const {
  const Footer& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.Footer)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated .orc.proto.StripeInformation stripes = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_stripes_size();
      for (const auto& msg : this_._internal_stripes()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.Type types = 4;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1UL * this_._internal_types_size();
      for (const auto& msg : this_._internal_types()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.UserMetadataItem metadata = 5;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1UL * this_._internal_metadata_size();
      for (const auto& msg : this_._internal_metadata()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // repeated .orc.proto.ColumnStatistics statistics = 7;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1UL * this_._internal_statistics_size();
      for (const auto& msg : this_._internal_statistics()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // optional string software_version = 12;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_software_version());
    }
    // optional .orc.proto.Encryption encryption = 10;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.encryption_);
    }
    // optional uint64 header_length = 1;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_header_length());
    }
    // optional uint64 content_length = 2;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_content_length());
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000f00U)) {
    // optional uint64 number_of_rows = 6;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_number_of_rows());
    }
    // optional uint32 row_index_stride = 8;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_row_index_stride());
    }
    // optional uint32 writer = 9;
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_writer());
    }
    // optional .orc.proto.CalendarKind calendar = 11;
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_calendar());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Footer::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Footer*>(&to_msg);
  auto& from = static_cast<const Footer&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.Footer)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_stripes()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_stripes());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_types()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_types());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_internal_mutable_metadata()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_metadata());
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_internal_mutable_statistics()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_statistics());
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_internal_set_software_version(from._internal_software_version());
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      ABSL_DCHECK(from._impl_.encryption_ != nullptr);
      if (_this->_impl_.encryption_ == nullptr) {
        _this->_impl_.encryption_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.encryption_);
      } else {
        _this->_impl_.encryption_->MergeFrom(*from._impl_.encryption_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.header_length_ = from._impl_.header_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _this->_impl_.content_length_ = from._impl_.content_length_;
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000f00U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      _this->_impl_.number_of_rows_ = from._impl_.number_of_rows_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      _this->_impl_.row_index_stride_ = from._impl_.row_index_stride_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000400U)) {
      _this->_impl_.writer_ = from._impl_.writer_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000800U)) {
      _this->_impl_.calendar_ = from._impl_.calendar_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Footer::CopyFrom(const Footer& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.Footer)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Footer::InternalSwap(Footer* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.stripes_.InternalSwap(&other->_impl_.stripes_);
  _impl_.types_.InternalSwap(&other->_impl_.types_);
  _impl_.metadata_.InternalSwap(&other->_impl_.metadata_);
  _impl_.statistics_.InternalSwap(&other->_impl_.statistics_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.software_version_, &other->_impl_.software_version_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Footer, _impl_.calendar_)
      + sizeof(Footer::_impl_.calendar_)
      - PROTOBUF_FIELD_OFFSET(Footer, _impl_.encryption_)>(
          reinterpret_cast<char*>(&_impl_.encryption_),
          reinterpret_cast<char*>(&other->_impl_.encryption_));
}

::google::protobuf::Metadata Footer::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

PostScript::PostScript(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PostScript_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.PostScript)
}
PROTOBUF_NDEBUG_INLINE PostScript::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::PostScript& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        version_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::orc::proto::PostScript,
              PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.version_)>()
          , from.version_
        }
        ,
        _version_cached_byte_size_{0},
        magic_(arena, from.magic_) {}

PostScript::PostScript(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const PostScript& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, PostScript_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  PostScript* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, footer_length_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, footer_length_),
           offsetof(Impl_, stripe_statistics_length_) -
               offsetof(Impl_, footer_length_) +
               sizeof(Impl_::stripe_statistics_length_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.PostScript)
}
PROTOBUF_NDEBUG_INLINE PostScript::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        version_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::orc::proto::PostScript,
            PROTOBUF_FIELD_OFFSET(::orc::proto::PostScript, _impl_.version_)>()
         }
        ,
        _version_cached_byte_size_{0},
        magic_(arena) {}

inline void PostScript::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, footer_length_),
           0,
           offsetof(Impl_, stripe_statistics_length_) -
               offsetof(Impl_, footer_length_) +
               sizeof(Impl_::stripe_statistics_length_));
}
PostScript::~PostScript() {
  // @@protoc_insertion_point(destructor:orc.proto.PostScript)
  SharedDtor(*this);
}
inline void PostScript::SharedDtor(MessageLite& self) {
  PostScript& this_ = static_cast<PostScript&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.magic_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull PostScript_class_data_ =
        PostScript::InternalGenerateClassData_(PostScript_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
PostScript::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&PostScript_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(PostScript_class_data_.tc_table);
  return PostScript_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
PostScript::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&PostScript_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&PostScript_globals_));
  return PostScript_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const PostScript::ParseTableT_
    PostScript::_table_ =
        PostScript::InternalGenerateParseTable_(PostScript_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void PostScript::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.PostScript)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.version_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.magic_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x000000fcU)) {
    ::memset(&_impl_.footer_length_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.stripe_statistics_length_) -
        reinterpret_cast<char*>(&_impl_.footer_length_)) + sizeof(_impl_.stripe_statistics_length_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL PostScript::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const PostScript& this_ = static_cast<const PostScript&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL PostScript::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const PostScript& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.PostScript)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional uint64 footer_length = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this_._internal_footer_length(), target);
  }

  // optional .orc.proto.CompressionKind compression = 2;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteEnumToArray(
        2, this_._internal_compression(), target);
  }

  // optional uint64 compression_block_size = 3;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this_._internal_compression_block_size(), target);
  }

  // repeated uint32 version = 4 [packed = true];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    {
      int byte_size = this_._impl_._version_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteUInt32Packed(
            4, this_._internal_version(), byte_size, target);
      }
    }
  }

  // optional uint64 metadata_length = 5;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this_._internal_metadata_length(), target);
  }

  // optional uint32 writer_version = 6;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this_._internal_writer_version(), target);
  }

  // optional uint64 stripe_statistics_length = 7;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        7, this_._internal_stripe_statistics_length(), target);
  }

  // optional string magic = 8000;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    const ::std::string& _s = this_._internal_magic();
    target = stream->WriteStringMaybeAliased(8000, _s, target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.PostScript)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t PostScript::ByteSizeLong(const MessageLite& base) {
  const PostScript& this_ = static_cast<const PostScript&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t PostScript::ByteSizeLong() const {
  const PostScript& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.PostScript)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated uint32 version = 4 [packed = true];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          ::_pbi::WireFormatLite::UInt32SizeWithPackedTagSize(
              this_._internal_version(), 1,
              this_._impl_._version_cached_byte_size_);
    }
    // optional string magic = 8000;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 3 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this_._internal_magic());
    }
    // optional uint64 footer_length = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_footer_length());
    }
    // optional uint64 compression_block_size = 3;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_compression_block_size());
    }
    // optional .orc.proto.CompressionKind compression = 2;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 +
                    ::_pbi::WireFormatLite::EnumSize(this_._internal_compression());
    }
    // optional uint32 writer_version = 6;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this_._internal_writer_version());
    }
    // optional uint64 metadata_length = 5;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_metadata_length());
    }
    // optional uint64 stripe_statistics_length = 7;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_stripe_statistics_length());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void PostScript::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<PostScript*>(&to_msg);
  auto& from = static_cast<const PostScript&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.PostScript)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_version()->MergeFrom(from._internal_version());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_set_magic(from._internal_magic());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.footer_length_ = from._impl_.footer_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.compression_block_size_ = from._impl_.compression_block_size_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _this->_impl_.compression_ = from._impl_.compression_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      _this->_impl_.writer_version_ = from._impl_.writer_version_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      _this->_impl_.metadata_length_ = from._impl_.metadata_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      _this->_impl_.stripe_statistics_length_ = from._impl_.stripe_statistics_length_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void PostScript::CopyFrom(const PostScript& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.PostScript)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void PostScript::InternalSwap(PostScript* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.version_.InternalSwap(&other->_impl_.version_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.magic_, &other->_impl_.magic_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(PostScript, _impl_.stripe_statistics_length_)
      + sizeof(PostScript::_impl_.stripe_statistics_length_)
      - PROTOBUF_FIELD_OFFSET(PostScript, _impl_.footer_length_)>(
          reinterpret_cast<char*>(&_impl_.footer_length_),
          reinterpret_cast<char*>(&other->_impl_.footer_length_));
}

::google::protobuf::Metadata PostScript::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

FileTail::FileTail(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FileTail_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:orc.proto.FileTail)
}
PROTOBUF_NDEBUG_INLINE FileTail::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::orc::proto::FileTail& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

FileTail::FileTail(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const FileTail& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, FileTail_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  FileTail* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.postscript_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.postscript_)
                : nullptr;
  _impl_.footer_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.footer_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, file_length_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, file_length_),
           offsetof(Impl_, postscript_length_) -
               offsetof(Impl_, file_length_) +
               sizeof(Impl_::postscript_length_));

  // @@protoc_insertion_point(copy_constructor:orc.proto.FileTail)
}
PROTOBUF_NDEBUG_INLINE FileTail::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void FileTail::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, postscript_),
           0,
           offsetof(Impl_, postscript_length_) -
               offsetof(Impl_, postscript_) +
               sizeof(Impl_::postscript_length_));
}
FileTail::~FileTail() {
  // @@protoc_insertion_point(destructor:orc.proto.FileTail)
  SharedDtor(*this);
}
inline void FileTail::SharedDtor(MessageLite& self) {
  FileTail& this_ = static_cast<FileTail&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.postscript_;
  delete this_._impl_.footer_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull FileTail_class_data_ =
        FileTail::InternalGenerateClassData_(FileTail_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FileTail::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FileTail_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(FileTail_class_data_.tc_table);
  return FileTail_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
FileTail::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&FileTail_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&FileTail_globals_));
  return FileTail_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const FileTail::ParseTableT_
    FileTail::_table_ =
        FileTail::InternalGenerateParseTable_(FileTail_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void FileTail::Clear() {
// @@protoc_insertion_point(message_clear_start:orc.proto.FileTail)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(_impl_.postscript_ != nullptr);
      _impl_.postscript_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.footer_ != nullptr);
      _impl_.footer_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000000cU)) {
    ::memset(&_impl_.file_length_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.postscript_length_) -
        reinterpret_cast<char*>(&_impl_.file_length_)) + sizeof(_impl_.postscript_length_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL FileTail::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const FileTail& this_ = static_cast<const FileTail&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL FileTail::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const FileTail& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:orc.proto.FileTail)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // optional .orc.proto.PostScript postscript = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        1, *this_._impl_.postscript_, this_._impl_.postscript_->GetCachedSize(), target,
        stream);
  }

  // optional .orc.proto.Footer footer = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.footer_, this_._impl_.footer_->GetCachedSize(), target,
        stream);
  }

  // optional uint64 file_length = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this_._internal_file_length(), target);
  }

  // optional uint64 postscript_length = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this_._internal_postscript_length(), target);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:orc.proto.FileTail)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t FileTail::ByteSizeLong(const MessageLite& base) {
  const FileTail& this_ = static_cast<const FileTail&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t FileTail::ByteSizeLong() const {
  const FileTail& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:orc.proto.FileTail)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // optional .orc.proto.PostScript postscript = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.postscript_);
    }
    // optional .orc.proto.Footer footer = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.footer_);
    }
    // optional uint64 file_length = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_file_length());
    }
    // optional uint64 postscript_length = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this_._internal_postscript_length());
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void FileTail::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<FileTail*>(&to_msg);
  auto& from = static_cast<const FileTail&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:orc.proto.FileTail)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.postscript_ != nullptr);
      if (_this->_impl_.postscript_ == nullptr) {
        _this->_impl_.postscript_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.postscript_);
      } else {
        _this->_impl_.postscript_->MergeFrom(*from._impl_.postscript_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.footer_ != nullptr);
      if (_this->_impl_.footer_ == nullptr) {
        _this->_impl_.footer_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.footer_);
      } else {
        _this->_impl_.footer_->MergeFrom(*from._impl_.footer_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _this->_impl_.file_length_ = from._impl_.file_length_;
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _this->_impl_.postscript_length_ = from._impl_.postscript_length_;
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void FileTail::CopyFrom(const FileTail& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:orc.proto.FileTail)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void FileTail::InternalSwap(FileTail* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_length_)
      + sizeof(FileTail::_impl_.postscript_length_)
      - PROTOBUF_FIELD_OFFSET(FileTail, _impl_.postscript_)>(
          reinterpret_cast<char*>(&_impl_.postscript_),
          reinterpret_cast<char*>(&other->_impl_.postscript_));
}

::google::protobuf::Metadata FileTail::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace proto
}  // namespace orc
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_orc_5fproto_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
